/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_dev.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
 * sdla_dev.c	SDLA DEV Support Module.  Main module.
 *
 *		This module is a compeled with kernel sources.
 *
 * Author:	Alex Feldman	<al.feldman@sangoma.com>
 ****************************************************************************
 */

/*#define SDLA_DEV_DEBUG */ 		/* uncomment this for debug messages */
/*#define INTR_TEST*/		/* uncomment this to check that driver receive intr */

#include "sdla.h"		/* number of Sangoma cards */

#include <stddef.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/interrupt.h>

#include <sys/sysent.h>

#include <../../include/net/sdlapci.h>
#include <../../include/net/sdlasfm.h>
#include <../../include/net/sdla_dev.h>	/* special header for S508 board */

#include <i386/isa/isa_device.h>

#include <vm/vm.h>
#include <vm/pmap.h>
 
#include <pci/pcivar.h>
#include <pci/pcireg.h>

/* hardware related constant */
#define MAX_CARDS         8
#define S508_NUMPORTS     2
#define S508_ID_BYTE      0x11

/* SDLA ISA global constants */
#define MOD_VERSION       1
#define MOD_RELEASE       1
#define MOD_MINOR         0

/* Maximum number cards (each CPU like separate card ) */
#define MAX_SDLA_CARD NSDLA*2
/****** Global Data *******************************************************/

int       number_cards = 0;
sdladev_t sdladev[MAX_SDLA_CARD];

/* SDLA ISA device related entry point */
static int 	   sdla_isa_probe(struct isa_device *dev);
static int 	   sdla_isa_attach(struct isa_device *dev);
static ointhand2_t sdla_isa_intr;

/* isa_driver structure for autoconfiguration process */
struct isa_driver sdladriver = {
	sdla_isa_probe, sdla_isa_attach, SDLA_DEV_NAME
};

/* SDLA PCI device relative entry point */
static unsigned long sdla_pci_count = 0;
unsigned int  number_S514_cards = 0; 
static const char    *sdla_pci_probe __P((pcici_t, pcidi_t));
static void 	     sdla_pci_attach __P((pcici_t, int));
static pci_inthand_t sdla_pci_intr;

static struct pci_device sdlapcidev = {
	SDLA_DEV_NAME,
	sdla_pci_probe,
	sdla_pci_attach,
	&sdla_pci_count,
	NULL,
};
DATA_SET(pcidevice_set, sdlapcidev);

/* For debugging convenience */
#ifdef SDLA_DEV_DEBUG
 #define dlog(x) log x
#else
 #define dlog(x)
#endif /* SDLA_DEV_DEBUG */

/******* ISA Device Driver Entry Points ********************************/
/*
 ****************************************************************************
 * sdla_isa_probe() - called during system's autoconfiguration to probe for
 * 		     device presence. The function should return the number
 * 		     of i/o ports needed, or 0 if device not present.
 */
static int
sdla_isa_probe(struct isa_device *dev)
{
    int isa_unit = dev->id_unit;

    /* Check that number of cards not bigger than size of array (MAX_SDLA_CARD) */
    if (number_cards >= MAX_SDLA_CARD)
	return 0; 
    /* Check if the unit makes sense. */ 
    if (isa_unit >= MAX_CARDS)
    { 
        log(LOG_ERR, "%s? - bad unit number (%d)!\n", SDLA_DEV_NAME, isa_unit);
	return 0;
    }
    /* Grab some info */
    sdladev[number_cards].type           = SDLA_S508;
    sdladev[number_cards].isr            = NULL;
    sdladev[number_cards].u.isa.isa_unit = isa_unit;
    sdladev[number_cards].u.isa.irq      = ffs(dev->id_irq) - 1;
    sdladev[number_cards].u.isa.maddr    = dev->id_maddr; /* virtual address */
    sdladev[number_cards].u.isa.ioport   = dev->id_iobase; 
    
    /* Try see if device is these. Read iobase+3and check if se see
     * 0x11 which S508's identifier.
     */
    if (s508_read_port(dev->id_iobase, 3) != S508_ID_BYTE)
        return 0;

    /* OK, device is there, return number of ports */
    return S508_NUMPORTS; 	
}

/*
 **************************************************************************
 * sdla_isa_attach() - called if the probe is succeeded. Attaches wanpipe?
 *                    device to the stack.
 */
static int
sdla_isa_attach(struct isa_device *dev)
{
    if (number_cards >= MAX_SDLA_CARD)
	return 0;
    dev->id_ointr = sdla_isa_intr;
    sdladev[number_cards].isr = NULL;
    sdladev[number_cards].arg = NULL;
    /* increment number of SDLA adapters */
    number_cards++;
    return 1;
}

/*
 ****************************************************************************
 * sdla_isa_intr() - Dummy ISA device interrupt handler.
 */
static void sdla_isa_intr(int isa_unit)
{
    sdladev_t *isa_cfg = NULL;
    int              i = 0;
 
#ifdef INTR_TEST
    log(LOG_INFO, "%s%d: Interrupt from S508 Adapter!\n", 
						SDLA_DEV_NAME, isa_unit);
#endif
    if (isa_unit >= number_cards)
	return;
    for(i = 0; (i < number_cards && isa_cfg == NULL); i++)
    {
	if (sdladev[i].type != SDLA_S508)
	    continue;
	if (sdladev[i].u.isa.isa_unit != isa_unit)
	    continue;
	isa_cfg = &sdladev[i]; 
    }

    if (isa_cfg == NULL)
    {
	log(LOG_ERR, "%s%d: Uncorrect addapter type!\n", SDLA_DEV_NAME, isa_unit);
	return; 
    }
    if (isa_cfg->isr && (isa_cfg->arg != NULL))
	isa_cfg->isr(isa_cfg->arg);
    return;
}


/******* PCI Device Driver Entry Points ********************************/

static const char *sdla_pci_probe(pcici_t config_id, pcidi_t device_id)
{
    uint16_t PCI_subsys_vendor;

    /* Check that number of cards not bigger than size of array (MAX_SDLA_CARD) */
    if (number_cards >= MAX_SDLA_CARD)
	return NULL;
    if (number_S514_cards >= MAX_S514_CARDS)
    {
	log(LOG_ERR, "Unsupported number of PCI cards!\n");
	return NULL;
    }
    if ((device_id & 0xFFFF) != V3_VENDOR_ID || 
	(device_id >>16) & 0xFFFF != V3_DEVICE_ID)
	return NULL;
    PCI_subsys_vendor = pci_conf_read(config_id, PCI_SUBSYS_VENDOR_WORD);
    if (PCI_subsys_vendor != SANGOMA_SUBSYS_VENDOR)
	return NULL;

    return ("SDLA S514 Adapter");
}

static void sdla_pci_attach(pcici_t config_id, int pci_unit)
{
    sdladev_t      *pci_cfg = NULL,
               *pci_cfg_nxt = NULL;

    if (number_cards+1 >= MAX_SDLA_CARD)
    {
	log(LOG_ERR, "%s%d: Error in kernel configuration file.\n",
				SDLA_DEV_NAME, pci_unit);
	return;
    }
    /* Map interrupt handler for first CPU */
    pci_cfg = &sdladev[number_cards];
    pci_cfg->type         = SDLA_S514;
    pci_cfg->isr          = NULL;
    pci_cfg->u.pci.cfg    = config_id;
    pci_cfg->u.pci.cpu_no = ' ';
    if (!pci_map_int(config_id, sdla_pci_intr, (void*)pci_cfg, &net_imask))
    {
	log(LOG_ERR, "%s%d: Couldn't map interrupt!\n", 
						SDLA_DEV_NAME, pci_unit);
	return;
    }
 
    /* Map interrupt handler for second CPU */
    pci_cfg_nxt = &sdladev[number_cards+1];
    pci_cfg_nxt->type         = SDLA_S514;
    pci_cfg_nxt->isr          = NULL;
    pci_cfg_nxt->u.pci.cfg    = config_id;
    pci_cfg_nxt->u.pci.cpu_no = ' ';
    if (!pci_map_int(config_id, sdla_pci_intr, (void*)pci_cfg_nxt, &net_imask))
    {
	log(LOG_ERR, "%s%d: Couldn't map interrupt!\n", 
						SDLA_DEV_NAME, pci_unit);
	return;
    }
    /* Increment number of SDLA adapters, include CPU_A and CPU_B */
    number_cards += 2;
    /* Incremenet number of S514 cards */
    number_S514_cards++;
}
static void sdla_pci_intr(void *arg)
{
    sdladev_t *pci_cfg = (sdladev_t*)arg;

    if (pci_cfg == NULL)
    {
	log(LOG_ERR, "%s?: received interrupt from unknown device!\n",
						SDLA_DEV_NAME);
	return;
    }
#ifdef INTR_TEST
    log(LOG_INFO, "%s: Interrupt from S514 Adapter!\n", SDLA_DEV_NAME); 
#endif
    if (pci_cfg->isr != NULL && pci_cfg->arg != NULL)
	pci_cfg->isr(pci_cfg->arg); 
}
/****** End *****************************************************************/
