/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_chdlc.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
 * sdla_chdlc.c	WANPIPE(tm) Multiprotocol WAN Link Driver. Cisco HDLC module.
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ============================================================================
 * Dec 02, 1999 Alex Feldman    Updated to FreeBSD 3.2
 * Aug 07, 1998	David Fong	Initial version.
 ****************************************************************************
 */

#include "bpfilter.h"

#include <stddef.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>

#define _GNUC_
#include <net/wanpipe.h>		/* WANPIPE common user API definision */
#include <net/sdla_chdlc.h>		/* CHDLC firmware API definitions */

#include <netinet/in.h>
#include <netinet/in_var.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <machine/clock.h>
#include <machine/atomic.h>

/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#define WANDEBUG_CHDLC
#else
#define	STATIC		static
#endif

#define SYSTEM_TICK ticks

/* For debugging convenience */
#ifdef WANDEBUG_CHDLC
#define dlog(x) log x
#else
#define dlog(x)
#endif

/* reasons for enabling the timer interrupt on the adapter */
#define TMR_INT_ENABLED_UDP     0x0001
#define TMR_INT_ENABLED_UPDATE  0x0002

#define	CHDLC_DFLT_DATA_LEN	1500		/* default MTU */
#define CHDLC_HDR_LEN		1

#define IFF_POINTTOPOINT 0x10

#define WANPIPE   0x00
#define API       0x01
#define CHDLC_API 0x01

#define PORT(x)  (x == 0 ? "PRIMARY" : "SECONDARY")

/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with CHDLC specific data
 */

typedef struct chdlc_private_area
{
    sdla_t		*card;
    int 		TracingEnabled;		/* For enabling Tracing */
    unsigned long 	curr_trace_addr;	/* Used for Tracing */
    unsigned long 	start_trace_addr;
    unsigned long 	end_trace_addr;
    unsigned long 	base_addr_trace_buffer;
    unsigned long 	end_addr_trace_buffer;
    unsigned short 	number_trace_elements;
    unsigned  		available_buffer_space;

    unsigned long 	router_start_time;
    unsigned char 	route_status;
    unsigned char 	route_removed;
    unsigned long 	tick_counter;		/* For 5s timeout counter */
    unsigned long 	router_up_time;
    uint32_t            IP_address;
    uint32_t            IP_netmask;
    unsigned char       mc;			/* Multicast support on/off */
    unsigned 		short udp_pkt_lgth;
    char 		udp_pkt_src;
    char 		udp_pkt_data[MAX_LGTH_UDP_MGNT_PKT];
    unsigned short 	timer_int_enabled;
    char 		update_comms_stats;             /* update comms stats */
} chdlc_private_area_t;

/* Route Status options */
#define NO_ROUTE	0x00
#define ADD_ROUTE	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE	0x03

/* Array of Sangoma card structure */
extern sdla_t sdla_card_array[];

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;

/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update (wan_device_t* wandev);
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf);
static int del_if (wan_device_t* wandev, struct ifnet* ifp);

/* Network device interface */
static int chdlc_close(struct ifnet *ifp);

/* ifnet entry points */
static int 	chdlc_output		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		     struct rtentry *));
static 	void	chdlc_start		/* initiate output routine */
		__P((struct ifnet *));
static	int	chdlc_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static	void	chdlc_watchdog		/* timer routine */
		__P((struct ifnet *));
static	int	chdlc_init		/* init routine */
		__P((struct ifnet *, struct ifaddr *));

/* CHDLC Firmware interface functions */
static int chdlc_configure 	(sdla_t* card, void* data);
static int chdlc_comm_enable 	(sdla_t* card);
static int chdlc_comm_disable 	(sdla_t* card);
static int chdlc_read_version 	(sdla_t* card, char* str);
static int chdlc_set_intr_mode 	(sdla_t* card, unsigned mode);
static int chdlc_read_intr_mode (sdla_t* card);
static int chdlc_send(sdla_t* card, void* data, unsigned len);
static int chdlc_read_comm_err_stats (sdla_t* card);
static int chdlc_read_configure(sdla_t *card);
static int chdlc_read_op_stats (sdla_t* card);

/* Miscellaneous CHDLC Functions */
static int set_chdlc_config (sdla_t *card, struct ifaddr *ifa);
static void init_chdlc_tx_rx_buff( sdla_t* card, struct ifnet *ifp);
static int chdlc_error (sdla_t *card, int err, CHDLC_MAILBOX_STRUCT *mb);
static int process_chdlc_exception(sdla_t *card);
static int process_global_exception(sdla_t *card);
static int update_comms_stats(sdla_t* card, 
		  chdlc_private_area_t* chdlc_priv_area);
static int configure_ip (sdla_t* card);
static int unconfigure_ip (sdla_t* card);
static void process_route(void *arg);
static void port_set_state(sdla_t *card, int);

/* Interrupt handlers */
static void wpc_isr (sdla_t* card);
static void rx_intr (sdla_t* card);
static void timer_intr(sdla_t *);

/* Miscellaneous functions */
static int chk_bcast_mcast_addr(sdla_t* card, struct ifnet* ifp,
							    struct mbuf *m);
static int reply_udp(unsigned char *data, unsigned int mbox_len );
static int chdlc_intr_test(sdla_t* card, struct ifnet *ifp);
static int udp_pkt_type(caddr_t pkt, sdla_t* card);
static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, struct ifnet* ifp,
                                chdlc_private_area_t* chdlc_priv_area);
static int process_udp_mgmt_pkt(sdla_t* card, struct ifnet *ifp,
				chdlc_private_area_t* chdlc_priv_area); 
static unsigned short calc_checksum(char*, int);

/* Miscellaneous utilities function */
#ifdef WANDEBUG_CHDLC
static void chdlc_dump_mailbox(CHDLC_MAILBOX_STRUCT *);
static void chdlc_dump_udp(chdlc_udp_pkt_t *, unsigned int);
#endif

static int  Intr_test_counter;

int if_flag = 0;
/****** Public Functions ****************************************************/

/* 
 * ============================================================================
 * Cisco HDLC protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpc_init (sdla_t* card, wandev_conf_t* conf)
{
    volatile CHDLC_MAILBOX_STRUCT* mb = NULL;
    CHDLC_MAILBOX_STRUCT* mb1 = NULL;
    unsigned int timeout;
    unsigned char port_num;
    unsigned long max_permitted_baud = 0;
    union{
	char str[80];
    } u;
    int err;

    dlog((LOG_INFO, "%s: Starting CHDLC initialization\n", card->wandev.name));
    /* Verify configuration ID */
    if (conf->config_id != WANCONFIG_CHDLC) 
    {
        log(LOG_ERR, "%s: invalid configuration ID %u!\n",
		 			card->devname, conf->config_id);
	return -EINVAL;
    }

    /* Find out which Port to use */
    if ((conf->comm_port == WANOPT_PRI) || (conf->comm_port == WANOPT_SEC))
    {
	if (card->next)
	{
	    if (conf->comm_port != card->next->u.c.comm_port)
		card->u.c.comm_port = conf->comm_port;
	    else
	    {
		log(LOG_ERR, "%s: %s port busy: already in use! \n", 
				card->wandev.name, PORT(conf->comm_port));
		return -EINVAL;
	    }
	}
	else
	    card->u.c.comm_port = conf->comm_port;
    } 
    else
    {
	log(LOG_ERR, "%s: Invalid Port selected!\n", card->wandev.name);
	return -EINVAL;
    }
	
    /* Initialize protocol-specific fields */
    if (card->hw.type != SDLA_S514)
    {
   	if (card->u.c.comm_port == WANOPT_PRI)
	    card->mbox = (void*)card->hw.dpmbase;
	else
	    card->mbox = (void*)((u_char*)card->hw.dpmbase + 
			SEC_BASE_ADDR_MB_STRUCT - PRI_BASE_ADDR_MB_STRUCT);
    }
    else
    {
        /* for a S514 adapter, set a pointer to the actual mailbox in the */
	/* allocated virtual memory area */
	if (card->u.c.comm_port == WANOPT_PRI)
	    card->mbox = 
		(void*)((u_char*)card->hw.dpmbase + PRI_BASE_ADDR_MB_STRUCT);
	else
	    card->mbox = 
		(void*)((u_char*)card->hw.dpmbase + SEC_BASE_ADDR_MB_STRUCT);
    }

    mb = mb1 = card->mbox;

    port_num = card->u.c.comm_port;
    if (!card->configured && !card->wandev.piggyback)
    {
        /* The board will place an 'I' in the return code to indicate that it is
         * ready to accept commands.  We expect this to be completed in less
         * than 1 second. 
         */
        timeout = SYSTEM_TICK;
        while(mb->return_code != 'I') /* wait 1s for board initialization */
            if ((SYSTEM_TICK - timeout) > 1*hz) break;
        if (mb->return_code != 'I') 
        {
            log(LOG_ERR, "%s: Initialization not completed by adapter\n", 
						card->devname);
            log(LOG_ERR, "Please contact Sangoma representative.\n");
            return -EIO;
        }
        dlog((LOG_INFO, "%s: Received I from the board!!! Great\n",
						card->devname));
    }

    /* Read firmware version.  Note that when adapter initializes, it
     * clears the mailbox, so it may appear that the first command was
     * executed successfully when in fact it was merely erased. To work
     * around this, we execute the first command twice.
     */

    if (chdlc_read_version(card, u.str))
    {	
	return -EIO;
    }
    log(LOG_INFO, "%s: running Cisco HDLC firmware v%s\n", card->devname, u.str); 

    card->isr			= &wpc_isr;
    card->poll			= NULL;
    card->exec			= NULL;
    card->wandev.update		= &update;
    card->wandev.new_if		= &new_if;
    card->wandev.del_if		= &del_if;
    card->wandev.udp_port   = conf->udp_port;

    /* Set the flag pointers for both ports to NULL */
    card->u.c.flags = NULL;

    /* reset the number of times the 'update()' proc has been called */
    card->u.c.update_call_count = 0;

    card->wandev.ttl = conf->ttl;
    card->wandev.interface = conf->interface;

    if (card->u.c.comm_port == WANOPT_SEC && conf->interface == WANOPT_V35 &&
	card->hw.type != SDLA_S514)
    {
	log(LOG_ERR, "%s: V35 Interface not supported on S508 %s Port!\n",
			card->devname, PORT(card->u.c.comm_port));
	return -EIO;
    } 

    card->wandev.clocking = conf->clocking;

    if (card->wandev.clocking)
    {
	if (port_num == WANOPT_PRI) /* For primary port */
	    max_permitted_baud = 
		(card->hw.type == SDLA_S514) ? PRI_MAX_BAUD_RATE_S514 :
					       PRI_MAX_BAUD_RATE_S508;	
	else 
	    if (port_num == WANOPT_SEC) /* For secondary port */
	        max_permitted_baud = 
		    (card->hw.type == SDLA_S514) ? SEC_MAX_BAUD_RATE_S514 :
			  		           SEC_MAX_BAUD_RATE_S508;	
	if (conf->bps > max_permitted_baud)
	{
	    conf->bps = max_permitted_baud;
	    log(LOG_ERR, "%s: Baud too high!!!\n", card->wandev.name);
	    log(LOG_ERR, "%s: Baud rate set to %lu bps.\n",
			card->wandev.name, max_permitted_baud);
	}
	card->wandev.bps = conf->bps;
    }
    else
	card->wandev.bps = 0;
	
    /* Setup the Port MTU */
    if (port_num == WANOPT_PRI) /* For primary port 0 */
	card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?
			min(conf->mtu, PRI_MAX_NO_DATA_BYTES_IN_FRAME) :
			CHDLC_DFLT_DATA_LEN;
    else
	if (port_num == WANOPT_SEC) /* For secondary port 1 */
	    card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?
			min(conf->mtu, SEC_MAX_NO_DATA_BYTES_IN_FRAME) :
			CHDLC_DFLT_DATA_LEN;

    /* Set up the interrupt status area. 
     * Read the CHDLC_configuration and obtain:
     * Ptr to sahred memory info struct.
     * Use this pointer to calculate the value of card->u.c.flags!
     */
    mb1->buffer_length = 0;
    mb1->command = READ_CHDLC_CONFIGURATION;
    err = sdla_exec(mb1) ? mb1->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
    {
	chdlc_error(card, err, mb1);
	return -EIO;
    }

    if (card->hw.type == SDLA_S514)
        card->u.c.flags = (void*)((u_char*)card->hw.dpmbase +
		(((CHDLC_CONFIGURATION_STRUCT*)mb1->data)->
		ptr_shared_mem_info_struct));
    else
        card->u.c.flags = (void*)((u_char*)card->hw.dpmbase +
		(((CHDLC_CONFIGURATION_STRUCT*)mb1->data)->
		ptr_shared_mem_info_struct % SDLA_WINDOWSIZE));

    /* This is for the ports link state */
    card->wandev.state		= WAN_DUALPORT;
    card->u.c.state = WAN_DISCONNECTED;

    return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/* 
 * ============================================================================
 * Update device status & statistics (called when updating the PROC file system)
 */
static int update (wan_device_t* wandev)
{
    sdla_t* card = NULL;
    SHARED_MEMORY_INFO_STRUCT *flags = NULL;
    volatile chdlc_private_area_t *chdlc_priv_area = NULL;
    unsigned long timeout;
 
    /* sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL))
	return -EFAULT;

    if(wandev->state == WAN_UNCONFIGURED)
	return -ENODEV;

    card = wandev->private;

    /* more sanity checks */
    if (!card->u.c.flags)
	return -ENODEV;

    flags = card->u.c.flags;

    /* Only one interface must be in interface list */
    if (wandev->if_head == NULL || 
	wandev->if_head->ifp == NULL ||
	wandev->if_head->ifp->if_softc == NULL)
	return -ENODEV;
    chdlc_priv_area = wandev->if_head->ifp->if_softc;
    
    if (chdlc_priv_area->update_comms_stats)
	return -EAGAIN;

    chdlc_priv_area->update_comms_stats = 2;
    flags->interrupt_info_struct.interrupt_permission |= APP_INT_ON_TIMER;
    chdlc_priv_area->timer_int_enabled = TMR_INT_ENABLED_UPDATE;

    /* wait maximum of 1 second for the statistics to be updated */
    timeout = SYSTEM_TICK;
    for(;;)
    {
	if (chdlc_priv_area->update_comms_stats == 0)
	    break;
	if ((SYSTEM_TICK - timeout) > (1*hz))
	{
	    chdlc_priv_area->update_comms_stats = 0;
	    chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
	    return -EAGAIN;
	}
    }   
    return 0;
}

/* 
 * ============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf)
{
    sdla_t* card = NULL;
    chdlc_private_area_t* chdlc_priv_area = NULL;

    dlog((LOG_INFO, "%s: Initialize new interface %s\n", wandev->name, conf->name));	
    if ((wandev == NULL) || (wandev->private == NULL))
	return -ENODEV;
    card = wandev->private; 

    if ((conf->name[0] == '\0') || (strlen(conf->name) >= IFNAMSIZ)) 
    {
	log(LOG_ERR, "%s: invalid interface name!\n", card->devname);
	return -EINVAL;
    }
	
    /* allocate and initialize private data */
    chdlc_priv_area = malloc(sizeof(chdlc_private_area_t), M_DEVBUF, M_NOWAIT);
	
    if(chdlc_priv_area == NULL)
    {
        log(LOG_ERR, "%s: Failed allocating private data...\n", card->devname); 
        return -ENOMEM;
    }	
    bzero(chdlc_priv_area, sizeof(chdlc_private_area_t));

    chdlc_priv_area->card = card; 

    /* initialize data */
    sprintf(card->u.c.if_name, "%s%d", conf->name, 0);

    strcpy(card->u.c.if_name, conf->name);

    if (wandev->new_if_cnt > 0)
    {
	log(LOG_ERR, "%s: The one network interface is already initialized!\n", 
								card->devname);
        free(chdlc_priv_area, M_DEVBUF);
        return -EEXIST;	
    }

    wandev->new_if_cnt++;
    chdlc_priv_area->TracingEnabled = 0;
    chdlc_priv_area->route_status = NO_ROUTE;
    chdlc_priv_area->route_removed = 0;
     
    /* Setup protocol options */
    card->u.c.protocol_options = 0;

    if (conf->ignore_dcd == WANOPT_YES)
	card->u.c.protocol_options |= IGNORE_DCD_FOR_LINK_STAT;
    if (conf->ignore_cts == WANOPT_YES)
	card->u.c.protocol_options |= IGNORE_CTS_FOR_LINK_STAT;

    if (conf->ignore_keepalive == WANOPT_YES)
    {
	card->u.c.protocol_options |= IGNORE_KPALV_FOR_LINK_STAT;
	card->u.c.kpalv_tx = MIN_Tx_KPALV_TIMER;
	card->u.c.kpalv_rx = MIN_Rx_KPALV_TIMER;
	card->u.c.kpalv_err = MIN_KPALV_ERR_TOL;
    }
    else
    {  /* Do not ignore keepalives */
 	card->u.c.kpalv_tx = 
		((conf->keepalive_tx_tmr - MIN_Tx_KPALV_TIMER) >= 0) ?
		min(conf->keepalive_tx_tmr, MAX_Tx_KPALV_TIMER) :
		DEFAULT_Tx_KPALV_TIMER;

 	card->u.c.kpalv_rx = 
		((conf->keepalive_rx_tmr - MIN_Rx_KPALV_TIMER) >= 0) ?
		min(conf->keepalive_rx_tmr, MAX_Rx_KPALV_TIMER) :
		DEFAULT_Rx_KPALV_TIMER;

 	card->u.c.kpalv_err = 
		((conf->keepalive_err_margin - MIN_KPALV_ERR_TOL) >= 0) ?
		min(conf->keepalive_err_margin, MAX_KPALV_ERR_TOL) :
		DEFAULT_KPALV_ERR_TOL;
    }

    /* Setup slarp timer to control delay between slarps */
    card->u.c.slarp_timer = 
		((conf->slarp_timer - MIN_SLARP_REQ_TIMER) >= 0) ?
		min(conf->slarp_timer, MAX_SLARP_REQ_TIMER) :
		DEFAULT_SLARP_REQ_TIMER;

    /* If HDLC_STREAMING is enabled then IGNORE DCD, CTS and KEEPALIVES
     * are automatically ignored 
     */
    if (conf->hdlc_streaming == WANOPT_YES)
    {
	log(LOG_INFO, "%s: Enabling HDLC STREAMING Mode.\n", wandev->name);
	card->u.c.protocol_options = HDLC_STREAMING_MODE;
    }

    /* Setup wanpipe as arouter (WANPIPE) or as an API */
    if (strcmp(conf->usedby, "WANPIPE") == 0)
    {
	log(LOG_INFO, "%s: Running in WANPIPE mode!\n", wandev->name);
	card->u.c.usedby = WANPIPE;
    }
    else 
        if (strcmp(conf->usedby, "API") == 0)
        {
#ifdef CHDLC_API
	    log(LOG_INFO, "%s: Running in API mode!\n", wandev->name);
	    card->u.c.usedby = API;
#else
	    log(LOG_ERR, "%s: API mode not supported by the kernel bundled drivers",
								wandev->name);
	    log(LOG_ERR, "\nUse the drivers from Sangoma themselves\n");
            free(chdlc_priv_area, M_DEVBUF);
	    return -EINVAL;
#endif
        }
    /* Get Multicast Information */
    chdlc_priv_area->mc = conf->mc;

    /* prepare network device data space for registration */
    ifp->if_softc          = chdlc_priv_area;
    ifp->if_unit           = 0; /* description in sdlamain.c */
    ifp->if_name           = malloc(strlen(conf->name)+1, M_DEVBUF, M_NOWAIT);
    if (ifp->if_name == NULL)
    {
	log(LOG_ERR, "%s: Cannot allocate memory!\n", conf->name);
	ifp->if_softc = NULL;
        free(chdlc_priv_area, M_DEVBUF);
	return -ENOMEM;
    }
    strcpy(ifp->if_name, conf->name);
    ifp->if_mtu            = wandev->mtu; /* ETHERMTU */
    ifp->if_type           = IFT_OTHER;
    ifp->if_flags          = IFF_POINTOPOINT;
    /* Enable Multicasting if user parameters */
    if (chdlc_priv_area->mc == WANOPT_YES)
	ifp->if_flags |= IFF_MULTICAST; 
    ifp->if_output   	   = chdlc_output;
    ifp->if_start          = chdlc_start;
    ifp->if_watchdog       = chdlc_watchdog;
    ifp->if_init           = (void*)chdlc_init;
    ifp->if_ioctl	   = chdlc_ioctl;
    ifp->if_snd.ifq_len    = 0;
    ifp->if_snd.ifq_maxlen = ifqmaxlen;
	
    /* ???? This members was not initialize in 3.1 */
    ifp->if_hdrlen = CHDLC_HDR_LEN;

    dlog((LOG_INFO, "%s: New interface (%s) initialized\n", 
				card->devname, ifp->if_name));
    return 0;
}

/* 
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
    sdla_t                  *card = wandev->private;

    dlog((LOG_INFO, "%s: Deleting CHDLC interface %s...\n", 
					wandev->name, ifp->if_name));
    if (card == NULL)
	return -ENODEV;

    /* Initialize interrupt handler pointer */
    card->isr = NULL;

    /* Close current interface, set interrupt permission byte to 0 */ 
    chdlc_close(ifp);

    /* Initialize network interface */
    ifp->if_output = NULL;
    ifp->if_start = NULL;
    ifp->if_ioctl = NULL;
    ifp->if_watchdog = NULL;
    ifp->if_init = NULL;
    ifp->if_flags &= ~(IFF_UP | IFF_RUNNING | IFF_OACTIVE);
    if (ifp->if_softc) 
    {
        free(ifp->if_softc, M_DEVBUF);
        ifp->if_softc = NULL;
    }
    dlog((LOG_INFO, "%s: Interface %s is deleted\n", wandev->name, ifp->if_name));
    return 0;
}

/****** Network Device Interface ********************************************/

/* 
 * cpipe_ioctl() -
 *
 */
static int
chdlc_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
    chdlc_private_area_t* chdlc_priv_area = ifp->if_softc;
    sdla_t* card = chdlc_priv_area->card;
    struct ifaddr	      *ifa = (struct ifaddr *)data; /* address struct */
    int 		      s = 0;	    	    /* saved interrupt state */
    int		      error = 0;	    	    /* resulting status */

    s = splimp();

    switch (command) 
    {
    case SIOCSIFADDR:
    	dlog((LOG_INFO, "%s - ioctl(SIOCSIFADDR) called.\n", card->devname));
    	/* check the protocol family - we currently support IP only */
    	switch(ifa->ifa_addr->sa_family) 
	{
    	case AF_INET:
      	    /* if chdlc_init fails it means we had an invalid IP address */
      	    if (chdlc_init(ifp, ifa)) 
	  	error = EADDRNOTAVAIL;
       	    break;
    	default:
      	    error = EAFNOSUPPORT;
      	    break;
    	}
    	break;

    case SIOCGIFADDR:
    	dlog((LOG_INFO, "%s - ioctl(SIOCGIFADDR) called.\n", card->devname));
    	break;

    case SIOCSIFFLAGS:
    	dlog((LOG_INFO, "%s - ioctl(SIOCSIFFLAGS) called.\n", card->devname));
    	/*
     	 * If the interface is marked up - enable communications. If down - 
     	 * disable communications.  IFF_UP is taken care of before entering 
     	 * this function.
     	 */
    	if ((ifp->if_flags & IFF_UP) == 0) 
	{
      	    /* bring it down */
      	    log(LOG_INFO, "%s - bringing interface down.\n", ifp->if_name);
      	    ifp->if_flags &= ~IFF_RUNNING;
	    chdlc_comm_disable(card);
     	} 
	else 
	{ /* bring it up */
	    log(LOG_INFO, "%s - bringing interface up.\n", ifp->if_name);
	    /*
	     * If flag intr_tst is not set, run interrupt test. This test will
	     * run each time when the specific wanpipe will start. 
 	     */
	    splx(s);
	    if (!card->configured && !card->wandev.piggyback)
		if (chdlc_intr_test(card, ifp))
		    return -EIO;
	    s = splimp();

	    /* 
	     * ch_enable_comms() does not mark the interface IFF_RUNNING - this
	     * is done when the protocol-open interrupt is received. If the 
             * IFF_RUNNING is already set - this means that the communications
	     * are aldready enabled, actually that the interface is already up.
	     * In this case we dont wanna enable communications again.
	     */
	    if ((ifp->if_flags & IFF_RUNNING) == 0) 
	    {
		struct timeval tv;
		chdlc_comm_enable(card);
		/* update the up time ... */
		microtime(&tv);
		chdlc_priv_area->router_up_time = tv.tv_sec;
	    }
	}
        break;

    case SIOCGIFFLAGS:
    	dlog((LOG_INFO, "%s - ioctl(SIOCGIFFLAGS) called.\n", card->devname));
    	break;

    case SIOCSIFMTU:
    	dlog((LOG_INFO, "%s - ioctl(SIOCSIFMTU) called.\n", card->devname));
    	error = EINVAL;
    	break;

    default:
    	dlog((LOG_INFO, "%s - ioctl(default) called.\n", card->devname));
    	error = EINVAL;
    	break;

    }
    splx(s);
    /* done */
    return error;
}

/*
 *
 * ===========================================================================
 * chdlc_watchdog()
 *
 */
static void chdlc_watchdog(struct ifnet *ifp)
{
    chdlc_private_area_t *chdlc_priv_area = ifp->if_softc;
    sdla_t *card = chdlc_priv_area->card;

    dlog((LOG_INFO, "%s - Watchdog function called\n", ifp->if_name));
    if ((card->wandev.state != WAN_UNCONFIGURED) && card->poll)
	card->poll(card);
}

/* 
 * ============================================================================
 * Open network interface.
 * o enable communications and interrupts.
 * o prevent module from unloading by incrementing use count
 *
 * Return 0 if O.k. or errno.
 */
static int chdlc_init (struct ifnet *ifp, struct ifaddr *ifa)
{
    chdlc_private_area_t* chdlc_priv_area = ifp->if_softc;
    sdla_t *card = chdlc_priv_area->card;
    SHARED_MEMORY_INFO_STRUCT* flags = card->u.c.flags;
    struct timeval tv;
    int err = 0;
    int 		      	 s = 0;	       	/* saved int. state */

    dlog((LOG_INFO, "%s: Initialize interface...\n", ifp->if_name));
    s = splimp();
    if (set_chdlc_config(card, ifa)) 
    {
	splx(s); 
	return -EIO;
    }

    /* Initialize Rx/Tx buffer control fields */
    init_chdlc_tx_rx_buff(card, ifp);

    /* Set interrupt mode and mask */
    if (chdlc_set_intr_mode(card, 
                        APP_INT_ON_RX_FRAME | APP_INT_ON_GLOBAL_EXCEP_COND |
                        APP_INT_ON_TX_FRAME | APP_INT_ON_CHDLC_EXCEP_COND | 
	                APP_INT_ON_TIMER)) 
    {
        splx(s);
	return -EIO;
    }

    /* Mask the Transmit and Timer interrupt */
    flags->interrupt_info_struct.interrupt_permission &= 
			~(APP_INT_ON_TX_FRAME | APP_INT_ON_TIMER);
    /* Enable communications */
    if (chdlc_comm_enable(card)) 
    {
	splx(s);
	return -EIO;
    }

    port_set_state(card, WAN_CONNECTING);
    microtime(&tv);
    chdlc_priv_area->router_start_time = tv.tv_sec;

    wanpipe_open(card);
    splx(s);
    dlog((LOG_INFO, "%s: Initialization is passed\n", ifp->if_name));
    return err;
}

/* 
 * ============================================================================
 * Send a packet on a network interface.
 * chdlc_output() - The function tries to enqueue the packet and then if the 
 *		 device is not IFF_OATIVE calls chdlc_start(). The function 
 *		 returns an error value or 0 on success.
 *
 */
static int
chdlc_output(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst, struct rtentry *rt0)
{
    chdlc_private_area_t 	 *chdlc_priv_area = ifp->if_softc;
    sdla_t 			 *card = chdlc_priv_area->card;
    SHARED_MEMORY_INFO_STRUCT    *flags = card->u.c.flags;
    INTERRUPT_INFORMATION_STRUCT *chdlc_int = &flags->interrupt_info_struct;
    int 			 udp_type = 0;
    int 			 s= 0, error = 0;

    s = splimp();
    if (m == NULL) 
    {
	/* If we get here, some higher layer thinks we've missed an
	 * tx-done interrupt.
	 */
	log(LOG_ERR, "%s: interface %s got kicked\n", card->devname, ifp->if_name);
	splx(s);
	return 0;
    }

    if ((ifp->if_flags & IFF_OACTIVE) != 0) 
    {
	/* If our device stays busy for at least 5 seconds then we increment
	 * statistics. We expect that our device never stays busy more than 
	 * 5 seconds.
	 */
	if((SYSTEM_TICK - chdlc_priv_area->tick_counter) > (5 * hz))
	{ 
            ++ card->wandev.stats.collisions;
            log(LOG_INFO, "%s: Transmit times out \n", ifp->if_name);
	}
    }

    if (card->u.c.usedby != API && dst->sa_family != AF_INET)
    {
	log(LOG_ERR, "%s: Protocol family is not isupported ...\n", 
								ifp->if_name);
	splx(s);
	return EAFNOSUPPORT;
    }

    /* check the udp packet type */
    udp_type = udp_pkt_type(mtod(m, caddr_t), card);
    if (udp_type == UDP_CPIPE_TYPE) 
    {
	if (store_udp_mgmt_pkt(UDP_PKT_FRM_STACK, card, m, ifp,
    				                         chdlc_priv_area))
	    chdlc_int->interrupt_permission |= APP_INT_ON_TIMER;
	splx(s);
	return 0;
    }

    /* check to see if the source IP address is a broadcast or 
     * multicast IP address 
     */
    if (chk_bcast_mcast_addr(card, ifp, m))
    {
	splx(s);
        return 0;
    }

    if (card->u.c.state != WAN_CONNECTED)
    {
        card->wandev.stats.tx_dropped++;
	log(LOG_ERR, "%s: Device is not connected\n", ifp->if_name);
        error = ENETDOWN;	
    }
    else
    { 
	if (IF_QFULL(&ifp->if_snd))
	{ 
	    log(LOG_INFO, "%s: Send queue is full, drop this packet\n", 
								ifp->if_name);
	    /* send queue is full - drop the packet and exit */
	    IF_DROP(&ifp->if_snd);
            splx(s);
	    return ENOBUFS;
	}
    	IF_ENQUEUE(&ifp->if_snd, m);

    	if ((ifp->if_flags & IFF_OACTIVE) == 0) 
	    chdlc_start(ifp);
    }

    splx(s);
    return error;
}


/* 
 * ============================================================================
 * chdlc_start() -  called by the system when a packet is added to the 
 *		    output queue for the interface and the device flags
 *		    indicate the device is not currently busy ( IFF_OACTIVE
 *                  is clear ).
 *                  We will try to transmit the packet. In case there's no
 *  		    available buffer - we will set IFF_OACTIVE, enable 
 *		    transmit interrupt and exit. And then the tx interrupt 
 *		    handler will call chdlc_start() and once its done - it 
 *		    will reset the IFF_OACTIVE.
 */
static void
chdlc_start(struct ifnet * ifp)
{
    chdlc_private_area_t           *chdlc_priv_area = ifp->if_softc;
    sdla_t		           *card = chdlc_priv_area->card;
    CHDLC_DATA_TX_STATUS_EL_STRUCT *txbuf = NULL;
    SHARED_MEMORY_INFO_STRUCT      *flags = card->u.c.flags;
    INTERRUPT_INFORMATION_STRUCT   *chdlc_int = &flags->interrupt_info_struct;
    int 		      	   s = 0;	        /* saved int. state */
    struct mbuf		           *opkt = NULL;	/* the packet */
    register struct mbuf	   *m = NULL;		/* current mbuf */
    unsigned long 	           data_addr;	/* where to copy data*/
    unsigned short	           data_length;	/* data length */
    char 			   *buffer = NULL, 
				   *data = NULL;

    s = splimp();

    /* We're trying to drain the queue as much as possible here because 
     * if we manage to get more than one packet into transmit queue     
     * ( line-down/line-up scenario, or heavy load ) and take only one  
     * at a time out we get an effective delay ...		      
     * BF-0001 
     */ 

    while ((opkt = ifp->if_snd.ifq_head) != NULL) 
    {	   /* BF-0001 */
	/* calculate tx element address */
        txbuf = card->u.c.txbuf;
	
	/* check if the buffer is available */
	if (txbuf->opp_flag == 0x01) 
	{
	    /* mark an interface active */
	    ifp->if_flags |= IFF_OACTIVE; 

	    chdlc_priv_area->tick_counter = SYSTEM_TICK;
	    ++ card->wandev.stats.tx_errors;

	    /* enable transmit interrupts */
	    chdlc_int->interrupt_permission |= APP_INT_ON_TX_FRAME;

	    /* restore priority level and exit */
    	    splx(s);
	    return;
        }
  
	/* 
	 * ok, now when we've got an available buffer we can transmit something
	 */
	/* Calculate packet's length */ 
	data_length = 0;
	for( m = opkt; m != NULL; m = m->m_next ) 
	    data_length += m->m_len;
	/* Create local buffer */
 	buffer = malloc(data_length+1, M_DEVBUF, M_NOWAIT);
	if (buffer == NULL)
	{
	    /* FIXME: update error statistics */
	    log(LOG_ERR, "%s: Cannot allocate data buffer!\n", ifp->if_name);
	    splx(s);
	    return;
	}
	bzero(buffer, data_length+1);
	data = buffer;
	/* Copy the packet inside */
	data_length = 0;
	for( m = opkt; m != NULL; m = m->m_next ) 
 	{
	    bcopy(mtod(m, caddr_t), &data[data_length], m->m_len);
	    data_length += m->m_len;
	}
	data[data_length] = '\0';	
	if (card->u.c.usedby == API)
	{
	    api_tx_hdr_t *api_tx_hdr = NULL;
	    unsigned char attr;
	    if (data_length <= sizeof(api_tx_hdr_t))
	    {
		/* dequeue packet from IP queue */
		IF_DEQUEUE(&ifp->if_snd, opkt);
		/* remove mbuf */
		m_freem(opkt);
		if (buffer)
		    free(buffer, M_DEVBUF);
		buffer = NULL;
		card->wandev.stats.tx_dropped++;
		continue;
	    }	
	    api_tx_hdr = (api_tx_hdr_t*)data;
	    attr = api_tx_hdr->attr;
	    data += sizeof(api_tx_hdr_t);
	    data_length -= sizeof(api_tx_hdr_t);    
	}

	/* initialize some of our stuff */
	data_addr 	= txbuf->ptr_data_bfr;

	/* Copy data to the board */
	sdla_poke(&card->hw, data_addr, data, data_length);

	/* write data length to the transmit element - offset 1, length 2 */
	txbuf->frame_length = data_length;

	/* start transmission */
	txbuf->opp_flag = 1;

	/* dequeue packet from IP queue */
	IF_DEQUEUE(&ifp->if_snd, opkt);

	/* remove local buffer */
	if (buffer)
	    free(buffer, M_DEVBUF);
	buffer = NULL;

	/* report the packet to BPF if present and attached */
#if NBPFILTER > 0
	if (ifp->if_bpf != NULL) 
	{	/* BF-0002 */
	    struct mbuf m0;
	    u_int32_t af = AF_INET;

	    m0.m_next = opkt;
	    m0.m_len = 4;
	    m0.m_data = (char *)&af;

            bpf_mtap(ifp, &m0);
	}
#endif

	/* update statistics and free the packet */
	ifp->if_opackets++;
	m_freem(opkt);
 
	/* Update transmit buffer control fields */
	card->u.c.txbuf = ++txbuf;

	if ((void*)txbuf > card->u.c.txbuf_last)
	    card->u.c.txbuf = card->u.c.txbuf_base;
 
    } /* while there're packets in the queue */

    /* restore interrupt level and exit */
    splx(s);
    return;
}

/*
 * ============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int chdlc_close (struct ifnet* ifp)
{
    chdlc_private_area_t* chdlc_priv_area = ifp->if_softc;
    sdla_t* card = chdlc_priv_area->card;
    SHARED_MEMORY_INFO_STRUCT *flags = NULL;
    int 		      	 s = 0;		       	/* saved int. state */

    dlog((LOG_INFO, "%s: Close interface %s\n", card->devname, ifp->if_name));
    s = splimp();

    flags = card->u.c.flags;
    /* Set interrupt permission byte to 0 */
    flags->interrupt_info_struct.interrupt_permission = 0;
    wanpipe_close(card);
    port_set_state(card, WAN_DISCONNECTED);
    /* Disable all interrupts */
    chdlc_set_intr_mode(card, 0);
    chdlc_comm_disable(card);

    splx(s);
    return 0;
}

/* 
 * ============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 * Return 0 if not broadcast/multicast address, otherwise return 1.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp, struct mbuf *m)
{
    uint32_t src_ip_addr = 0;
    uint32_t broadcast_ip_addr = 0;
    struct ifaddr *ifa = TAILQ_FIRST(&ifp->if_addrhead);
    struct sockaddr_in *saddr = NULL;

    /* read the IP source address from the outgoing packet */
    src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 12);

    /* read the IP broadcast address for the device */
    if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL) 
        broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);

    /* check if the IP Source Address is a Broadcast address */
    if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)) 
    {
        log(LOG_INFO, "%s: Broadcast Source Address silently discarded\n",
							card->devname);
        m_freem(m);
        ++ card->wandev.stats.tx_dropped;
        return 1;
    } 

    /* check if the IP Source Address is a Multicast address */
    if ((ntohl(src_ip_addr) >= 0xE0000001) && 
	(ntohl(src_ip_addr) <= 0xFFFFFFFE))
    {
        log(LOG_INFO, "%s: Multicast Source Address silently discarded\n", 
					card->devname);
        m_freem(m);
	++ card->wandev.stats.multicast;
        ++ card->wandev.stats.tx_dropped;
        return 1;
    }

    return 0;
}

/* 
 * ============================================================================
 * Reply to UDP Management system.
 * Return length of reply.
 */
static int reply_udp( unsigned char *data, unsigned int mbox_len )
{
    unsigned short len, udp_length, temp, ip_length;
    unsigned long ip_temp;
    int even_bound = 0;
    chdlc_udp_pkt_t *c_udp_pkt = (chdlc_udp_pkt_t*)data;
 
    /* Set length of packet */
    len = sizeof(ip_pkt_t) + sizeof(udp_pkt_t) + sizeof(wp_mgmt_t) +
	  sizeof(cblock_t) + sizeof(trace_info_t) + mbox_len;

    /* fill in UDP reply */
    c_udp_pkt->wp_mgmt.request_reply = UDPMGMT_REPLY;

    /* fill in UDP length */
    udp_length = sizeof(udp_pkt_t) + sizeof(wp_mgmt_t) + sizeof(cblock_t) +
	         sizeof(trace_info_t) + mbox_len;

    /* put it on an even boundary */
    if ( udp_length & 0x0001 ) 
    {
	udp_length += 1;
	len += 1;
	even_bound = 1;
    } 

    temp = (udp_length<<8)|(udp_length>>8);
    c_udp_pkt->udp_pkt.udp_length = temp;

    /* swap UDP ports */
    temp = c_udp_pkt->udp_pkt.udp_src_port;
    c_udp_pkt->udp_pkt.udp_src_port = c_udp_pkt->udp_pkt.udp_dst_port;
    c_udp_pkt->udp_pkt.udp_dst_port = temp;

    /* add UDP pseudo header */
    temp = 0x1100;
    *((unsigned short*)(c_udp_pkt->data + mbox_len + even_bound)) = temp;
    temp = (udp_length<<8)|(udp_length>>8);
    *((unsigned short*)(c_udp_pkt->data + mbox_len + even_bound + 2)) = temp;

    /* calculate UDP checksum */
    c_udp_pkt->udp_pkt.udp_checksum = 0;
    c_udp_pkt->udp_pkt.udp_checksum = 
		calc_checksum(&data[UDP_OFFSET], udp_length + UDP_OFFSET);

    /* fill in IP length */
    ip_length = len;
    temp = (ip_length<<8)|(ip_length>>8);
    c_udp_pkt->ip_pkt.total_length = temp;

    /* swap IP addresses */
    ip_temp = c_udp_pkt->ip_pkt.ip_src_address;
    c_udp_pkt->ip_pkt.ip_src_address = c_udp_pkt->ip_pkt.ip_dst_address;
    c_udp_pkt->ip_pkt.ip_dst_address = ip_temp;
    
    /* fill in IP checksum */
    c_udp_pkt->ip_pkt.hdr_checksum = 0;
    c_udp_pkt->ip_pkt.hdr_checksum = calc_checksum(data, sizeof(ip_pkt_t)); 

    return len;
} /* reply_udp */

unsigned short calc_checksum(char *data, int len)
{
    unsigned short temp;
    unsigned long sum = 0;
    int i;
  
    for(i=0; i < len; i += 2)
    {
	bcopy(&data[i], &temp, 2);
	sum += (unsigned long)temp;
    }

    while(sum >> 16)
	sum = (sum & 0xffffUL) + (sum >> 16);

    temp = (unsigned short)sum;
    temp = ~temp;
 
    if (temp == 0)
	temp = 0xFFFF;

    return temp;
} 

/****** Cisco HDLC Firmware Interface Functions *******************************/

/* 
 * ============================================================================
 * Read firmware code version.
 *	Put code version as ASCII string in str. 
 */
static int chdlc_read_version (sdla_t* card, char* str)
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    int len;
    char err;

    dlog((LOG_INFO, "%s: Read CHDLC Version..\n", card->devname)); 
    mb->buffer_length = 0;
    mb->command = READ_CHDLC_CODE_VERSION;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;

#ifdef WANDEBUG_CHDLC
    chdlc_dump_mailbox(mb);
#endif
    if (err != COMMAND_OK) 
	chdlc_error(card,err,mb);
    else
        if (str) 
        {  /* is not null */
	    len = mb->buffer_length;
	    bcopy(mb->data, str, len);
	    str[len] = '\0';
        }
    return (err);
}

/* 
 * ============================================================================
 *  Configure CHDLC firmware.
 */
static int chdlc_configure (sdla_t* card, void* data)
{
    int err;
    CHDLC_MAILBOX_STRUCT *mailbox = card->mbox;
    int data_length = sizeof(CHDLC_CONFIGURATION_STRUCT);

    dlog((LOG_INFO, "%s: Setting CHDLC configuration...\n", card->devname));	
    mailbox->buffer_length = data_length; 
    bcopy(data, mailbox->data, data_length);
    mailbox->command = SET_CHDLC_CONFIGURATION;
    err = sdla_exec(mailbox) ? mailbox->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
    {
        chdlc_error (card, err, mailbox);
	log(LOG_ERR, "%s: Invalid value in offset %d\n", 
				card->devname, mailbox->buffer_length);  
    }
    return err;
}

/* 
 * ============================================================================
 * Set interrupt mode -- HDLC Version.
 */

static int chdlc_set_intr_mode (sdla_t* card, unsigned mode)
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    CHDLC_INT_TRIGGERS_STRUCT* int_data =
	 (CHDLC_INT_TRIGGERS_STRUCT *)mb->data;
    int err;

    dlog((LOG_INFO, "%s: Setting interrupt mode(0x%x)...\n", card->devname, mode));
    int_data->CHDLC_interrupt_triggers 	= mode;
    int_data->IRQ			= card->hw.irq;
    int_data->interrupt_timer           = 1;
   
    mb->buffer_length = sizeof(CHDLC_INT_TRIGGERS_STRUCT);
    mb->command = SET_CHDLC_INTERRUPT_TRIGGERS;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
	chdlc_error (card, err, mb);
    return err;
}

/*
 * ============================================================================
 * Get interrupt mode -- HDLC Version
 */
static int chdlc_read_intr_mode(sdla_t* card)
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    int err = 0;

     
    mb->command = READ_CHDLC_INTERRUPT_TRIGGERS;
    mb->buffer_length = 0x0;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
	chdlc_error (card, err, mb);
    else
    {
    	CHDLC_INT_TRIGGERS_STRUCT* int_data =
	 				(CHDLC_INT_TRIGGERS_STRUCT *)mb->data;
 	return int_data->CHDLC_interrupt_triggers;	
    }
     
    return err;
}


/* 
 * ============================================================================
 * Enable communications.
 */
static int chdlc_comm_enable (sdla_t* card)
{
    int err;
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;

    mb->buffer_length = 0;
    mb->command = ENABLE_CHDLC_COMMUNICATIONS;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
    {
        if (err == CHDLC_COMMS_ENABLED)
	    dlog((LOG_INFO, "%s: Communication already enabled\n", card->devname));
        else
            chdlc_error(card, err, mb);
    }
    else
        dlog((LOG_INFO, "%s: Communication enabled\n", card->devname));	
    return err;
}

/* 
 * ============================================================================
 * Disable communications.
 */
static int chdlc_comm_disable (sdla_t* card)
{
    int err;
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;

    mb->buffer_length = 0;
    mb->command = DISABLE_CHDLC_COMMUNICATIONS;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
    {
        if (err == CHDLC_COMMS_DISABLED)
            dlog((LOG_INFO, "%s: Communication already disabled\n", card->devname)); 
        else
            chdlc_error(card,err,mb);
    }
    else
        dlog((LOG_INFO, "%s: Communication disabled\n", card->devname));
    return err;
}

/* 
 * ============================================================================
 * Read communication error statistics.
 */
static int chdlc_read_comm_err_stats (sdla_t* card)
{
    int err;
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;

    mb->buffer_length = 0;
    mb->command = READ_COMMS_ERROR_STATS;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
	chdlc_error(card,err,mb);
    else
        dlog((LOG_INFO, "%s: Read communication error stat...ok\n", card->devname));
    return err;
}

/*
 ************************************************************************
 * Read CHDLC configureation.
 */
static int chdlc_read_configure(sdla_t *card)
{
    int err;
    CHDLC_MAILBOX_STRUCT *mb = card->mbox;

    mb->buffer_length = 0;
    mb->command = READ_CHDLC_CONFIGURATION;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;

    if(err != COMMAND_OK)
        chdlc_error(card, err, mb);
    else
	dlog((LOG_INFO, "%s: Read CHDLC configuration...ok\n", card->devname));
    return err;
}


/* 
 * ============================================================================
 * Read CHDLC operational statistics.
 */
static int chdlc_read_op_stats (sdla_t* card)
{
    int err;
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;

    mb->buffer_length = 0;
    mb->command = READ_CHDLC_OPERATIONAL_STATS;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != COMMAND_OK)
	chdlc_error(card,err,mb);
    else
	dlog((LOG_INFO, "%s: Read CHDLC operational stat...ok\n", card->devname));
    return err;
}

/* 
 * ============================================================================
 * Update communications error and general packet statistics.
 */
static int 
update_comms_stats(sdla_t* card, chdlc_private_area_t* chdlc_priv_area)
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    COMMS_ERROR_STATS_STRUCT *err_stats = NULL;
    CHDLC_OPERATIONAL_STATS_STRUCT *op_stats = NULL;

    dlog((LOG_INFO, "Updating communication statistics...\n"));
    /* On the first timer interrupt, read the comms error statistics */
    if (chdlc_priv_area->update_comms_stats == 2)
    {
 	if (chdlc_read_comm_err_stats(card))
	    return 1;
   	err_stats = (COMMS_ERROR_STATS_STRUCT*)mb->data;
        card->wandev.stats.rx_over_errors = err_stats->Rx_overrun_err_count;
        card->wandev.stats.rx_crc_errors = err_stats->CRC_err_count;
        card->wandev.stats.rx_frame_errors = err_stats->Rx_abort_count;
        card->wandev.stats.rx_fifo_errors = 
				err_stats->Rx_dis_pri_bfrs_full_count;
        card->wandev.stats.rx_missed_errors = card->wandev.stats.rx_fifo_errors;
        card->wandev.stats.tx_aborted_errors = err_stats->sec_Tx_abort_count;
    }
    else
    { /* on second timer interrupt, read the operational statistics */
	if (chdlc_read_op_stats(card))
	    return 1;
	op_stats = (CHDLC_OPERATIONAL_STATS_STRUCT*)mb->data;
	card->wandev.stats.rx_length_errors = 
			(op_stats->Rx_Data_discard_short_count +
			op_stats->Rx_Data_discard_long_count);
    }
    return 0;
}

/* 
 * ============================================================================
 * Send packet.
 *	Return:	0 - o.k.
 *		1 - no transmit buffers available
 */
static int 
chdlc_send (sdla_t* card, void* data, unsigned len)
{
    CHDLC_DATA_TX_STATUS_EL_STRUCT *txbuf = card->u.c.txbuf;

    if (txbuf->opp_flag) return 1;

    sdla_poke(&card->hw, txbuf->ptr_data_bfr, data, len);

    txbuf->frame_length = len;
    txbuf->opp_flag = 1;		/* start transmission */

    /* Update transmit buffer control fields */
    card->u.c.txbuf = ++txbuf;

    if ((void*)txbuf > card->u.c.txbuf_last)
	card->u.c.txbuf = card->u.c.txbuf_base;

    return 0;
}

/****** Firmware Error Handler **********************************************/

/* 
 * ============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int chdlc_error (sdla_t *card, int err, CHDLC_MAILBOX_STRUCT *mb)
{
    unsigned cmd = mb->command;

    switch (err) 
    {
    case CMD_TIMEOUT:
	log(LOG_ERR, "%s: command 0x%02x timed out!\n", card->devname, cmd);
	break;
    case S514_BOTH_PORTS_SAME_CLK_MODE:
	if (cmd == SET_CHDLC_CONFIGURATION)
	    log(LOG_ERR, "%s: Configure both ports for the same clock source\n",
						card->devname);
	else
	    log(LOG_ERR, "%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, cmd, err);
	break;
    default:
	log(LOG_ERR, "%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, cmd, err);
    }
    return 0;
}

/****** Interrupt Handlers **************************************************/

/* 
 * ============================================================================
 * Cisco HDLC interrupt service routine.
 */
static void wpc_isr (sdla_t* card)
{
    struct ifnet 	                  *ifp = NULL;
    chdlc_private_area_t      *chdlc_priv_area = NULL;
    SHARED_MEMORY_INFO_STRUCT 		*flags = NULL;
    int i, interrupt_serviced = 0;
    sdla_t *my_card = NULL;

    /* Check for which port the interrupt has been generated 
     * Since Secondary Port is piggiback on the Primary the check be done here
     */
    flags = card->u.c.flags;
    if (!flags->interrupt_info_struct.interrupt_type)
    {
	/* Check for second port (piggiback) */
	if ((my_card = card->next) != NULL )
	{
	    flags = my_card->u.c.flags;
	    if (flags->interrupt_info_struct.interrupt_type)
		card = my_card;
	}
    }

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
							card->devname, __LINE__);
        flags->interrupt_info_struct.interrupt_type = 0;
	return;
    }
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;
    card->in_isr = 1;

    /* On a 508 Card, if critical due to if_send Major Error */
    if (card->u.c.flags != NULL)
    {
        flags = card->u.c.flags;
	
        switch(readb(&flags->interrupt_info_struct.interrupt_type)) 
        {
        case RX_APP_INT_PEND:	/* 0x01: receive interrupt */
	    interrupt_serviced = 1;
	    rx_intr(card);
            break;

        case TX_APP_INT_PEND:	/* 0x02: transmit interrupt */
	    interrupt_serviced = 1;
	    flags->interrupt_info_struct.interrupt_permission &=
					~APP_INT_ON_TX_FRAME;
	    /* interface is not active anymore */
	    ifp->if_flags &= ~IFF_OACTIVE; 
	    /* Update tick_counter */
	    chdlc_priv_area->tick_counter = SYSTEM_TICK;
	       
	    /* check the interface output queue */
            if (ifp->if_snd.ifq_head != NULL) 
	    {
	        /* there's something to send - send it */
	        chdlc_start(ifp);
	    }
            break;
              
        case COMMAND_COMPLETE_APP_INT_PEND:/* 0x04: cmd cplt */
	    interrupt_serviced = 1;
	    ++ Intr_test_counter;
            break;

        case TIMER_APP_INT_PEND:
	    interrupt_serviced = 1;
	    timer_intr(card);
            break;

        case GLOBAL_EXCEP_COND_APP_INT_PEND:
	    interrupt_serviced = 1;
	    process_global_exception(card);
            break;

        case CHDLC_EXCEP_COND_APP_INT_PEND:	/* 0x20 */
	    interrupt_serviced = 1;
	    process_chdlc_exception(card);
            break;

        default:
	    break;
	}
    }

    if (!interrupt_serviced) 
    {
        log(LOG_INFO, "%s: spurious interrupt 0x%02x!\n", card->devname,
	                flags->interrupt_info_struct.interrupt_type);
        log(LOG_INFO, "Code name: ");
        for(i = 0; i < 4; i ++)
	    log(LOG_INFO, "%c", flags->global_info_struct.codename[i]); 
        log(LOG_INFO, "\nCode version: ");
        for(i = 0; i < 4; i ++)
	    log(LOG_INFO, "%c", flags->global_info_struct.codeversion[i]); 
        log(LOG_INFO, "\n");
    }

    card->in_isr = 0;
    flags->interrupt_info_struct.interrupt_type = 0;
}

/* 
 * ============================================================================
 * Receive interrupt handler.
 */
static void rx_intr (sdla_t* card)
{
    struct ifnet 	           *ifp = NULL;
    chdlc_private_area_t 	   *chdlc_priv_area = NULL;
    SHARED_MEMORY_INFO_STRUCT 	   *flags = card->u.c.flags;
    CHDLC_DATA_RX_STATUS_EL_STRUCT *rxbuf = card->u.c.rxmb;
    struct mbuf			   *m = NULL, 
				   *ipkt = NULL;
    unsigned long 		   addr = 0;
    char 			   * data = NULL, *data_ptr = NULL;
    unsigned 			   len = 0, api_hdr_len = 0;
    int 			   i, udp_type;
    int				   s = 0;
	
    if (rxbuf->opp_flag != 0x01) 
    {
	log(LOG_INFO, "%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
			card->devname, (unsigned)rxbuf, rxbuf->opp_flag);
        log(LOG_INFO, "Code name: ");
	for(i = 0; i < 4; i ++)
	    log(LOG_INFO, "%c", flags->global_info_struct.codename[i]);
        log(LOG_INFO, "\nCode version: ");
        for(i = 0; i < 4; i ++)
	    log(LOG_INFO, "%c", flags->global_info_struct.codeversion[i]);
        log(LOG_INFO, "\n");
	return;
    }
   
    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return;
    }

    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;
    addr = rxbuf->ptr_data_bfr;
    len  = rxbuf->frame_length;
    api_hdr_len = 0;

    if (card->u.c.usedby == API)
	api_hdr_len = sizeof(api_rx_hdr_t);

    /*
     * If current card not running API, driver allocate buffer only for the 
     * received data, otherwise the length will be include the api header.
     */
    data = malloc(len + api_hdr_len + 1, M_DEVBUF, M_NOWAIT);
    if (data == NULL)
    {
	/* FIXME: update error statistics */
	log(LOG_ERR, "%s: Cannot allocate data buffer!\n",
					ifp->if_name);
	return;
    }
    bzero(data, len + api_hdr_len + 1);
    
    if (addr + len <= card->u.c.rx_top + 1)
    {
        sdla_peek(&card->hw, addr, &data[api_hdr_len], len);
	addr += len;
    }
    else
    {
	int first_piece_len = card->u.c.rx_top + 1 - addr;
	int second_piece_len = len - first_piece_len;

	sdla_peek(&card->hw, addr, &data[api_hdr_len], first_piece_len);
	sdla_peek(&card->hw, card->u.c.rx_base, 
                  	&data[api_hdr_len]+first_piece_len, second_piece_len);
	addr = card->u.c.rx_base + second_piece_len; 
    }
    if (card->u.c.usedby == API)
    { /* Update API header */
	((api_rx_hdr_t*)data)->error_flag = rxbuf->error_flag;
	((api_rx_hdr_t*)data)->time_stamp = rxbuf->time_stamp;
    }
    /* Update packet length (For API, length will include header length) */
    udp_type = udp_pkt_type(&data[api_hdr_len], card);
    if (udp_type == UDP_CPIPE_TYPE) 
	data_ptr = data + api_hdr_len;
    else
    {
	len += api_hdr_len;
	data_ptr = data;
    }

    /* get the first ( header ) mbuf */
    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m != NULL) 
    {
        /* m will be running through the chain. 
         * ipkt always point to the header 
         */
        ipkt		     = m; 
        ipkt->m_pkthdr.rcvif = ifp;	/* our interface */
        ipkt->m_pkthdr.len   = len;/* length of the data */
        ipkt->m_len          = MHLEN;	    /* at leasts as of now */

        /* let's start filling in the chain */
        while( len > 0 ) 
        {
	    /* check if we can use cluster instead ... */
	    if ( len > MINCLSIZE ) 
	    {
	        /* all right - get the cluster */
	        MCLGET(m, M_DONTWAIT);	
	        if (m->m_flags & M_EXT) 
	        {
		    /* all right - it worked - set the new length */
		    m->m_len = MCLBYTES;
	        } 
	        else 
	        {
	            /* it failed - well, screw it - we'll try again later */
	            m_freem(ipkt);
	            log(LOG_ERR, "%s: Failed to allocate cluster(RX INTR) !!!\n",
								ifp->if_name);
	            ifp->if_iqdrops++;	
	            return;
	        }
	    }

	    /* 
	     * ok, at this point we've got mbuf pointed by 'm' which is 
	     * either a regular one, or a cluster depending on how much 
	     * data do we have left. The beauty is that from here we dont
	     * care what kind of mbuf it is since all the macros work 
	     * correctly on both.
	     */
	    m->m_len = min(m->m_len, len);
  
	    /* 
	     * Now, we can start copying data into mbuf data area given 
	     * by mtod() macro. We copy m->m_len bytes. Note that since 
	     * the S508s receive buffers are cyclic we might have to copy
	     * the packet in two pieces.
	     */
#ifdef NOT_API
	    if (addr + m->m_len <= card->u.c.rx_top + 1)
	    {
		sdla_peek(&card->hw, addr, mtod(m, caddr_t), m->m_len);
		addr += m->m_len;
		len -= m->m_len;
  	    }
	    else
	    {
		int first_piece_len = card->u.c.rx_top + 1 - addr;
		int second_piece_len = len - first_piece_len;
		sdla_peek(&card->hw, addr, mtod(m, caddr_t), first_piece_len);
		sdla_peek(&card->hw, card->u.c.rx_base, 
                  	mtod(m, caddr_t)+first_piece_len, second_piece_len);
	 	addr = card->u.c.rx_base + second_piece_len;
		len -= m->m_len;	
	    }
#else
            bcopy(data_ptr, mtod(m, caddr_t), m->m_len );
	    data_ptr += m->m_len;
    	    len -= m->m_len;
#endif 
	    /* done with this mbuf - allocate the next one if needed */
	    if (len > 0) 
	    {
	        MGET(m->m_next, M_DONTWAIT, MT_DATA);
	        if (m->m_next == NULL) 
	        {
	            /* it failed - well, screw it - we'll try again later */
	            m_freem(ipkt);
	            log(LOG_ERR, "%s: Failed to allocate mbuf (RX INTR)!!!\n",
								ifp->if_name);
                    ifp->if_iqdrops++;	
	            return;
	        }
	        /* set m to point to the new one, and set up initial 
	         * length
	         */
	        m = m->m_next;
	        m->m_len = MLEN;
	    }
        } /* while ... */			

        if (udp_type == UDP_CPIPE_TYPE) 
        {
            if (store_udp_mgmt_pkt(UDP_PKT_FRM_NETWORK, card,m, ifp, 
			  chdlc_priv_area))
                flags->interrupt_info_struct.interrupt_permission |=
				APP_INT_ON_TIMER; 
        } 
        else 
        {
            /* Ok, this is a regular packet - do what's said above */
#if NBPFILTER > 0
            if (ifp->if_bpf != NULL)  
            {			/* BF-0002 */
	        struct mbuf m0;
	        u_int32_t af = AF_INET;

                m0.m_next = ipkt;
	        m0.m_len = 4;
	        m0.m_data = (char *)&af; 

	        bpf_mtap(ifp, &m0);
	        /*bpf_mtap(ifp, ipkt);*/
            }
#endif
            /* 
             * Schedule IP input routine. Note that it makes much more 
             * sence to check first if the protocol queue is full - and
             * then to schedule the input routine, but i took a look at
             * kernel sources for ether_input() 
             * [ /usr/src/sys/net/if_ethersubr.c ] and they do it in this
             * order - so i figured they know better.
             */
            s = splimp();
            schednetisr(NETISR_IP);
            if (IF_QFULL(&ipintrq)) 
            {
                log(LOG_ERR, "%s: IP queue is full, drop packet (RX INTR)!!!\n", 
							ifp->if_name);
                /* oh, no - IP queue is full - well - we'll try again 
                 * later 
                 */ 
                IF_DROP(&ipintrq);
                m_freem(ipkt);
                splx(s);
                return;
            }
 
            /* ok enqueue the packet */
            IF_ENQUEUE(&ipintrq, ipkt);
            splx(s);
  
            /* update interface statistics */
            ifp->if_ipackets++;
            card->wandev.stats.rx_packets++;
        }
    } 
    else 
    {
        log(LOG_ERR, "%s: no socket buffers available!\n", card->devname);
        ++ card->wandev.stats.rx_dropped;
        ifp->if_iqdrops++;	
    }
    if (data)
	free(data, M_DEVBUF);

    /* Release buffer element and calculate a pointer to the next one */
    rxbuf->opp_flag = 0x00;
    card->u.c.rxmb = ++ rxbuf;
    if ((void*)rxbuf > card->u.c.rxbuf_last)
	card->u.c.rxmb = card->u.c.rxbuf_base;
}

/* 
 * ============================================================================
 * Miscellaneous Functions
 *	- set_chdlc_config() used to set configuration options on the board
 */
static int set_chdlc_config(sdla_t *card, struct ifaddr *ifa)
{
    struct ifnet *ifp = NULL;
    chdlc_private_area_t *chdlc_priv_area = NULL; 
    CHDLC_CONFIGURATION_STRUCT cfg;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return ENODEV;
    }
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;	
    bzero(&cfg, sizeof(CHDLC_CONFIGURATION_STRUCT));

    if (card->wandev.clocking)
	cfg.baud_rate = card->wandev.bps;

    cfg.line_config_options = 
			(card->wandev.interface == WANOPT_RS232) ?
			INTERFACE_LEVEL_RS232 : INTERFACE_LEVEL_V35;

    cfg.modem_config_options	    = 0;     
    cfg.modem_status_timer	    = 100;   
    cfg.CHDLC_protocol_options	    = card->u.c.protocol_options;
    cfg.percent_data_buffer_for_Tx  = 50;
    cfg.CHDLC_statistics_options    = 0;
    cfg.max_CHDLC_data_field_length = card->wandev.mtu;
    cfg.transmit_keepalive_timer    = card->u.c.kpalv_tx;
    cfg.receive_keepalive_timer	    = card->u.c.kpalv_rx;
    cfg.keepalive_error_tolerance   = card->u.c.kpalv_err;
    cfg.SLARP_request_timer	    = card->u.c.slarp_timer;

    if (cfg.SLARP_request_timer) 
    {
	cfg.IP_address	= 0;
	cfg.IP_netmask	= 0;
    }
    else 
    {
        struct sockaddr_in *saddr = NULL;
        unsigned long	local_ip = 0;	/* local IP address */
        unsigned long 	remote_ip = 0;	/* remote IP address */
        unsigned long	ip_mask = 0;	/* IP mask	    */

        /* get local IP address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_addr) != NULL) 
            local_ip = htonl(saddr->sin_addr.s_addr);

        /* get remote IP address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_dstaddr) != NULL)
	    remote_ip = htonl(saddr->sin_addr.s_addr);

        /* get  IP netmask address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_netmask) != NULL)
	    ip_mask = htonl(saddr->sin_addr.s_addr);

        cfg.IP_address 		    = local_ip;
        cfg.IP_netmask 		    = ip_mask;
        chdlc_priv_area->IP_address = local_ip;
        chdlc_priv_area->IP_netmask = ip_mask;

	/* FIXME: We must re-think this message in next release */
	/*if ((cfg->IP_address & 0x000000FF) > 2)
	 *{
	 *   log(LOG_WARN, "WARNING: %s configured win an invalid\n",
	 *						card->devname);
	 *   log(LOG_WARN, "local IP address. Slarp pragmatic will fail.\n");
	 *   log(LOG_WARN, "IP address should be of the format A.B.C.1 or \n");
	 *   log(LOG_WARN, "A.B.C.2\n");
	 *}
 	 */
    }
    return chdlc_configure(card, &cfg);
}

/* 
 * ============================================================================
 * Process global exception condition
 */
static int process_global_exception(sdla_t *card)
{
    CHDLC_MAILBOX_STRUCT* mbox = card->mbox;
    int err;

    mbox->buffer_length = 0;
    mbox->command = READ_GLOBAL_EXCEPTION_CONDITION;
    err = sdla_exec(mbox) ? mbox->return_code : CMD_TIMEOUT;

    if (err != CMD_TIMEOUT )
    {
	switch(mbox->return_code) 
	{
        case EXCEP_MODEM_STATUS_CHANGE:
	    log(LOG_INFO, "%s: Modem status change\n", card->devname);
	    switch(mbox->data[0] & (DCD_HIGH | CTS_HIGH)) 
	    {
	    case (DCD_HIGH):
		log(LOG_INFO, "%s: DCD high, CTS low\n", card->devname);
		break;
	    case (CTS_HIGH):
		log(LOG_INFO, "%s: DCD low, CTS high\n", card->devname);       
		break;
           case ((DCD_HIGH | CTS_HIGH)):
		log(LOG_INFO, "%s: DCD high, CTS high\n", card->devname);
                break;
	   default:
		log(LOG_INFO, "%s: DCD low, CTS low\n", card->devname);
                break;
	   }
	   break;

        case EXCEP_TRC_DISABLED:
	    log(LOG_INFO, "%s: Line trace disabled\n", card->devname);
            break;

	case EXCEP_IRQ_TIMEOUT:
	    log(LOG_INFO, "%s: IRQ timeout occurred\n", card->devname);
	    break;

        default:
	    log(LOG_INFO, "%s: global exception\n", card->devname);
            break;
	}
    }
    return 0;
}


/*
 * ============================================================================
 * Process chdlc exception condition
 */
static int process_chdlc_exception(sdla_t *card)
{
    struct ifnet *ifp = NULL;
    chdlc_private_area_t *chdlc_priv_area = NULL;
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    int err;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    { 	
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
        return ENODEV;
    }
    ifp = card->wandev.if_head->ifp;	
    chdlc_priv_area = ifp->if_softc;
    mb->buffer_length = 0;
    mb->command = READ_CHDLC_EXCEPTION_CONDITION;
    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
    if (err != CMD_TIMEOUT) 
    {
	switch (err) 
	{
	case EXCEP_LINK_ACTIVE:
	    log(LOG_INFO, "%s: CHDLC link became active ...\n", card->devname);
	    /* protocol is up - now if the slarp timer is zero, that 
	     * means that IP address was set at the beginning and 
	     * we are able for data transfers, so we will bring the  
             * interface up. If the slarp timer of station is not zero 
	     * - we have to wait till we discover out IP address ... 
	     */
	    if (card->u.c.slarp_timer == 0)
	    { 
                struct timeval tv;
                ifp->if_flags |= (IFF_UP|IFF_RUNNING);
	        log(LOG_INFO, "%s: Router has been brought up.\n", card->devname); 
	        microtime(&tv);
	        chdlc_priv_area->router_up_time = tv.tv_sec;
  	    } 	
	    port_set_state(card, WAN_CONNECTED);
	    break;

	case EXCEP_LINK_INACTIVE_MODEM:
	    log(LOG_INFO, "%s - link is down because of invalid modem state\n",
							card->devname);
	    port_set_state(card, WAN_DISCONNECTED);
	    unconfigure_ip(card);
	    break;

	case EXCEP_LINK_INACTIVE_KPALV:
	    log(LOG_INFO, "%s: link is down due to the station not receiving SLARP keepalive frames ...\n", card->devname); 
	    port_set_state(card, WAN_DISCONNECTED);
	    log(LOG_INFO, "%s: Keepalive timer expired.\n", card->devname); 
	    unconfigure_ip(card);
	    break;

        case EXCEP_IP_ADDRESS_DISCOVERED:
            log(LOG_INFO, "%s: IP has been discovered ...\n", card->devname);
	    if (configure_ip(card)) 
	        return -1;
	    ifp->if_flags |= (IFF_UP|IFF_RUNNING);
	    break;

        case EXCEP_LOOPBACK_CONDITION:
	    log(LOG_INFO, "%s: Loopback Condition Detected.\n", card->devname);
	    chdlc_comm_disable(card);
	    ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
	    break;

        case NO_CHDLC_EXCEP_COND_TO_REPORT:
	    log(LOG_INFO, "%s: No exceptions reported.\n", card->devname);
	    break;
        }

    }
    return 0;
}

/* 
 * ============================================================================
 * Configure IP from SLARP negotiation
 * This adds dynamic routes when SLARP has provided valid addresses
 */
static int configure_ip (sdla_t* card)
{
    struct ifnet *ifp = NULL;
    chdlc_private_area_t *chdlc_priv_area = NULL;
    int err = 0;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return ENODEV; 
    }
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;
    /* set to discover */
    if (card->u.c.slarp_timer != 0x0)
    {
	CHDLC_MAILBOX_STRUCT *mb = card->mbox;
	CHDLC_CONFIGURATION_STRUCT *cfg = NULL;

        err = chdlc_read_configure(card);
	if (err != COMMAND_OK) 
	    return -1;

	/* According to Cisco HDLC, if the point-to-point address is
	 * A.B.C.1, then we are the opposite (A.B.C.2), and vice-versa,
	 * ie 2 and 1 respectively 
	 */
	cfg = (CHDLC_CONFIGURATION_STRUCT *)mb->data;

        chdlc_priv_area->IP_address = cfg->IP_address;
        chdlc_priv_area->IP_netmask = cfg->IP_netmask;
    }

    /* Set flag to add route */ 
    chdlc_priv_area->route_status = ADD_ROUTE;

    /* The idea here is to add the route in the poll routine.
     * This way, we aren't in interrupt context when adding routes 
     */
    card->poll = process_route;		/* process_route(card); */
    ifp->if_timer = 1;
    
    return 0;
}

/* 
 * ============================================================================
 * Un-Configure IP negotiated by SLARP
 * This removes dynamic routes when the link becomes inactive.
 */
static int unconfigure_ip (sdla_t* card)
{
    struct ifnet *ifp = NULL;
    chdlc_private_area_t *chdlc_priv_area = NULL;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return ENODEV;
    }
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;

    if (chdlc_priv_area->route_status == ROUTE_ADDED) 
    {
	chdlc_priv_area->route_status = REMOVE_ROUTE;
	/* 
	 * The idea here is to delete the route in the poll routine.
         * This way, we aren't in interrupt context when adding routes 
	 */
        card->poll = process_route;		/* process_route(card); */
	ifp->if_timer = 1;
    }
    return 0;
}

/* 
 * ============================================================================
 * Routine to add/remove routes 
 * Called like a polling routine when Routes are flagged to be added/removed.
 */
static void process_route (void *arg)
{
    sdla_t       *card = (sdla_t*)arg;
    struct ifnet *ifp = NULL;
    chdlc_private_area_t *chdlc_priv_area = NULL;
    uint32_t local_IP_addr = 0, remote_IP_addr = 0, IP_netmask = 0, IP_addr = 0;
    struct ifaddr *ifa = NULL; 
    struct sockaddr_in *saddr = NULL;
    struct timeval tv;
    int err = 0;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return; 
    }
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;
    ifa = TAILQ_FIRST(&ifp->if_addrhead); 
    if ((chdlc_priv_area->route_status == ADD_ROUTE) &&
	((chdlc_priv_area->IP_address & ~chdlc_priv_area->IP_netmask) > 2)) 
    {
        /* check that the received IP address is A.B.C.1 or A.B.C.2 
	 */
        log(LOG_INFO, "%s: Dynamic route failure.\n", card->devname); 
        if(card->u.c.slarp_timer) 
        {
            log(LOG_INFO, "%s: Bad IP address %s received ", 
	         card->devname, in_ntoa(ntohl(chdlc_priv_area->IP_address)));
            log(LOG_INFO, "from remote station.\n");
        } 
        else 
        {
            log(LOG_INFO, "%s: Bad IP address %s issued",
	          card->devname, in_ntoa(ntohl(chdlc_priv_area->IP_address)));
            log(LOG_INFO, "to remote station.\n Local");
            log(LOG_INFO, "IP address must be A.B.C.1");	
            log(LOG_INFO, "or A.B.C.2.\n");
        }
	/* Remove the route due to the IP address error condition */
	chdlc_priv_area->route_status = REMOVE_ROUTE;
	err = 1;
    }

    /* If we are removing a route with bad IP address, then use the 
     * local configured IP addresses 
     */
    if ((chdlc_priv_area->route_status == REMOVE_ROUTE) && err)
    {
	struct sockaddr_in *saddr = NULL;
	/* do not remove a bad route that has already been removed */
	if (chdlc_priv_area->route_removed)
	{
	    card->poll = NULL;
	    return;
	}
        if ((saddr = (struct sockaddr_in *)ifa->ifa_addr) != NULL) 
            local_IP_addr = htonl(saddr->sin_addr.s_addr);
        if ((saddr = (struct sockaddr_in *)ifa->ifa_dstaddr) != NULL) 
            remote_IP_addr = htonl(saddr->sin_addr.s_addr);
        if ((saddr = (struct sockaddr_in *)ifa->ifa_netmask) != NULL) 
            IP_netmask = htonl(saddr->sin_addr.s_addr);
    }
    else
    {
        IP_netmask = ntohl(chdlc_priv_area->IP_netmask);
        remote_IP_addr = ntohl(chdlc_priv_area->IP_address);
        local_IP_addr = (remote_IP_addr & IP_netmask) + 
		    (~remote_IP_addr & ntohl(0x0003));

        if (!card->u.c.slarp_timer) 
        { 
	    IP_addr = local_IP_addr;
	    local_IP_addr = remote_IP_addr;
	    remote_IP_addr = IP_addr;
        }
    }
    log(LOG_INFO, "local_IP_addr: %s\n", in_ntoa(local_IP_addr));
    log(LOG_INFO, "remote_IP_addr: %s\n", in_ntoa(remote_IP_addr));
  
    /* ok, wicked - now we need to update the interface and routing 
     * tables ...							
     */	
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) == NULL ) 
    {
        card->poll = NULL;
        return;
    }

    switch (chdlc_priv_area->route_status) 
    {
    case ADD_ROUTE:

        if (card->u.c.slarp_timer) 
        {
	    /* Deleting old routing entry */
	    err = rtrequest((int)RTM_DELETE,	/* action      */
              		ifa->ifa_dstaddr,	/* destination */
              		ifa->ifa_addr,		/* gateway     */
              		ifa->ifa_netmask,	/* mask        */
              		RTF_HOST,		/* flag        */
              		(struct rtentry **)0);

	    /* set up new local and remote IP's and new netmask */
	    saddr = (struct sockaddr_in *)ifa->ifa_addr;
	    saddr->sin_addr.s_addr = local_IP_addr;

	    saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
	    saddr->sin_addr.s_addr = remote_IP_addr;

	    saddr = (struct sockaddr_in *)ifa->ifa_netmask;
	    saddr->sin_addr.s_addr = IP_netmask;

	    /* add new routing entry */
	    err = rtrequest((int)RTM_ADD, 		/* action */
	          	    ifa->ifa_dstaddr,		/* destination */
	              	    ifa->ifa_addr,		/* gateway */
	               	    ifa->ifa_netmask,		/* mask */
                       	    RTF_HOST,			/* flag */
                            (struct rtentry **)0);
	    if (err)
		log(LOG_ERR, "%s: Add route %s is failed (%d)\n",
				card->devname, in_ntoa(remote_IP_addr), err); 
            else
	    {
		chdlc_priv_area->route_status = ROUTE_ADDED;
		log(LOG_INFO, "%s: Dynamic route added\n", card->devname);
	    	log(LOG_INFO, "%s:   Local IP addr = %s\n",
				card->devname, in_ntoa(local_IP_addr));
	    	log(LOG_INFO, "%s:   Remote IP addr = %s\n",
				card->devname, in_ntoa(remote_IP_addr));
		chdlc_priv_area->route_removed = 0;	
		/* perfect! now we're ready to go - mark interface up 
	     	 * and running.
	     	 */
	    	ifp->if_flags |= (IFF_UP|IFF_RUNNING);
	    	/* update the up time ... */
	    	microtime(&tv);
	    	chdlc_priv_area->router_up_time = tv.tv_sec;
       	    }
	}
        break;

    case REMOVE_ROUTE:
	/* Deleting current routing entry */
	err = rtrequest((int)RTM_DELETE,	/* action      */
        	ifa->ifa_dstaddr,	/* destination */
              	ifa->ifa_addr,		/* gateway     */
              	ifa->ifa_netmask,	/* mask        */
               	RTF_HOST,		/* flag        */
               (struct rtentry **)0);
	
	if (err)
	    log(LOG_ERR, "%s: Remove route %s failed (%d)\n",
			card->devname, in_ntoa(remote_IP_addr), err);
	else
	{
            chdlc_priv_area->route_status = NO_ROUTE;
	    log(LOG_INFO, "%s: Dynamic route removed %s.\n", 
					card->devname, in_ntoa(local_IP_addr)); 
	    /* mark interface inactive */
	    ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
	    chdlc_priv_area->route_removed = 1;	
	    /* Set destination address to zero */ 
	    saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
	    saddr->sin_addr.s_addr = 0;
        }
	break;
    }

    /* Once we've processed the route, stop polling */
    card->poll = NULL;
}

/*
 * =============================================================================
 * Store a UDP management packet for later processing.
 */
static int 
store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card, struct mbuf *m,
                   struct ifnet *ifp, chdlc_private_area_t *chdlc_priv_area)
{
    int udp_pkt_stored = 0;
    int api_hdr_len = 0;

    /*
     * If current card running API and UDP packet arrived from NETWORK,
     * then remove API header, which was added in rx_intr function.
     * Otherwise, api_hdr_len will be 0. 
     */
    if (udp_pkt_src == UDP_PKT_FRM_NETWORK && card->u.c.usedby == API)
	api_hdr_len = sizeof(api_rx_hdr_t);

    if(!chdlc_priv_area->udp_pkt_lgth && (m->m_len <= MAX_LGTH_UDP_MGNT_PKT)) 
    {
	dlog((LOG_INFO, "%s: Storing UDP packet in private area..\n", 
						ifp->if_name));
	chdlc_priv_area->udp_pkt_lgth = m->m_len - api_hdr_len;
	chdlc_priv_area->udp_pkt_src = udp_pkt_src;
	bcopy(mtod(m, caddr_t) + api_hdr_len, 
		chdlc_priv_area->udp_pkt_data, m->m_len - api_hdr_len);
	chdlc_priv_area->timer_int_enabled = TMR_INT_ENABLED_UDP;
	udp_pkt_stored = 1;
    }
    else
 	log(LOG_ERR, "%s: Previous UDP packet is still not sended!\n", 
								ifp->if_name);	

    m_freem(m);
	
    return(udp_pkt_stored);
}

/* Finished
 * ============================================================================
 * Process UDP management packet.
 */
static int 
process_udp_mgmt_pkt(sdla_t* card, struct ifnet *ifp, 
		     chdlc_private_area_t *chdlc_priv_area)
{
    unsigned int frames;
    struct mbuf *m = NULL, *m_curr = NULL;
    unsigned short buffer_length, real_len;
    unsigned long data_ptr;
    unsigned data_length = 0;
    unsigned char *data_addr = NULL;
    int udp_mgmt_req_valid = 1;
    CHDLC_MAILBOX_STRUCT *mb = card->mbox;
    SHARED_MEMORY_INFO_STRUCT *flags = card->u.c.flags;
    chdlc_udp_pkt_t *chdlc_udp_pkt = NULL;
    struct timeval tv;
    int err;
    char ut_char;

    chdlc_udp_pkt = (chdlc_udp_pkt_t *)chdlc_priv_area->udp_pkt_data;

    switch(chdlc_udp_pkt->cblock.command) 
    {
    case FT1_MONITOR_STATUS_CTRL:
    case CPIPE_ENABLE_TRACING:
    case CPIPE_DISABLE_TRACING:
    case CPIPE_GET_TRACE_INFO:
    case SET_FT1_MODE:
	if (chdlc_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) 
	    udp_mgmt_req_valid = 0;
	break;
   
    default:
        break;
    } 

    if(!udp_mgmt_req_valid) 
    {   /* set length to 0 */
        chdlc_udp_pkt->cblock.buffer_length = 0;

        /* set return code */
        chdlc_udp_pkt->cblock.return_code = 0xCD;
    } 
    else
    { 
        unsigned long trace_status_cfg_addr = 0;
        TRACE_STATUS_EL_CFG_STRUCT trace_cfg_struct;
        TRACE_STATUS_ELEMENT_STRUCT trace_element_struct;

        switch(chdlc_udp_pkt->cblock.command) 
        {
        case CPIPE_ENABLE_TRACING:
            if (!chdlc_priv_area->TracingEnabled) 
  	    {
		/* OPERATE_DATALINE_MONITOR */
	        mb->buffer_length = sizeof(LINE_TRACE_CONFIG_STRUCT);
		mb->command = SET_TRACE_CONFIGURATION;

    		((LINE_TRACE_CONFIG_STRUCT *)mb->data)->
			trace_config = TRACE_ACTIVE;
		/* Trace delay mode is not used because it slows
		 * down transfer and results in a standoff situation
	         * when there is a lot of data 
		 */

		((LINE_TRACE_CONFIG_STRUCT *)mb->data)-> trace_config |=
				chdlc_udp_pkt->data[0];

		((LINE_TRACE_CONFIG_STRUCT *)mb->data)->
				trace_deactivation_timer = 4000;

		err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
		if (err != COMMAND_OK) 
		{
		    chdlc_error(card,err,mb);
		    card->TracingEnabled = 0;
		    chdlc_udp_pkt->cblock.return_code = err;
		    mb->buffer_length = 0;
		    break;
	        } 

		/* Get the base address of the trace element list */
		mb->buffer_length = 0;
		mb->command = READ_TRACE_CONFIGURATION;
		err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
		if (err != COMMAND_OK) 
		{
		    chdlc_error(card,err,mb);
		    chdlc_priv_area->TracingEnabled = 0;
		    chdlc_udp_pkt->cblock.return_code = err;
		    mb->buffer_length = 0;
		    break;
	    	} 	

	   	trace_status_cfg_addr =((LINE_TRACE_CONFIG_STRUCT *)
	                	mb->data) -> ptr_trace_stat_el_cfg_struct;

		sdla_peek(&card->hw, trace_status_cfg_addr,
				&trace_cfg_struct, sizeof(trace_cfg_struct));
		    
		chdlc_priv_area->start_trace_addr = 
			trace_cfg_struct.base_addr_trace_status_elements;

		chdlc_priv_area->number_trace_elements = 
			trace_cfg_struct.number_trace_status_elements;

		chdlc_priv_area->end_trace_addr = 
			(unsigned long)((TRACE_STATUS_ELEMENT_STRUCT *)
				chdlc_priv_area->start_trace_addr + 
				(chdlc_priv_area->number_trace_elements - 1));

		chdlc_priv_area->base_addr_trace_buffer = 
			trace_cfg_struct.base_addr_trace_buffer;

		chdlc_priv_area->end_addr_trace_buffer = 
			trace_cfg_struct.end_addr_trace_buffer;

	 	chdlc_priv_area->curr_trace_addr = 
			trace_cfg_struct.next_trace_element_to_use;

	    	chdlc_priv_area->available_buffer_space = 
			2000 - sizeof(ip_pkt_t) - sizeof(udp_pkt_t) -
			sizeof(wp_mgmt_t) - sizeof(cblock_t) - 
			sizeof(trace_info_t); 
	    }  
            chdlc_udp_pkt->cblock.return_code = COMMAND_OK;
	    mb->buffer_length = 0;
	    chdlc_priv_area->TracingEnabled = 1;
	    break;
	   
        case CPIPE_DISABLE_TRACING:
	    if (chdlc_priv_area->TracingEnabled) 
	    {
	        /* OPERATE_DATALINE_MONITOR */
	        mb->buffer_length = sizeof(LINE_TRACE_CONFIG_STRUCT);
	        mb->command = SET_TRACE_CONFIGURATION;
    	        ((LINE_TRACE_CONFIG_STRUCT *)mb->data)-> trace_config = 
		                                 TRACE_INACTIVE;
	        err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
	    }		

	    chdlc_priv_area->TracingEnabled = 0;
	    chdlc_udp_pkt->cblock.return_code = COMMAND_OK;
	    mb->buffer_length = 0;
	    break;
	   
        case CPIPE_GET_TRACE_INFO:
	    if (!chdlc_priv_area->TracingEnabled) 
	    {
	        chdlc_udp_pkt->cblock.return_code = 1;
	        mb->buffer_length = 0;
	        break;
	    }
	    chdlc_udp_pkt->trace_info.ismoredata = 0x0;
            buffer_length = 0;	/* offset of packet already occupied */
            for (frames=0; frames < chdlc_priv_area->number_trace_elements;
	                         frames++) 
	    {
	        trace_pkt_t *trace_pkt = (trace_pkt_t *)
	                       &chdlc_udp_pkt->data[buffer_length];
	        sdla_peek(&card->hw, chdlc_priv_area->curr_trace_addr,
                               (unsigned char *)&trace_element_struct,
	                       sizeof(TRACE_STATUS_ELEMENT_STRUCT));
	        if (trace_element_struct.opp_flag == 0x00) 
	            break;

	        /* get pointer to real data */
	        data_ptr = trace_element_struct.ptr_data_bfr;

	        /* See if there is actual data on the trace buffer */
	        if (data_ptr)
		    data_length = trace_element_struct.trace_length;
		else
		{
		    data_length = 0;
		    chdlc_udp_pkt->trace_info.ismoredata = 0x01;
		}
	
	        if( (chdlc_priv_area->available_buffer_space - buffer_length)
			      < ( sizeof(trace_pkt_t) + data_length) ) 
	        {
	            /* indicate there are more frames on board & exit */
	            chdlc_udp_pkt->trace_info.ismoredata = 0x01;
                    break;
                }
	        trace_pkt->status = trace_element_struct.trace_type;
	        trace_pkt->time_stamp = trace_element_struct.trace_time_stamp;
	        trace_pkt->real_length = trace_element_struct.trace_length;

	        /* see if we can fit the frame into the user buffer */
	        real_len = trace_pkt->real_length;

	        if (data_ptr == 0) 
	            trace_pkt->data_avail = 0x00;
	        else 
                {
	            unsigned tmp = 0;

	            /* get the data from circular buffer
                     * must check for end of buffer 
	             */
	            trace_pkt->data_avail = 0x01;

	            if ((data_ptr + real_len) >
	                      chdlc_priv_area->end_addr_trace_buffer + 1) 
	            {
                        tmp = chdlc_priv_area->end_addr_trace_buffer - 
                                              data_ptr + 1;
                        sdla_peek(&card->hw, data_ptr, trace_pkt->data, tmp);
		        data_ptr = chdlc_priv_area->base_addr_trace_buffer;
                    }
      	            sdla_peek(&card->hw, data_ptr,
		                    &trace_pkt->data[tmp], real_len - tmp);
	        }	

                /* zero the opp flag to show we got the frame */
	        ut_char = 0x00;
	        sdla_poke(&card->hw, chdlc_priv_area->curr_trace_addr,
				  &ut_char, 1);
                /* now move onto the next frame */
                chdlc_priv_area->curr_trace_addr += 
		                sizeof(TRACE_STATUS_ELEMENT_STRUCT);

	        /* check if we went over the last address */
	        if (chdlc_priv_area->curr_trace_addr > 
		                   chdlc_priv_area->end_trace_addr)
		    chdlc_priv_area->curr_trace_addr = 
		                     chdlc_priv_area->start_trace_addr;

	        /* update buffer length */ 
  	        if (trace_pkt->data_avail == 0x01) 
	            buffer_length += real_len - 1;

    	        /* for the header */
      	        buffer_length += sizeof(trace_pkt_t);

	    }  /* For Loop */
	    if (frames == chdlc_priv_area->number_trace_elements)
		chdlc_udp_pkt->trace_info.ismoredata = 0x01;
	    chdlc_udp_pkt->trace_info.num_frames = frames;

	    /* set the data length */
            mb->buffer_length = buffer_length;
	    chdlc_udp_pkt->cblock.buffer_length = buffer_length; 

            /* set return code */
	    chdlc_udp_pkt->cblock.return_code = COMMAND_OK; 
            break;

        case CPIPE_FT1_READ_STATUS:
	    ((unsigned char *)chdlc_udp_pkt->data )[0] =
		            flags->FT1_info_struct.parallel_port_A_input;

	    ((unsigned char *)chdlc_udp_pkt->data )[1] =
			    flags->FT1_info_struct.parallel_port_B_input;

	    chdlc_udp_pkt->cblock.return_code = COMMAND_OK;
	    mb->buffer_length = 2;
	    break;
	
        case CPIPE_ROUTER_UP_TIME:
	    microtime( &tv );
	    chdlc_priv_area->router_up_time = tv.tv_sec - 
	                        chdlc_priv_area->router_start_time;
	    *(unsigned long *)&chdlc_udp_pkt->data = 
	                        chdlc_priv_area->router_up_time;
	    mb->buffer_length = sizeof(unsigned long);
	    break;

        case FT1_MONITOR_STATUS_CTRL:
	    /* Enable FT1 MONITOR STATUS */
       	    if ((chdlc_udp_pkt->data[0] & ENABLE_READ_FT1_STATUS) ||  
		(chdlc_udp_pkt->data[0] & ENABLE_READ_FT1_OP_STATS))
	    {
	        if (rCount++ != 0)  
	        {
		    chdlc_udp_pkt->cblock.return_code = COMMAND_OK;
		    mb->buffer_length = 1;
		    break;
   	         }
   	     }

     	     /* Disable FT1 MONITOR STATUS */
      	     if (chdlc_udp_pkt->data[0] == 0)
	     {
      	         if (--rCount != 0) 
		 {
		     chdlc_udp_pkt->cblock.return_code = COMMAND_OK;
		     mb->buffer_length = 1;
		     break;
    	         } 
    	     } 	

        default:
	    /* it's a board command */
	    mb->command = chdlc_udp_pkt->cblock.command;
	    mb->buffer_length = chdlc_udp_pkt->cblock.buffer_length;
	    if (mb->buffer_length) 
	    {
	        bcopy((unsigned char *)chdlc_udp_pkt->data,
	                   	&mb->data, mb->buffer_length);
	    } 
	    /* run the command on the board */
            err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
	    if (err != COMMAND_OK) 
	    {
	        chdlc_error(card, err, mb);
	        break;
	    }

	    /* copy the result back to our buffer */
       	    bcopy(mb, &chdlc_udp_pkt->cblock, sizeof(cblock_t)); 

            if (mb->buffer_length) 
	        bcopy(&mb->data, &chdlc_udp_pkt->data, mb->buffer_length); 

        } /* end of switch */
    } /* end of else */

    /* Fill UDP TTL */
    chdlc_udp_pkt->ip_pkt.ttl = card->wandev.ttl; 

    data_length = reply_udp(chdlc_priv_area->udp_pkt_data, mb->buffer_length);

    if (chdlc_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) 
    {
        chdlc_send(card, chdlc_priv_area->udp_pkt_data, data_length);
    }
    else 
    {
	int s = 0;	
        /*
         * Ok, what we've got now in chdlc_priv_area->udp_pkt_data is a 
         * fully formatted IP packet which consists of responce to the 
         * UDP management request. The length of the packet is 
         * 'data_length'. We have to free the original mbuf and
         * allocate the new one. ( when i tried to re-use the same 
         * mbuf with m_pullup i ended up crashing the system somewhere
         * in IP stack after i passed the frame up - looks like they 
         * free the buffer anyways if the driver does not ).
         */  
        /* will run through the buffer */
        data_addr  = chdlc_priv_area->udp_pkt_data;	

        /*
         * Build a new mbuf chain 
         */

        /* get the first ( header ) mbuf */
        MGETHDR(m, M_DONTWAIT, MT_DATA);
        if (m == NULL) 
        {
            log(LOG_ERR, "%s: Failed to allocate mbuf (UDP)!!!\n", ifp->if_name);
            /* i guess nothing can be done ... the packet is gone */
            return ENOMEM;
        }

        /* fill in the header */
        m->m_pkthdr.rcvif = ifp;		/* our interface */
        m->m_pkthdr.len	  = data_length;	/* length of the data */
        m->m_len	  = MHLEN;	/* well, at least as of now */

        /* m_curr will be running through the chain. m points to 
         * the header 
         */
        m_curr = m; 

        /* let's start filling in the chain */
        while(data_length > 0) 
        {
            /* check if we can use cluster instead ... */
            if (data_length > MINCLSIZE ) 
            {
                /* all right - get the cluster */
                MCLGET(m_curr, M_DONTWAIT);
                if (m_curr->m_flags & M_EXT) 
                {
                    /* all right - it worked - set the new length */
                    m_curr->m_len = MCLBYTES;
                } 
	        else 
                {
                    /* it failed - well, screw it - what can we do ? */
                    m_freem(m);
                    log(LOG_ERR, "%s: Failed to allocate cluster (UDP)!!!\n",
								ifp->if_name);
                    return ENOMEM;
                }
            }

            /* 
             * ok, at this point we've got mbuf pointed by 'm_curr' which 
             * is either a regular one, or a cluster depending on how much
             * data do we have left. The beauty is that from here we dont 
             * care what kind of mbuf it is since all the macros work 
             * correctly on both.
             */
            m_curr->m_len = min( m_curr->m_len, data_length );
  
            /* 
             * Now, we can start copying data into mbuf data area given
             * by mtod() macro. We copy m_curr->m_len bytes.
             */
            bcopy( data_addr, mtod(m_curr, caddr_t), m_curr->m_len );
  
            /* adjust data pointer and remaining length */
            data_addr   += m_curr->m_len;
            data_length -= m_curr->m_len;
  
            /* done with this mbuf - allocate the next one if needed */
            if (data_length > 0) 
            {
                MGET(m_curr->m_next, M_DONTWAIT, MT_DATA);
                if (m_curr->m_next == NULL) 
                {
                    /* it failed - well, screw it - we cant do much here */
                    m_freem(m);
                    log(LOG_ERR, "%s: Failed to allocate mbuf (UDP)!!!\n",
							ifp->if_name);
                    return ENOMEM;
                }
                /* set m_curr to point to the new one, and set up initial length
                 */
		m_curr = m_curr->m_next;
                m_curr->m_len = MLEN;
            }
        } /* while ... */

        /* 
         * Ok, at this point we've got an mbuf chain pointed by m which contains
         * a UDP management reply - now we just have to pass it up the stack -
         * and that's it !!!
         * Note that we neither report those packets to BPF nore update any 
         * stats. Those packets are meant to be invisible ;-)
         * Also we run inside splimp already - do there's no need to call it 
         * again.
	 */
 	s = splimp();
        schednetisr(NETISR_IP);
        if (IF_QFULL(&ipintrq)) 
        {
            /* oh, no - IP queue is full - well - screw it. We cant do much 
             * here. 
	     */ 
            log(LOG_ERR, "%s: IP queue is full (UDP)!!!\n", ifp->if_name);
            IF_DROP(&ipintrq);
            m_freem(m);
	    splx(s); 
            return -EINVAL;
        }

        /* ok enqueue the packet */
        IF_ENQUEUE(&ipintrq, m);
 	splx(s); 
	dlog((LOG_INFO, "%s: UDP packet passed back to the stack\n", ifp->if_name));
    }

    chdlc_priv_area->udp_pkt_lgth = 0;

    return 0;
}

/* Finished
 * ============================================================================
 * Initialize Receive and Transmit Buffers.
 */

static void init_chdlc_tx_rx_buff( sdla_t* card, struct ifnet *ifp )
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    CHDLC_TX_STATUS_EL_CFG_STRUCT *tx_config = NULL;
    CHDLC_RX_STATUS_EL_CFG_STRUCT *rx_config = NULL;
    int err = 0;

    err = chdlc_read_configure(card);
    if(err != COMMAND_OK) 
	return;

    if(card->hw.type == SDLA_S514) 
    {
       tx_config = (CHDLC_TX_STATUS_EL_CFG_STRUCT *)((u_char*)card->hw.dpmbase +
                (((CHDLC_CONFIGURATION_STRUCT *)mb->data)->
                            ptr_CHDLC_Tx_stat_el_cfg_struct));
       rx_config = (CHDLC_RX_STATUS_EL_CFG_STRUCT *)((u_char*)card->hw.dpmbase +
                (((CHDLC_CONFIGURATION_STRUCT *)mb->data)->
                            ptr_CHDLC_Rx_stat_el_cfg_struct));

       /* Setup Head and Tails for buffers */
       card->u.c.txbuf_base = (void *)((u_char*)card->hw.dpmbase +
                tx_config->base_addr_Tx_status_elements);
       card->u.c.txbuf_last = (CHDLC_DATA_TX_STATUS_EL_STRUCT *)  
                card->u.c.txbuf_base +
		(tx_config->number_Tx_status_elements - 1);

       card->u.c.rxbuf_base = (void *)((u_char*)card->hw.dpmbase +
                rx_config->base_addr_Rx_status_elements);
       card->u.c.rxbuf_last = (CHDLC_DATA_RX_STATUS_EL_STRUCT *)
                card->u.c.rxbuf_base +
		(rx_config->number_Rx_status_elements - 1);

       /* Set up next pointer to be used */
       card->u.c.txbuf = (void *)((u_char*)card->hw.dpmbase +
                tx_config->next_Tx_status_element_to_use);
       card->u.c.rxmb = (void *)((u_char*)card->hw.dpmbase +
                rx_config->next_Rx_status_element_to_use);
    }
    else 
    {
       tx_config = (CHDLC_TX_STATUS_EL_CFG_STRUCT *)((u_char*)card->hw.dpmbase +
		(((CHDLC_CONFIGURATION_STRUCT *)mb->data)->
			ptr_CHDLC_Tx_stat_el_cfg_struct % SDLA_WINDOWSIZE));

       rx_config = (CHDLC_RX_STATUS_EL_CFG_STRUCT *)((u_char*)card->hw.dpmbase +
		(((CHDLC_CONFIGURATION_STRUCT *)mb->data)->
		ptr_CHDLC_Rx_stat_el_cfg_struct % SDLA_WINDOWSIZE));

       /* Setup Head and Tails for buffers */
       card->u.c.txbuf_base = (void *)((u_char*)card->hw.dpmbase +
		(tx_config->base_addr_Tx_status_elements % SDLA_WINDOWSIZE));
       card->u.c.txbuf_last =
	        (CHDLC_DATA_TX_STATUS_EL_STRUCT *)card->u.c.txbuf_base
		+ (tx_config->number_Tx_status_elements - 1);
       card->u.c.rxbuf_base = (void *)((u_char*)card->hw.dpmbase +
		(rx_config->base_addr_Rx_status_elements % SDLA_WINDOWSIZE));
       card->u.c.rxbuf_last = 
	       (CHDLC_DATA_RX_STATUS_EL_STRUCT *)card->u.c.rxbuf_base
		+ (rx_config->number_Rx_status_elements - 1);

       /* Set up next pointer to be used */
       card->u.c.txbuf = (void *)((u_char*)card->hw.dpmbase +
		(tx_config->next_Tx_status_element_to_use % SDLA_WINDOWSIZE));
       card->u.c.rxmb = (void *)((u_char*)card->hw.dpmbase +
		(rx_config->next_Rx_status_element_to_use % SDLA_WINDOWSIZE));
    }

    /* Setup Actual Buffer Start and end addresses */
    card->u.c.rx_base = rx_config->base_addr_Rx_buffer;
    card->u.c.rx_top  = rx_config->end_addr_Rx_buffer;

}

/* 
 * ============================================================================
 * Perform Interrupt Test by running READ_CHDLC_CODE_VERSION command MAX_INTR
 * _TEST_COUNTER times.
 */
static int chdlc_intr_test( sdla_t* card, struct ifnet *ifp )
{
    CHDLC_MAILBOX_STRUCT* mb = card->mbox;
    SHARED_MEMORY_INFO_STRUCT* flags = card->u.c.flags;
    int intr_mode = 0, intr_permission = 0;
    int err = 0, i;

    Intr_test_counter = 0;

    intr_permission = flags->interrupt_info_struct.interrupt_permission;
    intr_mode = chdlc_read_intr_mode(card);
    
    err = chdlc_set_intr_mode(card, APP_INT_ON_COMMAND_COMPLETE); 
    if (err == CMD_OK) 
    {
	for (i = 0; i < MAX_INTR_TEST_COUNTER; i ++) 
	{
	    mb->buffer_length  = 0;
	    mb->command = READ_CHDLC_CODE_VERSION;
	    err = sdla_exec(mb) ? mb->return_code : CMD_TIMEOUT;
	    if (err != CMD_OK) 
	        chdlc_error(card, err, mb);
	}
    	err = chdlc_set_intr_mode(card, intr_mode);
        flags->interrupt_info_struct.interrupt_permission = intr_permission;
    }

    if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER))
    {
	log(LOG_ERR, "%s: Interrupt test failed (%d)\n",
					card->devname, Intr_test_counter);
	log(LOG_ERR, "%s: Please choose another interrupt\n",
					card->devname);
	return -EIO;
    }
    log(LOG_INFO, "%s: Interrupt test passed(%d)\n",
					card->devname, Intr_test_counter);
    card->configured = 1;
    return 0;
}

/* 
 * =============================================================================
 * Determine what type of UDP call it is. CPIPEAB ?
 */
static int udp_pkt_type(caddr_t ptr, sdla_t* card)
{
    chdlc_udp_pkt_t *chdlc_udp_pkt = (chdlc_udp_pkt_t*)ptr; 
    unsigned int udp_port = card->wandev.udp_port;

    if (!strncmp(chdlc_udp_pkt->wp_mgmt.signature,UDPMGMT_SIGNATURE,8) &&
	(chdlc_udp_pkt->udp_pkt.udp_dst_port == ntohs(udp_port)) &&
	(chdlc_udp_pkt->ip_pkt.protocol == UDPMGMT_UDP_PROTOCOL) &&
	(chdlc_udp_pkt->wp_mgmt.request_reply == UDPMGMT_REQUEST))
    {
#ifdef WANDEBUG_CHDLC
        chdlc_dump_udp(chdlc_udp_pkt, udp_port);
#endif
	return UDP_CPIPE_TYPE;
    }
    return UDP_INVALID_TYPE;
}

/*
 * ============================================================================
 * Set PORT state.
 */
static void port_set_state (sdla_t *card, int state)
{
    if (card->u.c.state != state)
    {
        switch (state)
        {
        case WAN_CONNECTED:
 	    log(LOG_INFO, "%s: Link connected!\n", card->devname);
	break;
	case WAN_CONNECTING:
   	    log(LOG_INFO, "%s: Link connecting...\n", card->devname);
        break;
	case WAN_DISCONNECTED:
	    log(LOG_INFO, "%s: Link disconnected!\n", card->devname);
        break;
        }
        card->wandev.state = card->u.c.state = state;
    }
}

/* Finished
 * ============================================================================
 * Timer interrupt handler.
 */
void timer_intr(sdla_t *card)
{
    struct ifnet *ifp = NULL;
    chdlc_private_area_t* chdlc_priv_area = NULL;
    SHARED_MEMORY_INFO_STRUCT* flags = NULL;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	return;
    } 
    ifp = card->wandev.if_head->ifp;
    chdlc_priv_area = ifp->if_softc;
    /* process a udp call if pending */
    if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_UDP)
    {
	process_udp_mgmt_pkt(card, ifp, chdlc_priv_area);
        chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
    }
    /* read communicationstatistics if required */
    if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_UPDATE)
    {
	update_comms_stats(card, chdlc_priv_area);
	if (!(--chdlc_priv_area->update_comms_stats))
            chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
    }
    /* Only disable the timer interrupt if there are no udp or statistics 
     * updates pending
     */
    if (!chdlc_priv_area->timer_int_enabled)
    {
        flags = card->u.c.flags;
        flags->interrupt_info_struct.interrupt_permission &= ~APP_INT_ON_TIMER;
    }
}

#ifdef WANDEBUG_CHDLC
static void 
chdlc_dump_mailbox(CHDLC_MAILBOX_STRUCT *mb)
{
    log(LOG_INFO, "Opp Flag      = 0x%x\n", mb->opp_flag);
    log(LOG_INFO, "Command       = 0x%x\n", mb->command);
    log(LOG_INFO, "Buffer Length = 0x%x\n", mb->buffer_length);
    log(LOG_INFO, "Return code   = 0x%x\n", mb->return_code);
}
static void
chdlc_dump_udp(chdlc_udp_pkt_t *chdlc_udp_pkt, unsigned int udp_port)
{
    if (chdlc_udp_pkt->wp_mgmt.signature)
        log(LOG_INFO, "Signature     = %s(%s)\n", 
		chdlc_udp_pkt->wp_mgmt.signature, UDPMGMT_SIGNATURE);
    log(LOG_INFO, "UDP port      = %d(%d)\n", 
		chdlc_udp_pkt->udp_pkt.udp_dst_port, ntohs(udp_port));
    log(LOG_INFO, "Protocol      = %d(%d)\n", 
		chdlc_udp_pkt->ip_pkt.protocol, UDPMGMT_UDP_PROTOCOL);
    log(LOG_INFO, "Request reply = %d(%d)\n", 
		chdlc_udp_pkt->wp_mgmt.request_reply, UDPMGMT_REQUEST);
}
#endif
/****** End ****************************************************************/
