/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_fr.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
 * sdla_fr.c	WANPIPE(tm) Multiprotocol WAN Link Driver. Frame relay module.
 *
 * Author(s):	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ===========================================================================
 * Dec 10, 1999 Alex Feldman	o Updated to FreeBSD 3.2
 * Jun 02, 1999	Gideon Hack     o Updated for S514 support.
 * Jan 02, 1997	Gene Kozin	o Initial version.
 ****************************************************************************
 */

#include "bpfilter.h"

#include <stddef.h>          /* pretty standard includes follow ... */
#include <sys/param.h> 
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>


#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>

#define _GNUC_
#include <net/wanpipe.h>
#include <net/sdla_fr.h>

#include <netinet/in.h>
#include <netinet/in_var.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#ifdef WANDEBUG_FR
 #define dlog(x) log x
#else
 #define dlog(x)
#endif
 
/****** Defines & Macros ****************************************************/

#define SYSTEM_TICK ticks

#define	MAX_CMD_RETRY	10		/* max number of firmware retries */

#define	FR_HEADER_LEN	8		/* max encapsulation header size */
#define	FR_CHANNEL_MTU	1500		/* unfragmented logical channel MTU */

/* IP over Frame Relay encapsulation ( 2 bytes )*/
#define FR_IP_HEADER			0xCC03

/* InArp request retransmission timeout in seconds */
#define INARP_REQUEST_TIMEOUT 20

/* Q.922 frame types */
#define	Q922_UI		0x03		/* Unnumbered Info frame */
#define	Q922_XID	0xAF		

/* DLCI configured or not */
#define DLCI_NOT_CONFIGURED	0x00
#define DLCI_CONFIG_PENDING	0x01
#define DLCI_CONFIGURED		0x02

/* CIR enabled or not */
#define CIR_ENABLED	0x00
#define CIR_DISABLED	0x01

#define WANPIPE 0x00
#define API     0x01
#define FRAME_RELAY_API 1

/* For handle_IPXWAN() */
#define CVHexToAscii(b) (((unsigned char)(b) > (unsigned char)9) ? ((unsigned char)'A' + ((unsigned char)(b) - (unsigned char)10)) : ((unsigned char)'0' + (unsigned char)(b)))
 
/****** Data Structures *****************************************************/

/* This is an extention of the 'struct device' we create for each network
 * interface to keep the rest of channel-specific data.
 */
typedef struct fr_channel
{
	char name[WAN_IFNAME_SZ+1];	/* interface name, ASCIIZ */
	unsigned dlci_configured  ;	/* check whether configured or not */
	unsigned cir_status;		/* check whether CIR enabled or not */
	unsigned dlci;			/* logical channel number */
	unsigned cir;			/* committed information rate */
	unsigned bc;			/* committed burst size */
	unsigned be;			/* excess burst size */
	unsigned mc;			/* multicast support on or off */
	unsigned tx_int_status;		/* Transmit Interrupt Status */	
	unsigned short pkt_length;	/* Packet Length */
	unsigned long router_start_time;/* Router start time in seconds */
	unsigned long tick_counter;	/* counter for transmit time out */
	char dev_pending_devtint;	/* interface pending dev_tint() */
	char state;			/* channel state */
	void *dlci_int_interface;	/* pointer to the DLCI Interface */ 
	unsigned long IB_addr;		/* physical address of Interface Byte */
	unsigned long state_tick;	/* time of the last state change */
	unsigned char enable_IPX;	/* Enable/Disable the use of IPX */
	unsigned long network_number;	/* Internal Network Number for IPX*/
	sdla_t *card;			/* -> owner */
	unsigned route_flag;		/* Add/Rem dest addr in route tables */
	unsigned inarp;			/* Inverse Arp Request status */ 
	int inarp_interval;		/* Time between InArp Requests */
	unsigned long inarp_tick;	/* InArp jiffies tick counter */
	struct enet_statistics ifstats;
	if_send_stat_t drvstats_if_send;
        rx_intr_stat_t drvstats_rx_intr;
        pipe_mgmt_stat_t drvstats_gen;

	unsigned char usedby;  /* Used by WANPIPE or API */

	unsigned long router_up_time;

	unsigned short transmit_length;
	char transmit_buffer[FR_MAX_NO_DATA_BYTES_IN_FRAME];

        /* Special members for FreeBSD */
  	int         start;     /* 1/0 - opened/closed interface (dev->start) */ 
        int         fr_inarp_reply_expected; /* */
        int	    fr_inarp_reply_timer;    /* timer counter for reply */
        int         fr_resend_inarp;	     /* set when we need to resend*/
        arphdr_fr_t fr_inarp_pkt;	     /* this packet 		     */
	/* new destination address */
	unsigned long               dst_addr;
} fr_channel_t;

/* Route Flag options */
#define NO_ROUTE	0x00
#define ADD_ROUTE 	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE 	0x03

/* inarp options */
#define INARP_NONE		0x00
#define INARP_REQUEST		0x01
#define INARP_CONFIGURED	0x02

/* reasons for enabling the timer interrupt on the adapter */
#define TMR_INT_ENABLED_UDP   	0x01
#define TMR_INT_ENABLED_UPDATE 	0x02
#define TMR_INT_ENABLED_INARP 	0x03


typedef struct dlci_status
{
	unsigned short dlci	PACKED;
	unsigned char state	PACKED;
} dlci_status_t;

typedef struct dlci_IB_mapping
{
	unsigned short dlci		PACKED;
	unsigned long  addr_value	PACKED;
} dlci_IB_mapping_t;

/* This structure is used for DLCI list Tx interrupt mode.  It is used to
   enable interrupt bit and set the packet length for transmission
 */
typedef struct fr_dlci_interface 
{
	unsigned char gen_interrupt	PACKED;
	unsigned short packet_length	PACKED;
	unsigned char reserved		PACKED;
} fr_dlci_interface_t;

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;

extern int ip_rt_ioctl(unsigned int, void *);
extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);

/* variable for keeping track of number of interrupts generated during 
 * interrupt test routine 
 */
static int Intr_test_counter;
/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update(wan_device_t *wandev);
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf);
static int del_if(wan_device_t *wandev, struct ifnet *ifp);

/* WANPIPE-specific entry points */
static int wpf_exec(struct sdla *card, void *u_cmd, void *u_data);

/* Network device interface */
static int fr_close(struct ifnet *ifp);

/* ifnet entry points */
static int fr_output   		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		      struct rtentry*));		
static void fr_start 		/* initiate output routine  */
		__P((struct ifnet *));
static int fr_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static void fr_watchdog		/* timer routine */
		__P((struct ifnet*));
static int fr_init		/* init routine */
		__P((struct ifnet*));

static void fr_up(struct ifnet *);
static void fr_down(struct ifnet *);

static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp,
                                struct mbuf *m);

/* Interrupt handlers */
static void fr_isr(sdla_t *card);
static void rx_intr(sdla_t *card);
static void tx_intr(sdla_t *card);
static void timer_intr(sdla_t *card);
static void spur_intr(sdla_t *card);

/* Frame relay firmware interface functions */
static int fr_read_version(sdla_t *card, char *str);
static int fr_configure(sdla_t *card, fr_conf_t *conf);
static int fr_dlci_configure(sdla_t *card, fr_dlc_conf_t *conf, unsigned dlci);
static int fr_init_dlci(sdla_t *card, fr_channel_t *chan);
static int fr_set_intr_mode (sdla_t *card, unsigned mode, unsigned mtu, unsigned short timeout);
static int fr_comm_enable(sdla_t *card);
static int fr_comm_disable(sdla_t *card);
static int fr_get_err_stats(sdla_t *card);
static int fr_get_stats(sdla_t *card);
static int fr_add_dlci(sdla_t *card, int dlci);
static int fr_activate_dlci(sdla_t *card, int dlci);
static int fr_delete_dlci (sdla_t* card, int dlci);
static int fr_issue_isf(sdla_t *card, int isf);
static int fr_write(sdla_t *card, int dlci, unsigned char attr, int len);
static int fr_send(sdla_t *card, int dlci, unsigned char attr, int len,
	void *buf, int snd_encp);

/* Firmware asynchronous event handlers */
static int fr_event(sdla_t *card, int event, fr_mbox_t *mbox);
static int fr_modem_failure(sdla_t *card, fr_mbox_t *mbox);
static int fr_dlci_change(sdla_t *card, fr_mbox_t *mbox);

/* Miscellaneous functions */
static int update_chan_state(struct ifnet *ifp);
static void set_chan_state(struct ifnet *ifp, int state);
static struct ifnet *find_channel(sdla_t *card, unsigned dlci);
static int is_tx_ready(sdla_t *card, fr_channel_t *chan);
static unsigned int dec_to_uint(unsigned char *str, int len);
static int reply_udp( unsigned char *data, unsigned int mbox_len );

static int fr_intr_test(sdla_t* card);
static void init_chan_statistics( fr_channel_t* chan );
static void init_global_statistics( sdla_t* card );
static void read_DLCI_IB_mapping( sdla_t* card, fr_channel_t* chan );
static void setup_for_delayed_transmit(struct ifnet* ifp);
static void process_route (void *arg);


/* Inverse ARP and Dynamic routing functions */
int process_ARP(arphdr_1490_t *ArpPacket, sdla_t *card, struct ifnet *ifp);
int is_arp(void *buf);
static void send_inarp_request(sdla_t *card, struct ifnet *ifp);
static void send_inarp_packet(sdla_t *card, arphdr_fr_t *arphdr, 
						struct ifnet *ifp);

/* Udp management functions */
static int process_udp_mgmt_pkt(sdla_t *card);
static int udp_pkt_type(struct mbuf *m, sdla_t* card);
static int store_udp_mgmt_pkt(int udp_type, char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, int dlci);

/* IPX functions */
#ifdef SUPPORT_IPX
static void switch_net_numbers(unsigned char *sendpacket,
			unsigned long network_number, unsigned char incoming);
static int handle_IPXWAN(unsigned char *sendpacket, char *devname,
			unsigned char enable_IPX, unsigned long network_number);
#endif

unsigned short calc_checksum (char *, int);


/****** Public Functions ****************************************************/

/*============================================================================
 * Frame relay protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpf_init(sdla_t *card, wandev_conf_t *conf)
{
    int err;
    union
    {
	char str[80];
        fr_conf_t cfg;
    } u;
    fr_buf_info_t* buf_info;
    int i;
    int s = splimp();

    dlog((LOG_DEBUG, "%s: Starting Frame Relay initialization\n", 
						card->wandev.name));
    /* Verify configuration ID */
    if (conf->config_id != WANCONFIG_FR) 
    {
	log(LOG_ERR, "%s: invalid configuration ID %u!\n",
				card->devname, conf->config_id);
	splx(s);
	return -EINVAL;
    }

    /* Initialize protocol-specific fields of adapter data space */
    switch (card->hw.fwid) 
    {
    case SFID_FR508:
	card->mbox  = (void*)((unsigned char*)card->hw.dpmbase + 
			      FR508_MBOX_OFFS);
	card->flags = (void*)((unsigned char*)card->hw.dpmbase + 
			      FR508_FLAG_OFFS);
	if(card->hw.type == SDLA_S514) 
	{
	    card->mbox = (void*)((unsigned char*)card->mbox + 
				FR_MB_VECTOR);
            card->flags = (void*)((unsigned char*)card->flags +
				FR_MB_VECTOR);
	}
        card->isr = &fr_isr;
	break;

    default:
	splx(s);
	return -EINVAL;
    }

    /* Read firmware version.  Note that when adapter initializes, it
     * clears the mailbox, so it may appear that the first command was
     * executed successfully when in fact it was merely erased. To work
     * around this, we execute the first command twice.
     */
    if (fr_read_version(card, NULL) || fr_read_version(card, u.str))
    {
	splx(s);
	return -EIO;
    }

    log(LOG_INFO, "%s: running frame relay firmware v%s\n", card->devname, u.str);

    /* Adjust configuration */
    conf->mtu += FR_HEADER_LEN;
    conf->mtu = (conf->mtu >= MIN_LGTH_FR_DATA_CFG) ?
			min(conf->mtu, FR_MAX_NO_DATA_BYTES_IN_FRAME) :
                        FR_CHANNEL_MTU + FR_HEADER_LEN;
     
    conf->bps = min(conf->bps, 2048000);

    /* Initialze the configuration structure sent to the board to zero */
    bzero(&u.cfg, sizeof(u.cfg));

    bzero(card->u.f.dlci_to_dev_map, sizeof(card->u.f.dlci_to_dev_map));

    /* Configure adapter firmware */
    u.cfg.mtu	= conf->mtu;
    u.cfg.kbps	= conf->bps / 1000;

    u.cfg.cir_fwd = u.cfg.cir_bwd = 16;
    u.cfg.bc_fwd  = u.cfg.bc_bwd = 16;

    u.cfg.options	= 0x0000;
    log(LOG_INFO, "%s: Global CIR enabled by Default\n", card->devname);

    switch (conf->u.fr.signalling) 
    {
    case WANOPT_FR_ANSI:
        u.cfg.options = 0x0000; 
        break;		

    case WANOPT_FR_Q933:	
        u.cfg.options |= 0x0200; 
        break;

    case WANOPT_FR_LMI:	
        u.cfg.options |= 0x0400; 
        break;

    case WANOPT_NO:
        u.cfg.options |= 0x0800; 
        break;

    default:
	log(LOG_ERR, "%s: Illegal Signalling option\n", card->wandev.name);
	splx(s);
	return -EINVAL;
    }
    card->wandev.signalling = conf->u.fr.signalling;

    if (conf->station == WANOPT_CPE) 
    {
	if (conf->u.fr.signalling == WANOPT_NO)
	{
	    log(LOG_ERR, "%s: For NO signalling, station must be set to Node!\n",
							card->devname);
	    splx(s);
	    return -EINVAL; 
	}
	u.cfg.station = 0;
	u.cfg.options |= 0x8000;	/* auto config DLCI */
	card->u.f.dlci_num  = 0;
    } 
    else 
    {
	u.cfg.station = 1;	/* switch emulation mode */

	/* For switch emulation we have to create a list of dlci(s)
	 * that will be sent to be global SET_DLCI_CONFIGURATION 
	 * command in fr_configure() routine. 
	 */

	card->u.f.dlci_num  = min(max(conf->u.fr.dlci_num, 1), 100);

	for ( i = 0; i < card->u.f.dlci_num; i++) 
	    card->u.f.node_dlci[i] = (unsigned short) 
		              conf->u.fr.dlci[i] ? conf->u.fr.dlci[i] : 16;
    }

    if (conf->clocking == WANOPT_INTERNAL)
	u.cfg.port |= 0x0001;

    if (conf->interface == WANOPT_RS232)
	u.cfg.port |= 0x0002;

    if (conf->u.fr.t391)
	u.cfg.t391 = min(conf->u.fr.t391, 30);
    else
	u.cfg.t391 = 5;

    if (conf->u.fr.t392)
	u.cfg.t392 = min(conf->u.fr.t392, 30);
    else
	u.cfg.t392 = 15;

    if (conf->u.fr.n391)
	u.cfg.n391 = min(conf->u.fr.n391, 255);
    else
	u.cfg.n391 = 2;

    if (conf->u.fr.n392)
	u.cfg.n392 = min(conf->u.fr.n392, 10);
    else
	u.cfg.n392 = 3;	

    if (conf->u.fr.n393)
	u.cfg.n393 = min(conf->u.fr.n393, 10);
    else
	u.cfg.n393 = 4;

    if (fr_configure(card, &u.cfg))
    {
	splx(s);
	return -EIO;
    }

    if (card->hw.type == SDLA_S514) 
    {
        buf_info = (void*)((unsigned char*)card->hw.dpmbase + FR_MB_VECTOR +
						FR508_RXBC_OFFS);

        card->rxmb = (void*)(buf_info->rse_next + 
		    (unsigned long)card->hw.dpmbase);

        card->u.f.rxmb_base =
              (void*)(buf_info->rse_base + (unsigned long)card->hw.dpmbase);

        card->u.f.rxmb_last =
                   (void*)(buf_info->rse_base +
                   (buf_info->rse_num - 1) * sizeof(fr_rx_buf_ctl_t) +
                        (unsigned long)card->hw.dpmbase);
    }
    else 
    {
	buf_info = (void*)((unsigned char*)card->hw.dpmbase +  
			   FR508_RXBC_OFFS);

	card->rxmb = (void*)(buf_info->rse_next -
		FR_MB_VECTOR + (unsigned long)card->hw.dpmbase);
	
	card->u.f.rxmb_base =
			(void*)(buf_info->rse_base -
			FR_MB_VECTOR + (unsigned long)card->hw.dpmbase);
		
	card->u.f.rxmb_last =
			(void*)(buf_info->rse_base +
			(buf_info->rse_num - 1) * sizeof(fr_rx_buf_ctl_t) -
			FR_MB_VECTOR + (unsigned long)card->hw.dpmbase);
    }

    card->u.f.rx_base = buf_info->buf_base;
    card->u.f.rx_top  = buf_info->buf_top;

    card->u.f.tx_interrupts_pending = 0;

    card->wandev.mtu	   = conf->mtu;
    card->wandev.bps	   = conf->bps;
    card->wandev.interface = conf->interface;
    card->wandev.clocking  = conf->clocking;
    card->wandev.station   = conf->station;
    card->poll		   = NULL; 
    card->exec		   = &wpf_exec;
    card->wandev.update	   = &update;
    card->wandev.new_if	   = &new_if;
    card->wandev.del_if	   = &del_if;
    card->wandev.ttl	   = conf->ttl;
    card->wandev.udp_port  = conf->udp_port; 
    /* Update card state */
    card->wandev.state	   = WAN_DISCONNECTED;

    /* Intialize global statistics for a card */
    init_global_statistics( card );

    card->TracingEnabled          = 0;

    /* Interrupt Test */
    splx(s);

    Intr_test_counter = 0;
    card->intr_mode = INTR_TEST_MODE;
    err = fr_intr_test( card );

    if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER)) 
    {
	log(LOG_ERR, "%s: Interrupt Test Failed, Counter: %d\n", 
					card->devname, Intr_test_counter);
	log(LOG_ERR, "Please choose another interrupt\n");
	return -EIO;
    }

    log(LOG_INFO, "%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
    return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/*============================================================================
 * Update device status & statistics.
 */
static int update (wan_device_t* wandev)
{
    sdla_t          *card = NULL;
    unsigned long timeout;
    fr508_flags_t  *flags = NULL;

    /* sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL))
	return -EFAULT;

    if (wandev->state == WAN_UNCONFIGURED)
	return -ENODEV;

    card = wandev->private;
    flags = card->flags;

    card->u.f.update_comms_stats = 1;
    card->u.f.timer_int_enabled |= TMR_INT_ENABLED_UPDATE;
    flags->imask |= FR_INTR_TIMER;
    timeout = SYSTEM_TICK;
    for(;;) 
    {
	if (card->u.f.update_comms_stats == 0)
	    break;
        if ((SYSTEM_TICK - timeout) > (1 * hz))
	{
    	    card->u.f.update_comms_stats = 0;
 	    return -EAGAIN;
	}
    }
    return 0;
}

/*============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf)
{
    sdla_t* card = wandev->private;
    fr_channel_t* chan = NULL;
    int dlci = 0;
    int err = 0;

    if ((conf->name[0] == '\0') || (strlen(conf->name) > WAN_IFNAME_SZ)) 
    {
	log(LOG_ERR, "%s: invalid interface name!\n", card->devname);
	return -EINVAL;
    }

    /* allocate and initialize private data */
    chan = malloc(sizeof(fr_channel_t), M_DEVBUF, M_NOWAIT);

    if (chan == NULL)
	return -ENOMEM;

    bzero(chan, sizeof(fr_channel_t));
    strcpy(chan->name, conf->name);
    chan->card = card;

    /* verify media address */
    if (is_digit(conf->addr[0])) 
    {
	dlci = dec_to_uint(conf->addr, 0);

	if (dlci && (dlci <= HIGHEST_VALID_DLCI)) 
	    chan->dlci = dlci;
	else 
	{
	    log(LOG_ERR, "%s: invalid DLCI %u on interface %s!\n",
					wandev->name, dlci, chan->name);
	    err = -EINVAL;
	}
    } 
    else 
    {
	log(LOG_ERR, "%s: invalid media address on interface %s!\n",
						wandev->name, chan->name);
	err = -EINVAL;
    }

    /* Copy network interface name to chan */ 
    sprintf(chan->name, "%s%d", conf->name, dlci);

    /* Setup wanpipe as a router (WANPIPE) or as an API */
    if (strcmp(conf->usedby, "WANPIPE") == 0)
    {
        log(LOG_INFO, "%s: WANPIPE mode enabled for %s\n",
					wandev->name, chan->name);
        chan->usedby = WANPIPE;
    } 
    else 
	if(strcmp(conf->usedby, "API") == 0)
	{
#ifdef FRAME_RELAY_API
	    chan->usedby = API;
            log(LOG_INFO, "%s: API(%d) mode enabled for %s\n", 
			wandev->name, chan->usedby, chan->name);
#else
            log(LOG_INFO, "%s: API Mode is not supported !\n", wandev->name);
            log(LOG_INFO, "%s: API patch can be obtained from Sangoma Tech.\n",
							     	wandev->name);
            err = -EINVAL;
#endif
        }

    if (err) 
    {
        free(chan, M_DEVBUF);
        return err;
    }

    card->u.f.dlci_to_dev_map[dlci] = ifp;

    /* place cir,be,bc and other channel specific information into the
     * chan structure 
     */
    if (conf->cir) 
    {
        chan->cir = max( 1, min( conf->cir, 512 ) );
        chan->cir_status = CIR_ENABLED; 

        /* If CIR is enabled, force BC to equal CIR
         * this solves number of potential problems if CIR is 
         * set and BC is not 
         */
        chan->bc = chan->cir;

        if (conf->be)
            chan->be = max(0, min(conf->be, 511)); 
        else	
            conf->be = 0;

        log(LOG_INFO, "%s: CIR ENABLED for DLCI %d \n", wandev->name,chan->dlci);
        log(LOG_INFO, "%s:     CIR = %d ; BC = %d ; BE = %d\n",
			wandev->name,chan->cir,chan->bc,chan->be);
    }
    else
    {
        chan->cir_status = CIR_DISABLED;
        log(LOG_INFO, "%s: CIR_DISABLED for DLCI %d\n", wandev->name,chan->dlci);
    }

    chan->mc = conf->mc;

    chan->inarp = conf->inarp ? INARP_REQUEST : INARP_NONE;

    chan->inarp_interval = conf->inarp_interval ? conf->inarp_interval : 10;

    chan->dlci_configured = DLCI_NOT_CONFIGURED;

    /* FIXME: IPX disabled in this WANPIPE version */
    if (conf->enable_IPX == WANOPT_YES)
    {
	log(LOG_INFO, "%s: This version of WANPIPE doesn't support IPX\n",
							card->devname);
	free(chan, M_DEVBUF);
	return -EINVAL;
    }
    else
	chan->enable_IPX = WANOPT_NO;
    chan->enable_IPX = conf->enable_IPX;

    if (conf->network_number)
        chan->network_number = conf->network_number;
    else
        chan->network_number = 0xDEADBEEF;

    chan->route_flag = NO_ROUTE;

    init_chan_statistics(chan);

    chan->transmit_length = 0;

    /* prepare network device data space for registration */
    ifp->if_softc          = chan;
    ifp->if_unit           = chan->dlci; /*wandev->unit;*/ /* Alex: wandev->ndev */
    ifp->if_name           = malloc(strlen(conf->name)+1, M_DEVBUF, M_NOWAIT);
    if (ifp->if_name == NULL)
    {
	log(LOG_ERR, "%s%d: Cannot allocate memory!\n", conf->name, chan->dlci);
	free(chan, M_DEVBUF);
	ifp->if_softc = NULL;
	return -ENOMEM;
    } 
    strcpy(ifp->if_name, conf->name);
    ifp->if_mtu            = wandev->mtu;
    ifp->if_type           = IFT_OTHER;
    ifp->if_flags          = IFF_POINTOPOINT;
    ifp->if_output         = fr_output;
    ifp->if_start          = fr_start;
    ifp->if_watchdog       = fr_watchdog;
    ifp->if_init           = (void*)fr_init;
    ifp->if_ioctl          = fr_ioctl;
    ifp->if_snd.ifq_len    = 0;
    ifp->if_snd.ifq_maxlen = ifqmaxlen;

    ifp->if_hdrlen = FR_HEADER_LEN;
    /* For an API, the maximum number of bytes that the stack will pass
     * to the driver is (dev->mtu + dev->hard_header_len). So, adjust the
     * mtu so that a frame of maximum size can be transmitted by the API. 
     */
    if (chan->usedby == API) 
        ifp->if_mtu += (sizeof(api_tx_hdr_t) - FR_HEADER_LEN);

    set_chan_state(ifp, WAN_DISCONNECTED);

    /* Enable Interrupts and Communications */
    if (!wandev->new_if_cnt)
    {
	fr508_flags_t* flags = card->flags;
	wandev->new_if_cnt++;
	/* 
	 * If you enable comms and then set ints, you get a Tx int as you
	 * perform the SET_INT_TRIGGERS command. So, we only set int
	 * triggers and then adjust the interrupt mask (to disable Tx ints)
	 * before enabling comms. 
	 */	
        if (fr_set_intr_mode(card, (FR_INTR_RXRDY | FR_INTR_TXRDY |
		FR_INTR_DLC | FR_INTR_TIMER | FR_INTR_TX_MULT_DLCIs),
		card->wandev.mtu, 0)) 
        {
	    free(chan, M_DEVBUF);
	    return -EIO;
	}

	flags->imask &= ~(FR_INTR_TXRDY | FR_INTR_TIMER);
 
	if (fr_comm_enable(card)) 
	{
	    free(chan, M_DEVBUF);
	    return -EIO;
        }	
	wanpipe_set_state(card, WAN_CONNECTED);
    }
    return 0;
}

/*
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
    sdla_t *card = wandev->private;

    /* Execute shutdown very first time we enter del_if */
    dlog((LOG_DEBUG, "%s%d: Removing interface...\n", ifp->if_name, ifp->if_unit));
    if (!wandev->del_if_cnt) 
    {
	wandev->del_if_cnt++;
	wanpipe_set_state(card, WAN_DISCONNECTED);
	/* Initialize Interrupt handler pointer */
	card->isr = NULL;  
	fr_set_intr_mode(card, 0, 0, 0);
	fr_comm_disable(card);
    }

    /* Close current interface and remove DLCI */ 
    fr_close(ifp);

    /* Initialize newtwork interface */
    ifp->if_output = NULL; 
    ifp->if_start = NULL; 
    ifp->if_ioctl = NULL; 
    ifp->if_watchdog = NULL; 
    ifp->if_init = NULL; 
    ifp->if_flags &= ~(IFF_UP | IFF_RUNNING | IFF_OACTIVE); 
    if (ifp->if_softc)
    {
	free(ifp->if_softc, M_DEVBUF);
	ifp->if_softc = NULL;
    }
    dlog((LOG_DEBUG, "%s%d: Interface removed.\n", ifp->if_name, ifp->if_unit));
    return 0;
}

/****** WANPIPE-specific entry points ***************************************/

/*
 * ============================================================================
 * Execute adapter interface command.
 */
static int wpf_exec (struct sdla* card, void* u_cmd, void* u_data)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err, len;
    fr_cmd_t cmd;

    if (copyin(u_cmd, (void*)&cmd, sizeof(cmd)))
	return -EFAULT;

    /* execute command */
    do
    {
	bcopy(&cmd, &mbox->cmd, sizeof(cmd));
	
	if (cmd.length)
	    if (copyin(u_data, (void*)&mbox->data, cmd.length))
		return -EFAULT;
		
	if (sdla_exec(mbox))
	    err = mbox->cmd.result;
	else 
	    return -EIO;
	
    } while (err && retry-- && fr_event(card, err, mbox));

    /* return result */
    if (copyout((void*)&mbox->cmd, u_cmd, sizeof(fr_cmd_t)))
	return -EFAULT;

    len = mbox->cmd.length;

    if (len && u_data &&
        	copyout((void*)&mbox->data, u_data, len))
	return -EFAULT;

    return 0;
}

/****** Network Device Interface ********************************************/

/*
 * fr_ioctl() -
 *
 */
static int
fr_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
    struct ifaddr *ifa = (struct ifaddr*)data;
    int       s = 0;		    	    /* saved interrupt state */
    int		      error = 0;	    	    /* resulting status */

    s = splimp();

    switch (command) 
    {
    case SIOCSIFADDR:
        dlog((LOG_DEBUG, "%s%d - ioctl(SIOCSIFADDR) called\n",
					ifp->if_name,ifp->if_unit));
        /* check the protocol family - we currently support IP only */
        switch(ifa->ifa_addr->sa_family) 
        {
        case AF_INET:
	    if (fr_init(ifp))
		return EBUSY; 
            break;
        default:
            error = EAFNOSUPPORT;
            break;
        }
        break;

    case SIOCGIFADDR:
        dlog((LOG_DEBUG, "%s%d - ioctl(SIOCGIFADDR) called.\n", 
						ifp->if_name, ifp->if_unit));
        break;

    case SIOCSIFFLAGS:
        dlog((LOG_DEBUG, "%s%d - ioctl(SIOCSIFFLAGS) called.\n",
						ifp->if_name, ifp->if_unit));
        break;

    case SIOCGIFFLAGS:
        dlog((LOG_DEBUG, "%s%d - ioctl(SIOCGIFFLAGS) called.\n",
						ifp->if_name, ifp->if_unit));
        break;

    case SIOCSIFMTU:
        dlog((LOG_DEBUG, "%s%d - ioctl(SIOCSIFMTU) called.\n",
						ifp->if_name, ifp->if_unit));
        error = EINVAL;
        break;

    default:
        dlog((LOG_DEBUG, "%s%d - ioctl(default) called.\n", 
						ifp->if_name, ifp->if_unit));
        error = EINVAL;
        break;
    }

    splx(s);
    return error;
}

/*
 * ============================================================================
 * fr_up() - bring interface UP and RUNNING
 */
static void fr_up(struct ifnet *ifp)
{
    ifp->if_flags |= (IFF_UP | IFF_RUNNING);
}

/*
 *============================================================================
 * fr_down() - Bring interface not UP and NOT RUNNING
 */
static void fr_down(struct ifnet *ifp)
{
    ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
}

/*============================================================================
 * if_wachdog():
 *
 * 
 * 
 * 
 */
static void
fr_watchdog(struct ifnet * ifp)
{
    fr_channel_t *chan = ifp->if_softc;   /* our softc struct */
    sdla_t       *card = chan->card;

    /* see if we've perviously failed to send an InARP packet  - if so */
    /* resend it and reset the flag ...				   */
    if (chan->fr_resend_inarp) 
    {
	/* clear the flag */
	chan->fr_resend_inarp = 0;
	/* resend */
	dlog((LOG_DEBUG, "%s%d: Watchdog tries to send InARP packet ...\n",
						ifp->if_name, ifp->if_unit));
	send_inarp_packet(chan->card, &chan->fr_inarp_pkt, ifp);
    }

    /* see if we're awaiting for InARP reply - if so decrement the timer */
    /* and if it goes off - resend the request			     */
    if (chan->fr_inarp_reply_expected) 
    {
	if (--chan->fr_inarp_reply_timer == 0) 
	    send_inarp_request(chan->card, ifp);
	else 
 	    /* make sure the watchdog is called again in a second */
	    ifp->if_timer = 1;
    }

    if ((chan->route_flag == ADD_ROUTE || chan->route_flag == REMOVE_ROUTE) &&
	card->poll)
	card->poll(ifp);
    
    return;
}

/* Ok
 * ============================================================================
 * Initialize network interface.
 *
 * This routine is called only once for each interface, during Linux network
 * interface registration.  Returning anything but zero will fail interface
 * registration.
 */
static int fr_init (struct ifnet* ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    sdla_t* 	  card = chan->card;
    int 	     s = 0;
    int 	   err = 0;
    struct timeval  tv;

    if (chan->start)
	return -EBUSY;	/* only one open is allowed */
  
    s = splimp();

    if (card->wandev.signalling == WANOPT_NO)
    {
	log(LOG_INFO, "%s: Signalling set to NO: Mapping DLCI's\n", 
					card->wandev.name);
	if (fr_init_dlci(card, chan))
 	{
	    return -EAGAIN;
	}
    }

    if (card->wandev.station == WANOPT_CPE) 
    {
	/* CPE: issue full status enquiry */
	fr_issue_isf(card, FR_ISF_FSE);
    } 
    else 
    {	/* FR switch: activate DLCI(s) */
		
	/* For Switch emulation we have to ADD and ACTIVATE
	 * the DLCI(s) that were configured with the SET_DLCI_
	 * CONFIGURATION command. Add and Activate will fail if
	 * DLCI specified is not included in the list.
	 *
	 * Also If_open is called once for each interface. But
	 * it does not get in here for all the interface. So
 	 * we have to pass the entire list of DLCI(s) to add 
	 * activate routines.  
	 */ 

	fr_add_dlci(card, chan->dlci);
	fr_activate_dlci(card, chan->dlci);
    }

    log(LOG_INFO, "%s: IF OPEN Successful for DLCI %d\n", 
					card->devname, chan->dlci);

    chan->start = 1; /* identical to dev->start in Linux */
    wanpipe_open(card);
    update_chan_state(ifp);
    microtime(&tv);
    chan->router_start_time = tv.tv_sec;
    splx(s);
    return err;
}

/* 
 * ============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int fr_close (struct ifnet* ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    sdla_t*       card = chan->card;
    int              s = 0;

    s = splimp();

    chan->start = 0; /* identical to dev->start in Linux */
    wanpipe_close(card);
    if (card->wandev.station == WANOPT_NODE)
	fr_delete_dlci (card,chan->dlci);

    log(LOG_INFO, "%s: IF CLOSE: Successful for DLCI %d\n",
		card->devname, chan->dlci);

    splx(s);
    return 0;
}

/* 
 * ============================================================================
 * Send a packet on a network interface.
 * o set tbusy flag (marks start of the transmission) to block a timer-based
 *   transmit from overlapping.
 * o set critical flag when accessing board.
 * o check link state. If link is not up, then drop the packet.
 * o check channel status. If it's down then initiate a call.
 * o pass a packet to corresponding WAN device.
 * o free socket buffer
 *
 * Return:	0	complete (socket buffer must be freed)
 *		non-0	packet may be re-transmitted (tbusy must be set)
 *
 * Notes:
 * 1. This routine is called either by the protocol stack or by the "net
 *    bottom half" (with interrupts enabled).
 * 2. Setting tbusy flag will inhibit further transmit requests from the
 *    protocol stack and can be used for flow control with protocol layer.
 */
static int
fr_output(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst, struct rtentry *rt0)
{
    fr_channel_t         *chan = ifp->if_softc;
    sdla_t               *card = chan->card;
    fr508_flags_t *adptr_flags = card->flags;
    unsigned char  *sendpacket = NULL;
    void                 *data = NULL;
    int               udp_type = 0, 
		     send_data = 0;
    unsigned int           len = 0;
    int 		     s = 0;		       	/* saved int. state */
    int error = 0;

    chan->drvstats_if_send.if_send_entry ++;

    if (m == NULL) 
    {
	/* if we get here, some higher layer thinks we've missed an
	 * tx-done interrupt.
	 */
	log(LOG_ERR, "%s: interface %s got kicked!\n", 
						card->devname, ifp->if_name);
	chan->drvstats_if_send.if_send_mbuf_null ++;
	return 0;
    }

    if ((ifp->if_flags & IFF_OACTIVE) != 0) 
    {
        /* If our device stays busy for at least 5 seconds then we will
         * increment statistics. We expect that our device never stays busy more 
         * than 5 seconds. 
         */
	if ((SYSTEM_TICK - chan->tick_counter) > (5 * hz))
	{
	    ++chan->ifstats.collisions;
	    chan->drvstats_if_send.if_send_tbusy ++;
            chan->drvstats_if_send.if_send_tbusy_timeout ++;
            log(LOG_INFO, "%s: Transmit timed out\n", chan->name);
	} 
    }

    /* For API, the socket family not specified */
    if (chan->usedby != API && dst->sa_family != AF_INET)
    {
	/* protocol family is not supported */
	log(LOG_ERR, "Protocol family (%d) is not isupported ...\n", 
							dst->sa_family);
	return EAFNOSUPPORT;
    }

    len = m->m_len;
    sendpacket = mtod(m, caddr_t);
    data = mtod(m, caddr_t);

    /* check the udp packet type */
    udp_type = udp_pkt_type(m, card);
    if(udp_type != UDP_INVALID_TYPE) 
    {
        if(store_udp_mgmt_pkt(udp_type, UDP_PKT_FRM_STACK, card, m, 
						      chan->dlci)) 
	{
	    adptr_flags->imask |= FR_INTR_TIMER;
            if (udp_type == UDP_FPIPE_TYPE)
		chan->drvstats_if_send.if_send_PIPE_request ++;
        }
	return 0;
    }

    if ((chan->usedby == API) && (len <= sizeof(api_tx_hdr_t))) 
    {
	/* FIXME: increment some error statistic */
	m_freem(m);
	return 0;
    }
 
    /* FIXME: can we do better than sendpacket[2]? */
    if ((chan->usedby == WANPIPE) && (sendpacket[2] == 0x45)) 
    {
	/* check to see if the source IP address is a broadcast or */
        /* multicast IP address */
        if (chk_bcast_mcast_addr(card, ifp, m))
	    return 0;
    }

    if (card->wandev.state != WAN_CONNECTED) 
    {
	chan->drvstats_if_send.if_send_wan_disconnected ++;
	++chan->ifstats.tx_dropped;
        ++card->wandev.stats.tx_dropped;
    } 
    else 
        if (chan->state != WAN_CONNECTED) 
        {
	    chan->drvstats_if_send.if_send_dlci_disconnected ++;
	    update_chan_state(ifp);
            ++chan->ifstats.tx_dropped;
       	    ++card->wandev.stats.tx_dropped;
        } 
	else 
	{
	    send_data = 1;
	    /* If it's an IPX packet */
#ifdef SUPPORT_IPX
	    /* FIXME: can we use an IPX header structure for this logic? */
	    if ((chan->usedby == WANPIPE) &&
	        sendpacket[1] == 0x00 &&
    	        sendpacket[2] == 0x80 &&
    	        sendpacket[6] == 0x81 &&
    	        sendpacket[7] == 0x37) 
	    {
	        if( chan->enable_IPX ) 
	        {
	            switch_net_numbers(sendpacket, chan->network_number, 0);
	        } 
	        else 
	        {
	            /* FIXME: increment some statistic here! */
	            send_data = 0;
	        }
            }
#endif
            if (send_data) 
	    {
	        if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) 
	      				!= (IFF_UP|IFF_RUNNING)) 
	        {
	            /* device is not ready */
	            log(LOG_INFO, "%s%d: Device is not ready ...\n",
					ifp->if_name, ifp->if_unit);
	            return ENETDOWN;
	        }
    	        s = splimp();
	        if (IF_QFULL(&ifp->if_snd))
		{ 
		    /* send queue is full - drop the packet and exit */
		    IF_DROP(&ifp->if_snd);
		    splx(s);
		    log(LOG_INFO, "%s%d: Send queue is full ...\n",
					ifp->if_name, ifp->if_unit);
		    return ENOBUFS;
		}
	        IF_ENQUEUE(&ifp->if_snd, m);

	        if ((ifp->if_flags & IFF_OACTIVE) == 0) 
		    fr_start(ifp);
                splx(s);
	    }
        }
    return (error);
}

/*
 * ============================================================================
 * fr_start() -  called by the system when a packet is added to the 
 *	         output queue for the interface and the device flags
 *	         indicate the device is not currently busy ( IFF_OACTIVE
 *               is clear ).
 *               We will try to transmit the packet. In case there's no
 *  	         available buffer - we will set IFF_OACTIVE, enable 
 *	         transmit interrupt and exit. And then the tx interrupt 
 *	         handler will call if_start() and once its done - it 
 *	         will reset the IFF_OACTIVE.
 */
static void fr_start(struct ifnet * ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    sdla_t		      *card = chan->card;
    fr_tx_buf_ctl_t*     frbuf = NULL;
    fr_mbox_t* mbox = (fr_mbox_t*)((unsigned char*)card->mbox + 0x800);
    struct mbuf        *opkt = NULL;			/* the packet */
    register struct mbuf	 *m = NULL;		/* current mbuf */
    unsigned short	         data_length;	/* data length */
    unsigned long data_addr;
    unsigned char attr = 0;
    int 	   s = 0;	         /* saved int. state */
    int 	   err = 0;
    char *buffer = NULL, *data = NULL;

    s = splimp();

    /* We're trying to drain the queue as much as possible here because 
     * if we manage to get more than one packet into transmit queue     
     * ( line-down/line-up scenario, or heavy load ) and take only one  
     * at a time out we get an effective delay ...		      
     * BF-0001 
     */ 

    while ((opkt = ifp->if_snd.ifq_head) != NULL) 
    {	   /* BF-0001 */
	if (!is_tx_ready(card, chan))
	{
	    dlog((LOG_DEBUG, "%s%d: TR buffer are full. Enable TX INTR for DLCI %d\n", 
					ifp->if_name, ifp->if_unit, chan->dlci));
	    /* Enable transmit interrupt */ 
	    setup_for_delayed_transmit(ifp);
	    /* Set tick_counter */
	    chan->tick_counter = SYSTEM_TICK; 
	    chan->drvstats_if_send.if_send_no_bfrs ++;
	    /* Set interface is active */
	    ifp->if_flags |= IFF_OACTIVE;
            splx(s);
    	    return;
	}
	/* go through mbuf chain to calculate the total length and copy packet
	 * to local buffer.
	 */
	data_length = 0;
	for( m = opkt; m != NULL; m = m->m_next)
	    data_length += m->m_len;
	buffer = malloc(data_length+1, M_DEVBUF, M_NOWAIT);
	if (buffer == NULL)
	{
	    /* FIXME: update error statistics */
	    log(LOG_ERR, "%s%d: Cannot allocte buffer!\n", ifp->if_name, ifp->if_unit);
	    splx(s);
	    return; 
	}
	bzero(buffer, data_length+1);
	data = buffer;
	data_length = 0;
	for( m = opkt; m != NULL; m = m->m_next)
	{
	    bcopy(mtod(m, caddr_t), &data[data_length], m->m_len);
	    data_length += m->m_len;
	}

	attr = 0;
	/* 
	 * For an API transmission, get rid of the API header.
	 * For not API, add 2 bytes encapsulation FR header. 
	 */
	if (chan->usedby == API) 
	{
	    api_tx_hdr_t* api_tx_hdr;
            api_tx_hdr = (api_tx_hdr_t*)data;
            attr = api_tx_hdr->attr;
	    data += sizeof(api_tx_hdr_t);
	    data_length -= sizeof(api_tx_hdr_t);
	    err = fr_write(card, chan->dlci, attr, data_length);
	}
	else
	    err = fr_write(card, chan->dlci, attr, data_length+2);
	if (err) 
	{
	    switch(err)
	    {
	    case FRRES_DISABLED:
		log(LOG_ERR, "%s: Failed to send data. Communication disabled.\n",
							card->devname);
		break;

	    case FRRES_INOPERATIVE:
		log(LOG_ERR, "%s: Failed to send data. Channel is inoperative.\n",
							card->devname);
		break;

	    case FRRES_DLCI_INACTIVE:
		log(LOG_ERR, "%s: Failed to send data. DLCI %d is inactive.\n",
						card->devname, chan->dlci);
		break;

	    case FRRES_TOO_MANY:
		log(LOG_ERR, "%s: Failed to send data %d bytes frame on DLCI %d\n,",
				card->devname, data_length, chan->dlci);
		log(LOG_ERR, "due to it can't be fit in a single Tc period. \n");
		log(LOG_ERR, "Maximum frame size that can fit within Tc period\n");
		log(LOG_ERR, "is %d bytes. Please, either increase CIR\n", 
							mbox->cmd.length);
		log(LOG_ERR, "for this DLCI or decrease the MTU for device\n"); 
		break;
		
	    case FRRES_CIR_OVERFLOW:
	    case FRRES_BUFFER_OVERFLOW:
		dlog((LOG_INFO, "%s%d: Failed to send data (no buffers/CIR).\n",
						ifp->if_name, ifp->if_unit));
		dlog((LOG_INFO, "Enabling tx interrupt for DLCI %d\n", chan->dlci)); 
	        setup_for_delayed_transmit(ifp);
                chan->drvstats_if_send.if_send_adptr_bfrs_full++;
		/* mark interface is active */
		ifp->if_flags |= IFF_OACTIVE;
	        break;
	    default:
		log(LOG_ERR, "%s%d: Failed to send data (0x%02xh)\n", 
					ifp->if_name, ifp->if_unit, err);
                chan->drvstats_if_send.if_send_dlci_disconnected++;
	        chan->ifstats.tx_dropped++;
	        chan->card->wandev.stats.tx_dropped++;
	        break;
	    }
	    if (buffer)
		free(buffer, M_DEVBUF);
	    splx(s);
	    return;
	}

	if (card->hw.type == SDLA_S514)
	    frbuf = (void*)(*(unsigned long*)mbox->data + 
		    (unsigned long)card->hw.dpmbase);
	else
	    frbuf = (void*)(*(unsigned long*)mbox->data - 
			FR_MB_VECTOR + (unsigned long)card->hw.dpmbase);

	data_addr = frbuf->offset;
	/* 
	 * Add the encapsulation bytes for IP : 03 CC - reuse tx_elem.flag.
	 * (only for not API).
	 */
	if (chan->usedby != API)
	{
	    unsigned short ip_over_fr_header = FR_IP_HEADER;

            sdla_poke(&card->hw, data_addr, &ip_over_fr_header, 2 );
	    data_addr += 2;
        }

	/* Copy local buffer to the board */
	sdla_poke(&card->hw, data_addr, data, data_length);

	/* data length - offset 5, length 2 */
	if (chan->usedby == API)
	    frbuf->length = data_length;
	else
	    frbuf->length = data_length + 2;
	/* dlci number - offset 7, length 2 */
	frbuf->dlci	= chan->dlci;
	/* FECN/BECN bits - offset 9, length 1 */
	frbuf->attr	= 0;
	/* flip the opp. flag to transmit the frame */
	frbuf->flag = 1;

	/* dequeue packet from IP queue */
	IF_DEQUEUE(&ifp->if_snd, opkt);

	/* remove local buffer */
	if (buffer)
	    free(buffer, M_DEVBUF);
#if NBPFILTER > 0
	/* report the packet to BPF if present and attached */
  	if (ifp->if_bpf != NULL)
	{
	    struct mbuf m0;
	    u_int32_t af = AF_INET;

	    m0.m_next = opkt;
	    m0.m_len = 4;
	    m0.m_data = (char*)&af;
	    /*bpf_mtap(ifp, &m0)*/;
	    bpf_mtap(ifp, opkt);
	}	
#endif

	/* update statistics and free the packet */
	ifp->if_opackets++;
        ifp->if_obytes +=data_length;
	m_freem(opkt);
    } /* while there're packets in the queue */

    /* restore interrupt level and exit */
    splx(s);
    return;
}

/*
 * ============================================================================
 * Setup so that a frame can be transmitted on the occurence of a transmit
 * interrupt.
 */
static void 
setup_for_delayed_transmit (struct ifnet* ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    sdla_t* card = chan->card;
    fr508_flags_t* adptr_flags = card->flags;
    fr_dlci_interface_t* dlci_interface = chan->dlci_int_interface;


    if(chan->transmit_length) 
    {
	log(LOG_ERR, "%s: Big mess in setup_for_del...\n", card->devname);
        return;
    }
    chan->transmit_length = 1;

    dlci_interface->gen_interrupt |= FR_INTR_TXRDY;
    /*dlci_interface->packet_length = len; */
    adptr_flags->imask |= FR_INTR_TXRDY;

    card->u.f.tx_interrupts_pending ++;
}

/* 
 * ============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 * Return 0 if not broadcast/multicast address, otherwise return 1.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp,
                                struct mbuf *m)
{
    uint32_t src_ip_addr = 0;
    uint32_t broadcast_ip_addr = 0;
    struct ifaddr *ifa = TAILQ_FIRST(&ifp->if_addrhead);
    struct sockaddr_in *saddr = NULL;
    fr_channel_t* chan = ifp->if_softc;

    /* read the IP source address from the outgoing packet */
    src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 14);

    /* read the IP broadcast address for the device */
    if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL) 
        broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);

    /* check if the IP Source Address is a Broadcast address */
    if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)) 
    {
        log(LOG_INFO, "%s: Broadcast Source Address silently discarded\n",
							card->devname);
        m_freem(m);
        ++ chan->ifstats.tx_dropped;
        return 1;
     } 

     /* check if the IP Source Address is a Multicast address */
     if ((chan->mc == WANOPT_NO) && (ntohl(src_ip_addr) >= 0xE0000001) && 
		 (ntohl(src_ip_addr) <= 0xFFFFFFFE))
     {
         log(LOG_INFO, "%s: Multicast Source Address silently discarded\n", 
					card->devname);
         m_freem(m);
         ++ chan->ifstats.tx_dropped;
         return 1;
     }

    return 0;
}

/*
 * ============================================================================
 * Reply to UDP Management system.
 * Return nothing.
 */
static int reply_udp( unsigned char *data, unsigned int mbox_len ) 
{
    unsigned short len, udp_length, temp, ip_length;
    unsigned long ip_temp;
    int even_bound = 0;

    fr_udp_pkt_t *fr_udp_pkt = (fr_udp_pkt_t *)data; 

    /* Set length of packet */
    len = /*sizeof(fr_encap_hdr_t)+*/
    	  sizeof(ip_pkt_t)+ 
      	  sizeof(udp_pkt_t)+
          sizeof(wp_mgmt_t)+
          sizeof(cblock_t)+
          mbox_len;

    /* fill in UDP reply */
    fr_udp_pkt->wp_mgmt.request_reply = UDPMGMT_REPLY;

    /* fill in UDP length */
    udp_length = sizeof(udp_pkt_t)+ 
	         sizeof(wp_mgmt_t)+
	         sizeof(cblock_t)+
	         mbox_len; 

    /* put it on an even boundary */
    if ( udp_length & 0x0001 ) 
    {
	udp_length += 1;
	len += 1;
	even_bound = 1;
    }

    temp = (udp_length<<8)|(udp_length>>8);
    fr_udp_pkt->udp_pkt.udp_length = temp;

    /* swap UDP ports */
    temp = fr_udp_pkt->udp_pkt.udp_src_port;
    fr_udp_pkt->udp_pkt.udp_src_port = fr_udp_pkt->udp_pkt.udp_dst_port; 
    fr_udp_pkt->udp_pkt.udp_dst_port = temp;

    /* add UDP pseudo header */
    temp = 0x1100;
    *((unsigned short *)(fr_udp_pkt->data+mbox_len+even_bound)) = temp;	
    temp = (udp_length<<8)|(udp_length>>8);
    *((unsigned short *)(fr_udp_pkt->data+mbox_len+even_bound+2)) = temp;

    /* calculate UDP checksum */
    fr_udp_pkt->udp_pkt.udp_checksum = 0;
    fr_udp_pkt->udp_pkt.udp_checksum = 
		calc_checksum(&data[UDP_OFFSET/*+sizeof(fr_encap_hdr_t)*/],
		udp_length+UDP_OFFSET);

    /* fill in IP length */
    ip_length = udp_length + sizeof(ip_pkt_t);
    temp = (ip_length<<8)|(ip_length>>8);
    fr_udp_pkt->ip_pkt.total_length = temp;

    /* swap IP addresses */
    ip_temp = fr_udp_pkt->ip_pkt.ip_src_address;
    fr_udp_pkt->ip_pkt.ip_src_address = fr_udp_pkt->ip_pkt.ip_dst_address;
    fr_udp_pkt->ip_pkt.ip_dst_address = ip_temp;

    /* fill in IP checksum */
    fr_udp_pkt->ip_pkt.hdr_checksum = 0;
    fr_udp_pkt->ip_pkt.hdr_checksum = 
		calc_checksum(&data[0/*sizeof(fr_encap_hdr_t)*/],
	        sizeof(ip_pkt_t));

    return len;
} /* reply_udp */

unsigned short calc_checksum (char *data, int len)
{
    unsigned short temp; 
    unsigned long sum=0;
    int i;

    for( i = 0; i <len; i+=2 ) 
    {
	bcopy(&data[i], &temp, 2);
	sum += (unsigned long)temp;
    }

    while (sum >> 16) 
    {
	sum = (sum & 0xffffUL) + (sum >> 16);
    }

    temp = (unsigned short)sum;
    temp = ~temp;

    if (temp == 0) 
        temp = 0xffff;

    return temp;
}

/* 
 * ============================================================================
 * If incoming is 0 (outgoing)- if the net numbers is ours make it 0
 * if incoming is 1 - if the net number is 0 make it ours 
 *
 */
#ifdef SUPPORT_IPX
static void 
switch_net_numbers(unsigned char *sendpacket, unsigned long network_number, 
		   unsigned char incoming)
{
    unsigned long pnetwork_number;

    pnetwork_number = (unsigned long)(
		(sendpacket[14] << 24) + (sendpacket[15] << 16) + 
		(sendpacket[16] << 8) + sendpacket[17]);

    if (!incoming) 
    {
	/* If the destination network number is ours, make it 0 */
	if( pnetwork_number == network_number) 
	{
	    sendpacket[14] = sendpacket[15] = sendpacket[16] = 
				 sendpacket[17] = 0x00;
	}
    } 
    else 
    {
	/* If the incoming network is 0, make it ours */
	if( pnetwork_number == 0) 
	{
	    sendpacket[14] = (unsigned char)(network_number >> 24);
	    sendpacket[15] = (unsigned char)((network_number & 
			                      0x00FF0000) >> 16);
	    sendpacket[16] = (unsigned char)((network_number & 
			                      0x0000FF00) >> 8);
	   sendpacket[17] = (unsigned char)(network_number & 
			 		    0x000000FF);
	}
    }

    pnetwork_number = (unsigned long)(
		(sendpacket[26] << 24) + (sendpacket[27] << 16) + 
		(sendpacket[28] << 8) + sendpacket[29]);

    if (!incoming) 
    {
	/* If the source network is ours, make it 0 */
	if (pnetwork_number == network_number) 
	{
	    sendpacket[26] = sendpacket[27] = sendpacket[28] = 
			 sendpacket[29] = 0x00;
	}
    } 
    else 
    {
        /* If the source network is 0, make it ours */
        if (pnetwork_number == 0)
        {
	    sendpacket[26] = (unsigned char)(network_number >> 24);
	    sendpacket[27] = (unsigned char)((network_number & 
			 		      0x00FF0000) >> 16);
	    sendpacket[28] = (unsigned char)((network_number & 
			 		      0x0000FF00) >> 8);
	    sendpacket[29] = (unsigned char)(network_number & 
			 		     0x000000FF);
	}
    }
} /* switch_net_numbers */
#endif

/****** Interrupt Handlers **************************************************/

/* 
 * ============================================================================
 * S508 frame relay interrupt service routine.
 */
static void fr_isr (sdla_t* card)
{
    fr508_flags_t* flags = card->flags;
    int i,err;
    fr_mbox_t* mbox = card->mbox;
    char *ptr = &flags->iflag;

    /* 
     * This flag prevents nesting of interrupts.  See sdla_isr() routine
     * in sdlamain.c. 
     */
    card->in_isr = 1;

    ++card->statistics.isr_entry;

    switch (flags->iflag) 
    {
    case FR_INTR_RXRDY:  /* receive interrupt */
	++card->statistics.isr_rx;
        rx_intr(card);
        break;

    case FR_INTR_TXRDY:  /* transmit interrupt */
	++ card->statistics.isr_tx; 
	tx_intr(card); 
        break;

    case FR_INTR_READY:
	Intr_test_counter++;
	++card->statistics.isr_intr_test;
        break;	

    case FR_INTR_DLC: /* Event interrupt occured */
	mbox->cmd.command = FR_READ_STATUS;
	mbox->cmd.length = 0;
	mbox->cmd.dlci = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
	if (err)
	    fr_event(card, err, mbox);
	break;

    case FR_INTR_TIMER:  /* Timer interrupt */
	timer_intr(card);
	break;

    default:
 	++card->statistics.isr_spurious;
        spur_intr(card);
        log(LOG_ERR, "%s: Interrupt Type 0x%02x!\n", card->devname, flags->iflag);
        log(LOG_ERR, "%s: ID Bytes = ",card->devname);
	for(i = 0; i < 8; i ++)
	    log(LOG_ERR, "0x%02x ", *(ptr + 0x28 + i));
   	log(LOG_ERR, "\n");	
	break;
    }

    card->in_isr = 0;
    flags->iflag = 0;
}


/* 
 * ============================================================================
 * Receive interrupt handler.
 * When a receive interrupt occurs do the following:
 *	1- Find the structure for the dlci that the interrupt occured on
 *      2- If it doesn't exist then print appropriate msg and goto step 8. 
 * 	3- If it exist then copy data to a mbuf.
 * 	4- If mbuf contains Sangoma UDP data then process them
 *	5- If mbuf contains IPXWAN data then send IPXWAN reply packets
 *	6- If mbuf contains Inverse Arp data then send Inv Arp replies
 *	7- If mbuf contains any other data then decapsulate the packet and
 *	   send it to the stack.
 * 	8- Release the receive element and update receive pointers on the board 
 */
static void rx_intr (sdla_t* card)
{
    fr_rx_buf_ctl_t* frbuf = card->rxmb;
    fr508_flags_t* flags = card->flags;
    fr_channel_t* chan = NULL;
    struct mbuf* m = NULL, *ipkt = NULL;
    struct ifnet* ifp = NULL;
    int ip_pkt = 0; /* default value */
    unsigned dlci, len, offs;
    int i, udp_type, s = 0;	
    char *ptr = &flags->iflag;

    if (frbuf->flag != 0x01) 
    {
	log(LOG_ERR, "%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
			card->devname, (unsigned)frbuf, frbuf->flag);
      
	log(LOG_ERR, "%s: ID Bytes = ",card->devname);
 	for(i = 0; i < 8; i ++)
	    log(LOG_ERR, "0x%02x ", *(ptr + 0x28 + i));
	log(LOG_ERR, "\n");

	++card->statistics.rx_intr_corrupt_rx_bfr;
	return;
    }

    len  = frbuf->length;
    dlci = frbuf->dlci;
    offs = frbuf->offset;
    /* Find network interface for this packet */
    ifp = find_channel(card, dlci);

    if (ifp == NULL) 
    {
	/* unconfigured DLCI, so discard packet */
   	log(LOG_ERR, "%s: received data on unconfigured DLCI %d!\n",
                                               card->devname, dlci);
	++card->statistics.rx_intr_on_orphaned_DLCI; 
    } 
    else 
    {
        char *data = NULL, *data_ptr = NULL;
        unsigned int api_hdr_len = 0;
        arphdr_1490_t hdr;
	chan = ifp->if_softc;

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) 
	{
	    log(LOG_ERR, "%s - received data frame on DLCI %d however device %s%d
				is not UP and RUNNING.\n", card->devname, dlci, 
				ifp->if_name, ifp->if_unit);
	    chan->drvstats_rx_intr.rx_intr_dev_not_started ++;
	    ++chan->ifstats.rx_dropped;
	    goto rx_done;
	}

	if (chan->usedby == API)
	    api_hdr_len = sizeof(api_rx_hdr_t);
	
	/* Ok, let's see what kind of packet have we received - for that we will
	 * read-ahead the RFC1490 header. As of now we've got 2 options - its   
	 * either IP over Frame Relay or InARP packet. All the rest is discarded
	 */
	sdla_peek(&card->hw, offs, &hdr, sizeof(arphdr_1490_t));

        /* check the the packet type ... */
	if ((hdr.control == 0x03) && (hdr.pad == 0xcc)) 
	{
	    /* IP over Frmae Relay encapsulation ... skip first two bytes and
	     * go on with mbufs, UDP stuff and so on ...			
	     */
	    len -= 2; 
	    offs += 2;
	    ip_pkt = 1;
	}

	/*
	 * If current chan not running API, driver allocate buffer for the
	 * received data, otherwise the length will be include the api header.
	 */
	data = malloc(len + api_hdr_len + 1, M_DEVBUF, M_NOWAIT);
	if (data == NULL)
	{
	    log(LOG_ERR, "%s%d: Cannot allocate data buffer!\n", 
						ifp->if_name, ifp->if_unit);
	    /* FIXME: update error statistics */
	    return; 
	}
	bzero(data, len + api_hdr_len + 1);

	/*
	 * Now, we can start copying data into data area. 
	 * We len bytes. Note that since the S508s receive buffers are  
	 * cyclic we might have to copy the packet in two pieces. 
	 */
	if ((offs + len) <= card->u.f.rx_top + 1) 
	{
	    sdla_peek(&card->hw, offs, &data[api_hdr_len], len);
	    offs += len;
	}
	else
        {
	    unsigned int first = card->u.f.rx_top - offs + 1;
	    unsigned int second = len - first;
	    sdla_peek(&card->hw, offs, &data[api_hdr_len], first);
	    sdla_peek(&card->hw, card->u.f.rx_base, 
					&data[api_hdr_len + first], second); 
	    offs = card->u.f.rx_base + second;
	}

	if (chan->usedby == API)
	{ /* Update API header */
            ((api_rx_hdr_t*)data)->attr = frbuf->attr;
    	    ((api_rx_hdr_t*)data)->time_stamp = frbuf->tmstamp;
	}
	data_ptr = data;
	len += api_hdr_len;

	/* we get here with IP packets only ... get the first ( header ) mbuf */
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL) 
	{
	    log(LOG_ERR, "%s: no socket buffers available!\n", card->devname);
   	    chan->drvstats_rx_intr.rx_intr_no_socket ++;
	    ifp->if_iqdrops++;
	    return;
	}
        else
	{

	/* m will be running through the chain. 
         * ipkt always point to the header.
	 */
	ipkt = m; 
	ipkt->m_pkthdr.rcvif = ifp;	/* our interface */
	ipkt->m_pkthdr.len   = len;	/* length of the data */
	ipkt->m_len	     = MHLEN;	/* ... well, at leasts as of now */

	/* let's start filling in the chain */
	while( len > 0 ) 
	{
	    /* check if we can use cluster instead ... */
	    if (len > MINCLSIZE) 
	    {
	        /* all right - get the cluster */
	        MCLGET(m, M_DONTWAIT);
	        if (m->m_flags & M_EXT) 
	        {
	            /* all right - it worked - set the new length */
	            m->m_len = MCLBYTES;
	        } 
		else 
		{
		    /* it failed - well, screw it - we'll try again later */
		    m_freem(ipkt);
		    log(LOG_ERR, "%s%d: Failed to allocate cluster !!!\n",
						ifp->if_name, ifp->if_unit);
		    ifp->if_iqdrops++;
	 	    return;
		}
	    }
	
            /* 
	     * ok, at this point we've got mbuf pointed by 'm' which is either 
	     * a regular one, or a cluster depending on how much data do we have
	     * left. The beauty is that from here we dont care what kind of mbuf
	     * it is since all the macros work correctly on both.
	     */
	    m->m_len = min(m->m_len, len);
   
	    bcopy(data_ptr, mtod(m, caddr_t), m->m_len);
	    data_ptr += m->m_len;
	    len      -= m->m_len;

	    /* done with this mbuf - allocate the next one if needed */
	    if (len > 0) 
	    {
		MGET(m->m_next, M_DONTWAIT, MT_DATA);
		if (m->m_next == NULL) 
		{
		    /* it failed - well, screw it - we'll try again later */
		    m_freem(ipkt);
		    log(LOG_ERR, "%s%d: Failed to allocate mbuf !!!\n", 
						ifp->if_name, ifp->if_unit);
		    ifp->if_iqdrops++;	
		    return;
		}
		/* set m to point to the new one, and set up initial length */
		m = m->m_next;
		m->m_len = MLEN;
	    }
	} /* while ... */

	if (data)
	    free(data, M_DEVBUF);

	udp_type = udp_pkt_type( ipkt, card );
        if (udp_type != UDP_INVALID_TYPE) 
	{
	    if (store_udp_mgmt_pkt(udp_type,UDP_PKT_FRM_NETWORK, card, 
				  ipkt, dlci)) 
	    {
	        flags->imask |= FR_INTR_TIMER;
                if (udp_type == UDP_FPIPE_TYPE)
		    chan->drvstats_rx_intr.rx_intr_PIPE_request ++;
	    }			
	}
	else
	{
	    int pass_to_stack = 1;
 
            if (chan->usedby != API) 
	    {
#ifdef SUPPORT_IPX 
	        if (handle_IPXWAN(mtod(ipkt, caddr_t), chan->name,	
			          chan->enable_IPX, chan->network_number)) 
		{
		    if (chan->enable_IPX) 
		       fr_send(card, dlci, 0, ipkt->m_len, 
			       mtod(ipkt, caddr_t), 1);
		    m_freem(m);
		    pass_to_stack = 0;
		} 
		else
#endif
		    if (is_arp(mtod(ipkt, caddr_t))) 
		    {
	                if (process_ARP((arphdr_1490_t *)mtod(ipkt, caddr_t), 
					card, ifp)) 
			    log(LOG_ERR, "%s: Error processing ARP Packet.\n", 
							card->devname);
		        m_freem(ipkt);
		        pass_to_stack = 0;
		    } 
		    else 
			if (!ip_pkt) 
			{
			    log(LOG_INFO, "%s: Non IETF packet discarded\n", 
							card->devname);		
			    m_freem(ipkt);
		            pass_to_stack = 0;
			}
            }
	    if (pass_to_stack)
	    {
#if NBPFILTER > 0
		/* report the packet to BPF for a dlci_specific device */
		if (ifp->if_bpf != NULL)
		{
		    struct mbuf m0;
		    u_int32_t af = AF_INET;
		    m0.m_next = ipkt;
		    m0.m_len = 4;
		    m0.m_data = (char*)&af;
		    /*bpf_mtap(ifp, &m0);*/
		    bpf_mtap(ifp, ipkt);
		}
#endif
		/* 
		 * Schedule IP input routine. Note that it makes 
		 * much more sence to check first if the protocol 
		 * queue is full - and then to schedule the input 
		 * routine, but i took a look at kernel sources
		 * for ether_input() 
		 * [ /usr/src/sys/net/if_ethersubr.c ] and they
		 * do it in this order - so i figured they know 
		 * better.
		 */
		s = splimp();
		schednetisr(NETISR_IP);
		if (IF_QFULL(&ipintrq)) 
		{
		    /* oh, no - IP queue is full - well - we'll 
		     * try again later 
		     */ 
		    log(LOG_ERR, "%s%d: IP queue is full, drop packet !!!\n",
							ifp->if_name, ifp->if_unit);
		    IF_DROP(&ipintrq);
		    m_freem(ipkt);
		    splx(s);
		    return;
		}
	  
		/* ok enqueue the packet */
		IF_ENQUEUE(&ipintrq, ipkt);
		splx(s);

                /* update interface statistics */
	        ifp->if_ipackets++;
		chan->drvstats_rx_intr.rx_intr_bfr_passed_to_stack ++;
		chan->ifstats.rx_packets ++;
                card->wandev.stats.rx_packets ++;
	    }
        }
	}
    }

rx_done:

    /* Release buffer element and calculate a pointer to the next one */ 
    frbuf->flag = 0;
    card->rxmb = ++frbuf;
    if ((void*)frbuf > card->u.f.rxmb_last)
	card->rxmb = card->u.f.rxmb_base;
}

/* 
 * ============================================================================
 * Transmit interrupt handler.
 */
static void tx_intr(sdla_t *card)
{
    fr508_flags_t* flags = card->flags;
    fr_tx_buf_ctl_t* bctl = NULL;
    struct ifnet* ifp = NULL;
    fr_channel_t* chan = NULL;

    if(card->hw.type == SDLA_S514)
	bctl = (void*)(flags->tse_offs + (unsigned long)card->hw.dpmbase);
    else
	bctl = (void*)(flags->tse_offs - FR_MB_VECTOR + 
		       (unsigned long)card->hw.dpmbase);

    /* Find the structure and make it unbusy */
    ifp = find_channel(card, flags->dlci);

    if (ifp == NULL)
    {
	/* no suppose to happened - but just to check */
	log(LOG_ERR, "Tx interrupt for DLCI %d which is not configured.\n", 
						flags->dlci);
	return;
    }
    chan = ifp->if_softc;


    if(!chan->transmit_length) 
    {
	log(LOG_ERR, "%s: tx int error - transmit length zero\n", 
						card->wandev.name);
        return;
    }
   
    {
	int transmit = 1, s = 0;
	struct mbuf *opkt = NULL, *m = NULL;
	unsigned long data_addr;
	unsigned short data_length = 0;

        s = splimp();

	/* check if we've got anything to tranmit at all */
	if ((opkt = ifp->if_snd.ifq_head) == NULL) 
	{
	    /* not supposed to happen - but who knows */
	    log(LOG_ERR, "%s%d: Nothing to transmit on DLCI %d.\n", 
				ifp->if_name, ifp->if_unit, flags->dlci);
	    /* set up the no-transmit flag */
	    transmit = 0;
	} 
	else 
	{
	    /* ok, now we actually gonna copy the data onto the board and   */
	    /* prepare the rest of tx status element (length only actually) */

	    /* dequeue packet from IP queue */
	    IF_DEQUEUE(&ifp->if_snd, opkt);

	    /* initialize some of our stuff */
	    data_addr = bctl->offset;
	    data_length	= 0;

	    /* Add the encapsulation bytes for IP  */
	    {
	        unsigned short ip_over_fr_header = FR_IP_HEADER;

	        sdla_poke(&card->hw, data_addr, &ip_over_fr_header, 2 );
	        data_addr += 2; 
	        data_length += 2;
	    }

	    /* go through mbuf chain and copy data to the board */
	    for( m = opkt; m != NULL; m = m->m_next ) 
	    {
		/* copy data to the board */
	        sdla_poke(&card->hw, data_addr, mtod(m, caddr_t), m->m_len );
		data_length += m->m_len;
		data_addr += m->m_len;
	    }
	}

	/* see if we need to transmit packet - if yes - set the flag to 0x80 
         * if not - set the flag to 0xa0				  
	 */
	if (transmit) 
	{
	    /* 
	     * set up the rest of tx element and transmit the frame ....
	     */
	    bctl->dlci   = flags->dlci;
            bctl->length = data_length;
	    bctl->attr   = 0;
	    bctl->flag   = 0xC0;

#if NBPFILTER > 0
	    /* report the packet to BPF if present and attached */
	    if (ifp->if_bpf != NULL) 
	    {	/* BF-0002 */
	        struct mbuf m0;
	        u_int32_t af = AF_INET;

	        m0.m_next = opkt;
	        m0.m_len = 4;
	        m0.m_data = (char *)&af; 

	        /*bpf_mtap(ifp, &m0);*/
	        bpf_mtap(ifp, opkt);
	    }
#endif

	    /* update statistics and free the packet */
	    ifp->if_opackets++;
	    ifp->if_obytes +=data_length;
	    m_freem(opkt);

	    /* reset the ACTIVE flag */
	    ifp->if_flags &= ~IFF_OACTIVE;

	    /* Update tick_counter */
	    chan->tick_counter = SYSTEM_TICK;

	    ++chan->ifstats.tx_packets;
	    ++card->wandev.stats.tx_packets;

       	    chan->transmit_length = 0;

            /* if any other interfaces have transmit interrupts pending, 
	     * do not disable the global transmit interrupt 
	     */
	    if (!(-- card->u.f.tx_interrupts_pending))
	        flags->imask &= ~FR_INTR_TXRDY;
	}
	else
	{
	    bctl->flag   = 0xA0;
	
	}
        splx(s);
    }
}


/*
 * ============================================================================
 * Timer interrupt handler.
 * FIXME: update comments as we modify the code
 * The timer interrupt is used for three purposes:
 *    1) Processing udp calls from 'fpipemon'.
 *    2) Processing update calls from /proc file system
 *    2) Reading board-level statistics for updating the proc file system.
 *    3) Sending inverse ARP request packets.
 */
static void timer_intr(sdla_t *card)
{
    fr508_flags_t* flags = card->flags;

    if(card->u.f.timer_int_enabled & TMR_INT_ENABLED_UDP) 
    {
	if ((card->u.f.udp_type == UDP_FPIPE_TYPE) && 
            (process_udp_mgmt_pkt(card)))
            card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
    }

    if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_UPDATE) 
    {
	fr_get_err_stats(card);
	fr_get_stats(card);
	card->u.f.update_comms_stats = 0;
	card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
    }

    if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_INARP) 
    {
    	fr_channel_t *chan = NULL;
        struct ifnet_list *ifnode = NULL;

      	for (ifnode = card->wandev.if_head; ifnode && ifnode->ifp; 
	 	ifnode = ifnode->pnext)
	{
	    chan = ifnode->ifp->if_softc;

	    if (chan->inarp == INARP_REQUEST && chan->state == WAN_CONNECTED &&
	        chan->dlci_configured == DLCI_CONFIGURED)
	        send_inarp_request(card, ifnode->ifp);
	}
	card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_INARP;
    }

    if (!card->u.f.timer_int_enabled)
	flags->imask &= ~FR_INTR_TIMER;
}


/*============================================================================
 * Spurious interrupt handler.
 * o print a warning
 * o 
 */
static void spur_intr (sdla_t* card)
{
    log(LOG_ERR, "%s: spurious interrupt!\n", card->devname);
}


/*===========================================================================
 *  Return 0 for non-IPXWAN packet
 *         1 for IPXWAN packet or IPX is not enabled!
 *  FIXME: Use a IPX structure here not offsets
 */
#ifdef SUPPORT_IPX
static int 
handle_IPXWAN(unsigned char *sendpacket, char *devname, 
	      unsigned char enable_IPX, unsigned long network_number)
{
    int i;

    if (sendpacket[1] == 0x00 &&
        sendpacket[2] == 0x80 &&
        sendpacket[6] == 0x81 &&
        sendpacket[7] == 0x37) 
    { 
	/* It's an IPX packet */
	if (!enable_IPX)
	    /* Return 1 so we don't pass it up the stack. */
	    return 1;
    } 
    else /* It's not IPX so return and pass it up the stack. */
	return 0;

    if (sendpacket[24] == 0x90 && sendpacket[25] == 0x04)
    {
	/* It's IPXWAN */
	if (sendpacket[10] == 0x02 && sendpacket[42] == 0x00)
	{
	    /* It's a timer request packet */
	    log(LOG_INFO, "%s: Received IPXWAN Timer Request packet\n",devname);

	    /* Go through the routing options and answer no to every
	     * option except Unnumbered RIP/SAP
	     */
	    for(i = 49; sendpacket[i] == 0x00; i += 5)
	    {
		/* 0x02 is the option for Unnumbered RIP/SAP */
		if (sendpacket[i + 4] != 0x02)
		    sendpacket[i + 1] = 0;
	    }

	    /* Skip over the extended Node ID option */
	    if (sendpacket[i] == 0x04)
		i += 8;

	    /* We also want to turn off all header compression opt.
	     */
	    for(; sendpacket[i] == 0x80 ;)
	    {
	        sendpacket[i + 1] = 0;
	        i += (sendpacket[i + 2] << 8) + (sendpacket[i + 3]) + 4;
	    }

	    /* Set the packet type to timer response */
	    sendpacket[42] = 0x01;

	    log(LOG_INFO, "%s: Sending IPXWAN Timer Response\n",devname);
        }
        else 
            if (sendpacket[42] == 0x02)
   	    {
	        /* This is an information request packet */
                log(LOG_INFO, "%s: Received IPXWAN Information Request packet\n",
								devname);

		/* Set the packet type to information response */
		sendpacket[42] = 0x03;

		/* Set the router name */
		sendpacket[59] = 'F';
		sendpacket[60] = 'P';
		sendpacket[61] = 'I';
		sendpacket[62] = 'P';
		sendpacket[63] = 'E';
		sendpacket[64] = '-';
		sendpacket[65] = CVHexToAscii(network_number >> 28);
	        sendpacket[66] = 
			CVHexToAscii((network_number & 0x0F000000) >> 24);
	        sendpacket[67] = 
			CVHexToAscii((network_number & 0x00F00000) >> 20);
	        sendpacket[68] = 
			CVHexToAscii((network_number & 0x000F0000) >> 16);
		sendpacket[69] = 
			CVHexToAscii((network_number & 0x0000F000) >> 12);
	        sendpacket[70] = 
			CVHexToAscii((network_number & 0x00000F00) >> 8);
		sendpacket[71] = 
			CVHexToAscii((network_number & 0x000000F0) >> 4);
	 	sendpacket[72] = 
			CVHexToAscii(network_number & 0x0000000F);
		for(i = 73; i < 107; i+= 1)
		    sendpacket[i] = 0;

		log(LOG_INFO, "%s: Sending IPXWAN Information Response packet\n",
								devname);
	    }
	    else
	    {
		log(LOG_ERR, "%s: Unknown IPXWAN packet!\n",devname);
		return 0;
	    }

        /* Set the WNodeID to our network address */
        sendpacket[43] = (unsigned char)(network_number >> 24);
	sendpacket[44] = (unsigned char)((network_number & 0x00FF0000) >> 16);
        sendpacket[45] = (unsigned char)((network_number & 0x0000FF00) >> 8);
	sendpacket[46] = (unsigned char)(network_number & 0x000000FF);
	return 1;
    }

    /* If we get here, its an IPX-data packet so it'll get passed up the 
     * stack.
     * switch the network numbers 
     */
    switch_net_numbers(sendpacket, network_number ,1);
    return 0;
}
#endif

/*
 * ============================================================================
 * Process Route.
 * This routine is called as a polling routine to dynamically add/delete routes
 * negotiated by inverse ARP.  It is in this "task" because we don't want routes
 * to be added while in interrupt context.
 */
static void process_route (void *arg)
{
    fr_channel_t *chan = NULL;
    sdla_t *card = NULL;
    struct ifnet *ifp = (struct ifnet*)arg;
    int err = 0;
    struct ifaddr	     *ifa = TAILQ_FIRST(&ifp->if_addrhead);
    struct sockaddr_in *saddr = NULL;		/* addr. struct */

    chan = ifp->if_softc;
    card = chan->card;
    /* i have no idea why but it seems that the local and remote IP's 
     * are stored not inside ifa but inside ifa->ifa_next. We'd like  
     * to make sure we are not writing to NULL though.		
     */
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) == NULL) 
    {
        /* that shouldnt happen ... */
        return;
    }

    /* delete old routing entry */
    err = rtrequest((int)RTM_DELETE, 		/* action */
		ifa->ifa_dstaddr,		/* destination */
		ifa->ifa_addr,		    /* gateway */
		ifa->ifa_netmask,		/* mask */
		RTF_HOST,			    /* flag */
		(struct rtentry **)0);

    switch(chan->route_flag) 
    {
    case ADD_ROUTE:
        /* set up new destination address */
        saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
        saddr->sin_addr.s_addr = chan->dst_addr;

        /* add new routing entry */
        err = rtrequest((int)RTM_ADD, 	  /* action */
		ifa->ifa_dstaddr,	  /* destination */
		ifa->ifa_addr,		  /* gateway */
		ifa->ifa_netmask,	  /* mask */
		RTF_HOST,			  /* flag */
		(struct rtentry **)0);
	if (err)
	{
	    log(LOG_ERR, "%s: Adding of route failed (%d)\n",
				card->devname, err);
	    log(LOG_ERR, "%s: Address %s\n", card->devname, 
				in_ntoa(saddr->sin_addr.s_addr));
	}
	else
            chan->route_flag = ROUTE_ADDED;
        break;

    case REMOVE_ROUTE:
	if (err)
	{
	    log(LOG_ERR, "%s: Deleting of route failed (%d)\n",
						card->devname, err);
	    log(LOG_ERR, "%s: Address %s\n", card->devname, 
		in_ntoa(((struct sockaddr_in*)ifa->ifa_dstaddr)->sin_addr.s_addr));
	}
	else
            chan->route_flag = NO_ROUTE;
    	break;
    }

    card->poll = NULL;
}


/****** Frame Relay Firmware-Specific Functions *****************************/

/* 
 * ============================================================================
 * Read firmware code version.
 * o fill string str with firmware version info. 
 */
static int fr_read_version (sdla_t* card, char* str)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Read firmware version info ...\n", card->devname));
    do
    {
	mbox->cmd.command = FR_READ_CODE_VERSION;
	mbox->cmd.length = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));
	
    if (!err && str) 
    {
	int len = mbox->cmd.length;
	bcopy(mbox->data, str, len);
        str[len] = '\0';
    }
    return err;
}


/* 
 * ============================================================================
 * Set global configuration.
 */
static int fr_configure (sdla_t* card, fr_conf_t *conf)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int dlci_num = card->u.f.dlci_num;
    int err, i;

    dlog((LOG_DEBUG, "%s: Set global configuration ...\n", card->devname));
    do
    {
	bcopy(conf, mbox->data, sizeof(fr_conf_t));

	if (dlci_num)
	    for (i = 0; i < dlci_num; ++i)
		((fr_conf_t*)mbox->data)->dlci[i] = card->u.f.node_dlci[i]; 
	
	mbox->cmd.command = FR_SET_CONFIG;
	mbox->cmd.length = sizeof(fr_conf_t) + dlci_num * sizeof(short);

	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;

    } while (err && retry-- && fr_event(card, err, mbox));

    return err;
}

/* 
 * ============================================================================
 * Set DLCI configuration.
 */
static int fr_dlci_configure (sdla_t* card, fr_dlc_conf_t *conf, unsigned dlci)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Configure DLCI %d...\n", card->devname, dlci));
    do
    {
	bcopy(conf, mbox->data, sizeof(fr_dlc_conf_t));
	mbox->cmd.dlci = (unsigned short) dlci; 
	mbox->cmd.command = FR_SET_CONFIG;
	mbox->cmd.length = sizeof(fr_dlc_conf_t);
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry--);
	
    return err;
}


/* 
 * ============================================================================
 * Set interrupt mode.
 */
static int 
fr_set_intr_mode (sdla_t* card, unsigned mode, unsigned mtu, 
				  unsigned short timeout)
{
    fr_mbox_t* mbox = card->mbox;
    fr508_intr_ctl_t* ictl = (void*)mbox->data;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Set interrupt mode...\n", card->devname));
    do
    {
	bzero(ictl, sizeof(fr508_intr_ctl_t));
	ictl->mode   = mode;
	ictl->tx_len = mtu;
	ictl->irq    = card->hw.irq;

	/* indicate timeout on timer */
	if (mode & 0x20) ictl->timeout = timeout; 

	mbox->cmd.length = sizeof(fr508_intr_ctl_t);
	mbox->cmd.command = FR_SET_INTR_MODE;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;

    } while (err && retry-- && fr_event(card, err, mbox));
	
    return err;
}

/* 
 * ============================================================================
 * Enable communications.
 */
static int fr_comm_enable (sdla_t* card)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Enable communications...\n", card->devname));
    do
    {
	mbox->cmd.command = FR_COMM_ENABLE;
	mbox->cmd.length = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));
	
    return err;
}

/*
 * ============================================================================
 * Disable communications. 
 */
static int fr_comm_disable (sdla_t* card)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Disable communications...\n", card->devname));
    do
    {
	mbox->cmd.command = FR_COMM_DISABLE;
	mbox->cmd.length = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    retry = MAX_CMD_RETRY;

    do 
    {
	mbox->cmd.command = FR_SET_MODEM_STATUS;
	mbox->cmd.length = 1;
	mbox->data[0] = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));
	
    dlog((LOG_DEBUG, "%s: fr_comm_disable end()\n", card->devname));
    return err;
}

/*
 * ============================================================================
 * Get communications error statistics. 
 */
static int fr_get_err_stats (sdla_t* card)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Get error statistics...\n", card->devname));
    do
    {
	mbox->cmd.command = FR_READ_ERROR_STATS;
	mbox->cmd.length = 0;
	mbox->cmd.dlci = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    if (!err) 
    {
	fr_comm_stat_t* stats = (void*)mbox->data;
	card->wandev.stats.rx_over_errors    = stats->rx_overruns;
	card->wandev.stats.rx_crc_errors     = stats->rx_bad_crc;
	card->wandev.stats.rx_missed_errors  = stats->rx_aborts;
	card->wandev.stats.rx_length_errors  = stats->rx_too_long;
	card->wandev.stats.tx_aborted_errors = stats->tx_aborts;
    }

    return err;
}

/* 
 * ============================================================================
 * Get statistics. 
 */
static int fr_get_stats (sdla_t* card)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Get statistics...\n", card->devname));
    do
    {
	mbox->cmd.command = FR_READ_STATISTICS;
	mbox->cmd.length = 0;
	mbox->cmd.dlci = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    if (!err) 
    {
	fr_link_stat_t* stats = (void*)mbox->data;
	card->wandev.stats.rx_frame_errors = stats->rx_bad_format;
	card->wandev.stats.rx_dropped =
				stats->rx_dropped + stats->rx_dropped2;
    }

    return err;
}

/*
 * ============================================================================
 * Add DLCI(s) (Access Node only!).
 * This routine will perform the ADD_DLCIs command for the specified DLCI.
 */
static int fr_add_dlci (sdla_t* card, int dlci)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Add new DLCI %d...\n", card->devname, dlci));
    do
    {
	unsigned short* dlci_list = (void*)mbox->data;

	mbox->cmd.length  = sizeof(short);
	dlci_list[0] = dlci;
	mbox->cmd.command = FR_ADD_DLCI;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    return err;
}

/* 
 * ===========================================================================
 * Activate DLCI(s) (Access Node only!). 
 * This routine will perform the ACTIVATE_DLCIs command with a DLCI number. 
 */
static int fr_activate_dlci (sdla_t* card, int dlci)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Active DLCI %d...\n", card->devname, dlci));
    do
    {
	unsigned short* dlci_list = (void*)mbox->data;

	mbox->cmd.length  = sizeof(short);
	dlci_list[0] = dlci;
	mbox->cmd.command = FR_ACTIVATE_DLCI;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    return err;
}

/* 
 * ============================================================================
 * Delete DLCI(s) (Access Node only!). 
 * This routine will perform the DELETE_DLCIs command with a DLCI number. 
 */
static int fr_delete_dlci (sdla_t* card, int dlci)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Delete DLCI %d...\n", card->devname, dlci));
    do
    {
	unsigned short* dlci_list = (void*)mbox->data;

	mbox->cmd.length  = sizeof(short);
	dlci_list[0] = dlci;
	mbox->cmd.command = FR_DELETE_DLCI;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));
	
    return err;
}


/*
 * ============================================================================
 * Issue in-channel signalling frame. 
 */
static int fr_issue_isf (sdla_t* card, int isf)
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    dlog((LOG_DEBUG, "%s: Issue signalling frame...\n", card->devname));
    do
    {
	mbox->data[0] = isf;
	mbox->cmd.length  = 1;
	mbox->cmd.command = FR_ISSUE_IS_FRAME;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));
	
    return err;
}


/*
 * ===========================================================================
 * Send data to the card... 
 */
static int fr_write(sdla_t *card, int dlci, unsigned char attr, int len)
{
    fr_mbox_t* mbox = (fr_mbox_t*)((unsigned char*)card->mbox + 0x800);
    int retry = MAX_CMD_RETRY;
    int err;

    do
    {
	mbox->cmd.dlci    = dlci;
	mbox->cmd.attr    = attr;
	mbox->cmd.length  = len;
	mbox->cmd.command = FR_WRITE;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    return err;   
}


/* 
 * ============================================================================
 * Send a frame on a selected DLCI.  
 */
static int 
fr_send (sdla_t* card, int dlci, unsigned char attr, int len, void *buf, int snd_encp)
{
    fr_mbox_t* mbox = (fr_mbox_t*)((unsigned char*)card->mbox + 0x800);
    int err;
    int s = splimp();

    if (snd_encp)
    	err = fr_write(card, dlci, attr, len+2);
    else
	err = fr_write(card, dlci, attr, len);
    if (!err) 
    {
	fr_tx_buf_ctl_t *frbuf = NULL;
	unsigned long data_addr = 0l;

	if(card->hw.type == SDLA_S514)
	    frbuf = (void*)(*(unsigned long*)mbox->data + 
			    (unsigned long)card->hw.dpmbase);
	else
	    frbuf = (void*)(*(unsigned long*)mbox->data - FR_MB_VECTOR + 
			    (unsigned long)card->hw.dpmbase);
	data_addr = frbuf->offset;
	/*Add encapsulation header if snd_encp flag set to 1 !!! */
	if (snd_encp)
	{
	    unsigned short ip_over_fr_header = FR_IP_HEADER;
	    sdla_poke(&card->hw, data_addr, &ip_over_fr_header, 2);
	    data_addr += 2;
	}
	sdla_poke(&card->hw, data_addr, buf, len);
	frbuf->flag = 0x01;
    }

    splx(s);
    return err;
}

/****** Firmware Asynchronous Event Handlers ********************************/

/*
 * ============================================================================
 * Main asyncronous event/error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_event (sdla_t *card, int event, fr_mbox_t* mbox)
{
    fr508_flags_t* flags = card->flags;
    int i;
    char *ptr = &flags->iflag;

    dlog((LOG_DEBUG, "%s: Asyncronous event/error (0x%x)\n", card->devname, event)); 
    switch (event) 
    {
    case FRRES_MODEM_FAILURE:
        return fr_modem_failure(card, mbox);

    case FRRES_CHANNEL_DOWN:
    {
        struct ifnet_list *ifnode = NULL;

        /* Remove all routes from associated DLCI's */
        for (ifnode = card->wandev.if_head; ifnode && ifnode->ifp; 
             ifnode = ifnode->pnext) 
        {
	    struct ifnet *ifp = ifnode->ifp;
            fr_channel_t *chan = ifp->if_softc;
            if (chan->route_flag == ROUTE_ADDED) 
            {
                chan->route_flag = REMOVE_ROUTE;
		card->poll = &process_route;
		ifp->if_timer = 1;
	    }

            if (chan->inarp == INARP_CONFIGURED) 
	    {
		chan->inarp = INARP_REQUEST;
	    }
	}
	wanpipe_set_state(card, WAN_DISCONNECTED);
	return 1;
    }

    case FRRES_CHANNEL_UP:
    {
	struct ifnet_list *ifnode = NULL;
	int num_requests = 0;

	/* Remove all routes from associated DLCI's */
	for (ifnode = card->wandev.if_head; ifnode && ifnode->ifp; 
 	     ifnode = ifnode->pnext) 
	{
	    struct ifnet *ifp = ifnode->ifp;
	    fr_channel_t *chan = ifp->if_softc;	

	    if( chan->inarp == INARP_REQUEST )
	    {
		num_requests++;
		chan->inarp_tick = SYSTEM_TICK;
	    }
	}

	/* Allow timer interrupts */
	if (num_requests)
	{
	    flags->imask |= FR_INTR_TIMER;
	}
	wanpipe_set_state(card, WAN_CONNECTED);
	return 1;
    }

    case FRRES_DLCI_CHANGE:
	return fr_dlci_change(card, mbox);

    case FRRES_DLCI_MISMATCH:
	return 1;

    case CMD_TIMEOUT:
	log(LOG_ERR, "%s: command 0x%02x timed out!\n", 
				card->devname, mbox->cmd.command);
	log(LOG_INFO, "%s: ID Bytes = ",card->devname);
   	for(i = 0; i < 8; i ++)
	    log(LOG_INFO, "0x%02x ", *(ptr + 0x18 + i));
    	log(LOG_INFO, "\n");
	break;

    case FRRES_DLCI_INACTIVE:
	break;
 
    case FRRES_CIR_OVERFLOW:
	break;

    case FRRES_BUFFER_OVERFLOW:
	break;
 
    default:
	log(LOG_INFO, "%s: command 0x%02x returned 0x%02x!\n",
			card->devname, mbox->cmd.command, event);
    } 

    return 0;
}

/* 
 * ============================================================================
 * Handle modem error.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_modem_failure (sdla_t *card, fr_mbox_t* mbox)
{
    log(LOG_INFO, "%s: physical link down! (modem error 0x%02x)\n",
			card->devname, mbox->data[0]);

    switch (mbox->cmd.command)
    {
	case FR_WRITE:
	case FR_READ:
	    return 0;
    }
    return 1;
}

/* 
 * ============================================================================
 * Handle DLCI status change.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_dlci_change (sdla_t *card, fr_mbox_t* mbox)
{
    dlci_status_t* status = (void*)mbox->data;
    int cnt = mbox->cmd.length / sizeof(dlci_status_t);
    fr_channel_t *chan = NULL;
    struct ifnet_list *ifnode = NULL;
    fr508_flags_t *flags = card->flags;
    int num_requests = 0;

    for (; cnt; --cnt, ++status) 
    {
	unsigned short dlci= status->dlci;
	struct ifnet* ifp = find_channel(card, dlci);
	
	if (ifp == NULL)
	{
	    log(LOG_INFO, "%s: CPE contains unconfigured DLCI= %d\n", 
	     					card->devname, dlci); 	
            log(LOG_INFO, "%s: unconfigured DLCI %d reported by network\n", 
						card->devname, dlci);
 	}
	else
	{
	    chan = ifp->if_softc;
	    if (status->state == FR_LINK_INOPER) 
	    {
		log(LOG_INFO, "%s: DLCI %u is inactive!\n", card->devname, dlci);

	        if (chan->start) 
	            set_chan_state(ifp, WAN_DISCONNECTED); 
	    }
	
            if (status->state & FR_DLCI_DELETED) 
	    {
		log(LOG_INFO, "%s: DLCI %u has been deleted!\n", 
						card->devname, dlci);
		if (chan->start)
		{
		    if (chan->route_flag == ROUTE_ADDED) 
		    {
			chan->route_flag = REMOVE_ROUTE;
			card->poll = &process_route;
			ifp->if_timer = 1;
		    }

		    if (chan->inarp == INARP_CONFIGURED) 
			chan->inarp = INARP_REQUEST;

		    set_chan_state(ifp, WAN_DISCONNECTED);
		}
	    } 
	    else 
        	if (status->state & FR_DLCI_ACTIVE) 
		{
		    /* This flag is used for configuring specific 
		     *  DLCI(s) when they become active.
 		     */ 
		    chan->dlci_configured = DLCI_CONFIG_PENDING;

		    if (chan->start)
		        set_chan_state(ifp, WAN_CONNECTED);
	        }
        }
    }
	
    for (ifnode = card->wandev.if_head; ifnode && ifnode->ifp; 
	 ifnode = ifnode->pnext)
    {
	chan = ifnode->ifp->if_softc;

	if (chan->dlci_configured == DLCI_CONFIG_PENDING)
	    if (!fr_init_dlci(card, chan))
		num_requests++;
    }
    /* Allow timer interrupts */
    if (num_requests)
    {
	card->u.f.timer_int_enabled = TMR_INT_ENABLED_INARP;
        flags->imask |= FR_INTR_TIMER;
    }
    return 1;
}

/*
 * =============================================================================
 * fr_init_dlci() - 
 *
 */
static int fr_init_dlci(sdla_t *card, fr_channel_t *chan)
{
    fr_dlc_conf_t cfg;
	    
    bzero(&cfg, sizeof(cfg));

    if ( chan->cir_status == CIR_DISABLED) 
    {
        cfg.cir_fwd = cfg.cir_bwd  = 16;
        cfg.bc_fwd = cfg.bc_bwd = 16;
        cfg.conf_flags = 0x0001;
    } 
    else 
        if (chan->cir_status == CIR_ENABLED) 
        {
            cfg.cir_fwd = cfg.cir_bwd = chan->cir;
            cfg.bc_fwd  = cfg.bc_bwd  = chan->bc;
            cfg.be_fwd  = cfg.be_bwd  = chan->be;
            cfg.conf_flags = 0x0000;
        }

    if (fr_dlci_configure( card, &cfg , chan->dlci))
    {
        log(LOG_ERR, "%s: DLCI Configure failed for %d\n",
					card->devname, chan->dlci);
        return 1;
    }

    chan->dlci_configured = DLCI_CONFIGURED;

    /* Allow timer interrupts */
    if (chan->inarp == INARP_REQUEST && 
	card->wandev.state == WAN_CONNECTED)
    {
        chan->inarp_tick = SYSTEM_TICK;
    }	

    /* Read the interface byte mapping into the channel 
     * structure.
     */
    read_DLCI_IB_mapping(card, chan);
    return 0;
 }

/******* Miscellaneous ******************************************************/

/* 
 * ===========================================================================
 * Update channel state. 
 */
static int update_chan_state (struct ifnet* ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    sdla_t* card = chan->card;
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;
    int err;

    do
    {
	mbox->cmd.command = FR_LIST_ACTIVE_DLCI;
	mbox->cmd.length = 0;
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    if (!err) 
    {
	unsigned short* list = (void*)mbox->data;
	int cnt = mbox->cmd.length / sizeof(short);

	for (; cnt; --cnt, ++list) 
	{
	    if (*list == chan->dlci) 
	    {
 	        set_chan_state(ifp, WAN_CONNECTED);
	        break;
	    }
        }
    }

    return err;
}

/* 
 * ============================================================================
 * Set channel state.
 */
static void set_chan_state (struct ifnet* ifp, int state)
{
    fr_channel_t* chan = ifp->if_softc;
    unsigned long flags;

    flags = read_eflags();
    disable_intr();

    if (chan->state != state) 
    {
	switch (state) 
	{
	case WAN_CONNECTED:
	    log(LOG_INFO, "%s%d: Set DLCI %d connected\n",
				ifp->if_name, ifp->if_unit, chan->dlci);
	    fr_up(ifp);
	    break;

	case WAN_CONNECTING:
	    log(LOG_INFO, "%s%d: Set DLCI %d connecting...\n",
				ifp->if_name, ifp->if_unit, chan->dlci);
	    ifp->if_flags |= IFF_RUNNING;
	    break;

	case WAN_DISCONNECTED:
	    log(LOG_INFO, "%s%d: Set DLCI %d disconnected!\n",
				ifp->if_name, ifp->if_unit, chan->dlci);
	    fr_down(ifp);
   	    break;
	}

	chan->state = state;
    }

    chan->state_tick = SYSTEM_TICK;
    write_eflags(flags);
}

/* 
 *============================================================================
 * Find network device by its channel number.
 */
static struct ifnet* find_channel (sdla_t* card, unsigned dlci)
{
    if (dlci > HIGHEST_VALID_DLCI)
	return NULL;
    return(card->u.f.dlci_to_dev_map[dlci]);
}

/* 
 * ============================================================================
 * Check to see if a frame can be sent. If no transmit buffers available,
 * enable transmit interrupts.
 *
 * Return:	1 - Tx buffer(s) available
 *		0 - no buffers available
 */
static int is_tx_ready (sdla_t* card, fr_channel_t* chan)
{
    unsigned char sb;

    if (card->hw.type == SDLA_S514)
	return 1;

    sb = inb(card->hw.port);
    if (sb & 0x02) 
	return 1;

    return 0;
}

/* 
 * ============================================================================
 * Convert decimal string to unsigned integer.
 * If len != 0 then only 'len' characters of the string are converted.
 */
static unsigned int dec_to_uint (unsigned char* str, int len)
{
    unsigned val;

    if (!len) 
	len = strlen(str);

    for (val = 0; len && is_digit(*str); ++str, --len)
	val = (val * 10) + (*str - (unsigned)'0');

    return val;
}


/* 
 * =============================================================================
 * Store a UDP management packet for later processing.
 */
static int 
store_udp_mgmt_pkt(int udp_type, char udp_pkt_src, sdla_t* card, 
				   struct mbuf *m, int dlci)
{
    int udp_pkt_stored = 0;
    int data_length = m->m_pkthdr.len;

    if(!card->u.f.udp_pkt_lgth && (data_length <= MAX_LGTH_UDP_MGNT_PKT)) 
    {
	card->u.f.udp_pkt_lgth = data_length;
        card->u.f.udp_type = udp_type;
	card->u.f.udp_pkt_src = udp_pkt_src;
        card->u.f.udp_dlci = dlci;
	m_copydata(m, 0, data_length, card->u.f.udp_pkt_data);
	card->u.f.timer_int_enabled = TMR_INT_ENABLED_UDP;
	udp_pkt_stored = 1;
    }
    else
	log(LOG_ERR, "%s: UDP packet not stored for DLCI %d\n", 
						card->devname, dlci);

    m_freem(m);
	
    return (udp_pkt_stored);
}

/* 
 * =============================================================================
 * Process UDP call of type FPIPE8ND
 */
static int process_udp_mgmt_pkt(sdla_t* card)
{
    int c_retry = MAX_CMD_RETRY;
    unsigned char frames;
    unsigned char *data_addr = NULL;
    unsigned int data_length = 0;
    unsigned short buffer_length = 0;
    struct mbuf *m = NULL, *m_curr = NULL;
    fr_mbox_t* mbox = card->mbox;
    int err;
    struct timeval tv;
    int udp_mgmt_req_valid = 1;
    struct ifnet* ifp = NULL;
    fr_channel_t* chan = NULL;
    fr_udp_pkt_t *fr_udp_pkt = NULL;
    unsigned short num_trc_els;
    fr_trc_el_t* ptr_trc_el = NULL;
    fr_trc_el_t trc_el;
    fpipemon_trc_t* fpipemon_trc = NULL;
    char udp_pkt_src = card->u.f.udp_pkt_src; 
    int dlci = card->u.f.udp_dlci;

    /* Find network interface for this packet */
    ifp = find_channel(card, dlci);
    chan = ifp->if_softc;

    fr_udp_pkt = (fr_udp_pkt_t *)card->u.f.udp_pkt_data;

    /*
     * If the UDP packet is from the network, we are going to have to transmit
     * a response. Before doing so, we must check to see that we are not 
     * currently transmitting a frame (in 'if_send()') and that we are not 
     * already in a 'delayed transmit' state.
     */
    if (udp_pkt_src == UDP_PKT_FRM_NETWORK)
    {
	if ((ifp->if_flags & IFF_OACTIVE) || (card->u.f.tx_interrupts_pending))
	    return 0; 
    }

    fr_udp_pkt = (fr_udp_pkt_t *)card->u.f.udp_pkt_data;
 
    switch(fr_udp_pkt->cblock.command) 
    {
    case FPIPE_ENABLE_TRACING:
    case FPIPE_DISABLE_TRACING:
    case FPIPE_GET_TRACE_INFO:
    case FR_SET_FT1_MODE:
        if (udp_pkt_src == UDP_PKT_FRM_NETWORK) 
        {
	    chan->drvstats_gen.UDP_PIPE_mgmt_direction_err ++;
	    udp_mgmt_req_valid = 0;
	    break;
        }
        break; 
    case FPIPE_FT1_READ_STATUS:
    case FR_FT1_STATUS_CTRL:
        if(card->hw.fwid != SFID_FR508) 
        {
            chan->drvstats_gen.UDP_PIPE_mgmt_adptr_type_err ++;
	    udp_mgmt_req_valid = 0;
	    break;
        }
        break;
    }

    if(!udp_mgmt_req_valid) 
    {
	/* set length to 0 */
	fr_udp_pkt->cblock.length = 0;
	/* set return code */
	fr_udp_pkt->cblock.result = (card->hw.fwid != SFID_FR508) ? 0x1F : 0xCD;
    } 
    else 
    {   
	switch(fr_udp_pkt->cblock.command) 
        {
	case FPIPE_ENABLE_TRACING:
	    if (!card->TracingEnabled) 
	    {
		do 
		{
		    mbox->cmd.command = FR_SET_TRACE_CONFIG;
                    mbox->cmd.length = 1;
                    mbox->cmd.dlci = 0x00;
               	    mbox->data[0] = fr_udp_pkt->data[0] | RESET_TRC;
		    err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
                } while (err && c_retry-- && fr_event(card, err, mbox));

                if(err) 
		{
		    card->TracingEnabled = 0;
		    /* set the return code */
		    fr_udp_pkt->cblock.result =	mbox->cmd.result;
		    mbox->cmd.length = 0;
		    break;
		}

		sdla_peek(&card->hw, NO_TRC_ELEMENTS_OFF, &num_trc_els, 2);
		sdla_peek(&card->hw, BASE_TRC_ELEMENTS_OFF,
				&card->u.f.trc_el_base, 4);
		card->u.f.curr_trc_el = card->u.f.trc_el_base;
                card->u.f.trc_el_last = (unsigned char*)card->u.f.curr_trc_el + 
				 ((num_trc_els - 1) * sizeof(fr_trc_el_t));
   
		/* Calculate the maximum trace data area in */
		/* the UDP packet */
	        card->u.f.trc_bfr_space=(MAX_LGTH_UDP_MGNT_PKT -
				/*sizeof(fr_encap_hdr_t) -*/
				sizeof(ip_pkt_t) -
				sizeof(udp_pkt_t) -
				sizeof(wp_mgmt_t) -
				sizeof(cblock_t));

		/* set return code */
		fr_udp_pkt->cblock.result = 0;

            } 
	    else 
	    { 
		/* set return code to line trace already enabled */
		fr_udp_pkt->cblock.result = 1;
            }
	    mbox->cmd.length = 0;
	    card->TracingEnabled = 1;
	    break;

        case FPIPE_DISABLE_TRACING:
	    if(card->TracingEnabled) 
	    {
		do 
		{
		    mbox->cmd.command = FR_SET_TRACE_CONFIG;
		    mbox->cmd.length = 1;
		    mbox->cmd.dlci = 0x00;
		    mbox->data[0] = ~ACTIVATE_TRC;
		    err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
		} while (err && c_retry-- && fr_event(card, err, mbox));
            }

            /* set return code */
	    fr_udp_pkt->cblock.result = 0;
	    mbox->cmd.length = 0;
	    card->TracingEnabled = 0;
	    break;

        case FPIPE_GET_TRACE_INFO:
	    /* Line trace cannot be performed on the 502 */
            if (!card->TracingEnabled) 
	    {
		/* set return code */
                fr_udp_pkt->cblock.result = 1;
                mbox->cmd.length = 0;
                break;
            }

	    (void *)ptr_trc_el = card->u.f.curr_trc_el;

            buffer_length = 0;
	    fr_udp_pkt->data[0x00] = 0x00;

            for(frames = 0; frames < MAX_FRMS_TRACED; frames ++) 
	    {
	        sdla_peek(&card->hw, (unsigned long)ptr_trc_el, 
				(void *)&trc_el.flag, sizeof(fr_trc_el_t));
                if (trc_el.flag == 0x00) 
		    break;
                if ((card->u.f.trc_bfr_space - buffer_length) < 
		    sizeof(fpipemon_trc_hdr_t)) 
		{ 
		    fr_udp_pkt->data[0x00] |= MORE_TRC_DATA;
                    break;
                }

		fpipemon_trc = 
			(fpipemon_trc_t *)&fr_udp_pkt->data[buffer_length]; 
		fpipemon_trc->fpipemon_trc_hdr.status = trc_el.attr;
               	fpipemon_trc->fpipemon_trc_hdr.tmstamp = trc_el.tmstamp;
		fpipemon_trc->fpipemon_trc_hdr.length = trc_el.length;

		if (!trc_el.offset || !trc_el.length) 
		    fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x00;
		else 
		    if ((trc_el.length + sizeof(fpipemon_trc_hdr_t) + 1) >
		        (card->u.f.trc_bfr_space - buffer_length))
		    {
			fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x00;
                        fr_udp_pkt->data[0x00] |= MORE_TRC_DATA;
		    }
		    else 
		    {
			fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x01;
                        sdla_peek(&card->hw, trc_el.offset, fpipemon_trc->data,
							 trc_el.length);
		    }	

		trc_el.flag = 0x00;
                sdla_poke(&card->hw, (unsigned long)ptr_trc_el, &trc_el.flag,1);
		ptr_trc_el ++;
		if ((void *)ptr_trc_el > card->u.f.trc_el_last)
		    (void*)ptr_trc_el = card->u.f.trc_el_base;

		buffer_length += sizeof(fpipemon_trc_hdr_t);
                if (fpipemon_trc->fpipemon_trc_hdr.data_passed) 
		    buffer_length += trc_el.length;

		if (fr_udp_pkt->data[0x00] & MORE_TRC_DATA) 
		    break;
            }
                      
	    if(frames == MAX_FRMS_TRACED) 
		fr_udp_pkt->data[0x00] |= MORE_TRC_DATA;
       
	    card->u.f.curr_trc_el = (void *)ptr_trc_el;

            /* set the total number of frames passed */
            fr_udp_pkt->data[0x00] |= ((frames << 1) & (MAX_FRMS_TRACED << 1));

	    /* set the data length and return code */
	    fr_udp_pkt->cblock.length = mbox->cmd.length = buffer_length;
            fr_udp_pkt->cblock.result = 0;
            break;

	case FPIPE_FT1_READ_STATUS:
	    sdla_peek(&card->hw, 0xF020, &fr_udp_pkt->data[0x00] , 2);
	    fr_udp_pkt->cblock.length = 2;
	    fr_udp_pkt->cblock.result = 0;
	    break;

	case FPIPE_FLUSH_DRIVER_STATS:
	    init_chan_statistics(chan);
	    init_global_statistics(card);
	    mbox->cmd.length = 0;
	    break;

	case FPIPE_ROUTER_UP_TIME:
	    microtime(&tv);
	    chan->router_up_time = tv.tv_sec - chan->router_start_time;
	    *(unsigned long *)&fr_udp_pkt->data = chan->router_up_time;	
	    mbox->cmd.length = 4;
	    break;

	case FR_FT1_STATUS_CTRL:
	    if (fr_udp_pkt->data[0] == 1) 
	    {
		if (rCount++ != 0)
		{
	            fr_udp_pkt->cblock.result = 0;
	            mbox->cmd.length = 1;
		    break;
		} 
	    }
           
	    /* Disable FT1 MONITOR STATUS */
            if (fr_udp_pkt->data[0] == 0) 
	    {
	        if ( --rCount != 0) 
		{
		    fr_udp_pkt->cblock.result = 0;
		    mbox->cmd.length = 1;
		    break;
	        } 
	    }  
	    break;
	case FPIPE_DRIVER_STAT_IFSEND:
	    bcopy(&chan->drvstats_if_send.if_send_entry, fr_udp_pkt->data, 
		   sizeof(if_send_stat_t));
	    mbox->cmd.length = sizeof(if_send_stat_t);
	    break;
	case FPIPE_DRIVER_STAT_INTR:
	    bcopy(&card->statistics.isr_entry, fr_udp_pkt->data,
		   sizeof(global_stats_t));
	    bcopy(&chan->drvstats_rx_intr.rx_intr_no_socket,
	          &fr_udp_pkt->data[sizeof(global_stats_t)],
		   sizeof(rx_intr_stat_t));
	    mbox->cmd.length = sizeof(global_stats_t) + sizeof(rx_intr_stat_t);
	    break;
	case FPIPE_DRIVER_STAT_GEN:
            bcopy(&chan->drvstats_gen.UDP_PIPE_mgmt_malloc_err,
	    	  fr_udp_pkt->data, sizeof(pipe_mgmt_stat_t)); 
	    bcopy(&card->statistics,
	          &fr_udp_pkt->data[sizeof(pipe_mgmt_stat_t)],
		  sizeof(global_stats_t));
	    fr_udp_pkt->cblock.result = 0;
	    fr_udp_pkt->cblock.length = sizeof(global_stats_t) +
					sizeof(rx_intr_stat_t);
	    mbox->cmd.length = fr_udp_pkt->cblock.length; 
	    break;
        default:
 	    do 
	    {
	        bcopy(&fr_udp_pkt->cblock.command, &mbox->cmd,  
							sizeof(fr_cmd_t));
	        if (mbox->cmd.length) 
		    bcopy((char *)fr_udp_pkt->data, &mbox->data,  
							mbox->cmd.length);
	        err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
	    } while (err && c_retry-- && fr_event(card, err, mbox));

	    if (!err)
	        chan->drvstats_gen.UDP_PIPE_mgmt_adptr_cmnd_OK ++;
	    else
	        chan->drvstats_gen.UDP_PIPE_mgmt_adptr_cmnd_timeout ++;

	    /* copy the result back to our buffer */
	    bcopy(&mbox->cmd, &fr_udp_pkt->cblock.command, sizeof(fr_cmd_t));

	    if (mbox->cmd.length) 
	        bcopy(&mbox->data, &fr_udp_pkt->data, mbox->cmd.length);
        } 
    }
   
    /* Fill UDP TTL */
    fr_udp_pkt->ip_pkt.ttl = card->wandev.ttl;
    data_length = reply_udp(card->u.f.udp_pkt_data, mbox->cmd.length);

    if (udp_pkt_src == UDP_PKT_FRM_NETWORK) 
    {
	err = fr_send(card, dlci, 0, data_length, card->u.f.udp_pkt_data, 1);
	if (err) 
	    chan->drvstats_gen.UDP_PIPE_mgmt_adptr_send_passed ++;
	else
	    chan->drvstats_gen.UDP_PIPE_mgmt_adptr_send_failed ++;
    } 
    else 
    {
	int s = 0;

	data_addr = card->u.f.udp_pkt_data;
	/* get the first ( header ) mbuf */
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL) 
	{
	    chan->drvstats_gen.UDP_PIPE_mgmt_no_socket ++;
            log(LOG_ERR, "%s: UDP mgmt cmnd, no socket buffers available!\n",
							card->devname);
	    /* i guess nothing can be done ... the packet is gone */
	    return ENOMEM;
	}

	/* fill in the header */
	m->m_pkthdr.rcvif = ifp;     /* our fr? interface */
	m->m_pkthdr.len	  = data_length;     /* length of the data */
	m->m_len	  = MHLEN;   /* well, at least as of now */

	/* m_curr will be running through the chain. m points to the header */
	m_curr = m; 

	/* let's start filling in the chain */
        while(data_length > 0)  
	{
	    /* check if we can use cluster instead ... */
	    if (data_length > MINCLSIZE) 
	    {
		/* all right - get the cluster */
		MCLGET(m_curr, M_DONTWAIT);
		if (m_curr->m_flags & M_EXT) 
		{
		    /* all right - it worked - set the new length */
		    m_curr->m_len = MCLBYTES;
		} 
		else 
		{
		    chan->drvstats_gen.UDP_PIPE_mgmt_no_socket ++;
		    /* it failed - well, screw it - what can we do ? */
		    m_freem(m);
		    log(LOG_ERR, "%s: Failed to allocate cluster !!!\n", 
							card->devname);
		    return ENOMEM;
		}
	    }

	    /* 
	     * ok, at this point we've got mbuf pointed by 'm_curr' which is 
	     * eitehr a regular one, or a cluster depending on how much data do 
	     * we have left. The beauty is that from here we dont care what kind
	     * of mbuf it is since all the macros work correctly on both.
	     */
	    m_curr->m_len = min(m_curr->m_len, data_length);
    
	    /* 
	     * Now, we can start copying data into mbuf data area given by 
	     * mtod() macro. We copy m_curr->m_len bytes.
	     */
	    bcopy( data_addr, mtod(m_curr, caddr_t), m_curr->m_len );
   
            /* adjust data pointer and remaining length */
	    data_addr += m_curr->m_len;
	    data_length -= m_curr->m_len;
   
	    /* done with this mbuf - allocate the next one if needed */
	    if (data_length > 0) 
	    {
	        MGET(m_curr->m_next, M_DONTWAIT, MT_DATA);
	        if (m_curr->m_next == NULL) 
	        {
		    chan->drvstats_gen.UDP_PIPE_mgmt_no_socket ++;
		    /* it failed - well, screw it - we cant do much here */
		    m_freem(m);
		    log(LOG_ERR, "%s: Failed to allocate mbuf !!!\n",
							card->devname);
		    return ENOMEM;
	        }
	        /* set m_curr to point to the new one, and set up initial length
		 */
	        m_curr        = m_curr->m_next;
	        m_curr->m_len = MLEN;
	    }
        } /* while ... */

	/* 
	 * Ok, at this point we've got an mbuf chain pointed by m which contains
	 * a UDP management reply - now we just have to pass it up the stack -
	 * and that's it !!!
	 * Note that we neither report those packets to BPF nore update any 
	 * stat. Those packets are meant to be invisible ;-)
	 * Also we run inside splimp already - do there's no need to call it 
	 * again.
	 */
	s = splimp();
	schednetisr(NETISR_IP);
	if (IF_QFULL(&ipintrq)) 
	{
	    chan->drvstats_gen.UDP_PIPE_mgmt_not_passed_to_stack ++;
	    /* oh, no - IP queue is full - well - screw it. 
	     * We can't do much here. 
	     */ 
	    log(LOG_ERR, "%s: IP queue is full, drop packet!\n", card->devname);
	    IF_DROP(&ipintrq);
	    m_freem(m);
	    splx(s);
	    return ENOMEM;
	}
  
	/* ok enqueue the packet */
	IF_ENQUEUE(&ipintrq, m);
	splx(s);
	dlog((LOG_DEBUG, "%s: UDP packet passed back on IP's queue\n", ifp->if_name));
	chan->drvstats_gen.UDP_PIPE_mgmt_passed_to_stack ++;
    }

    card->u.f.udp_pkt_lgth = 0;

    return 1;
}

/* 
 * ============================================================================
 * send_inarp_request()- sends out an InARP request packet on specified
 *			    interface
 */
static void 
send_inarp_request(sdla_t *card, struct ifnet *ifp)
{
    struct ifaddr	     *ifa = TAILQ_FIRST(&ifp->if_addrhead);
    struct sockaddr_in *saddr = NULL;	/* the address */
    arphdr_fr_t        arphdr;

    dlog((LOG_DEBUG, "%s%d: Sending InARP request...\n", ifp->if_name, ifp->if_unit));

    /* set up the pointer to the local address */
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) == NULL) 
		return;
    saddr = (struct sockaddr_in *)ifa->ifa_addr;

    /* InARP request */
    arphdr.ar_hrd = 0x0F00;	            /* Frame Relay HW type */
    arphdr.ar_pro = 0x0008;	       	    /* IP Protocol         */
    arphdr.ar_hln = 2;		 	    /* HW addr length      */
    arphdr.ar_pln = 4;		 	    /* IP addr length      */
    arphdr.ar_op = htons(0x08);	 	    /* InARP Request       */
    arphdr.ar_sha = 0;		 	    /* src HW DLCI - Doesn't matter */
    arphdr.ar_sip = saddr->sin_addr.s_addr; /* Local Address       */
    arphdr.ar_tha = 0;			    /* dst HW DLCI - Doesn't matter */
    arphdr.ar_tip = 0;			    /* Remote Address -- what we want */
  
    /* send it out ... */
    send_inarp_packet(card, &arphdr, ifp);

    /* done */
    return;
}

/* 
 * ============================================================================
 * send_inarp_packet() - sends previously formatted InARP packet on 
 *			    specified interface 
 */
static void     
send_inarp_packet(sdla_t *card, arphdr_fr_t * arphdr, struct ifnet *ifp)
{
    fr_channel_t* chan = ifp->if_softc;
    arphdr_1490_t      *ArpPacket = NULL;
    unsigned short	 data_length;	 /* data length */
    int err = 0;

    data_length = sizeof(arphdr_1490_t) + sizeof(arphdr_fr_t);

    /* allocate space for the packet */
    if ((ArpPacket = malloc(data_length, M_DEVBUF, M_NOWAIT)) == NULL) 
    {
        /* oops ... well we'll retry later */
	goto send_failed;
    }

    /* prepare the header */
    ArpPacket->control = 0x03;
    ArpPacket->pad     = 0x00;
    ArpPacket->NLPID   = 0x80;
    ArpPacket->OUI[0]  = 0;
    ArpPacket->OUI[1]  = 0;
    ArpPacket->OUI[2]  = 0;
    ArpPacket->PID     = 0x0608;

    /* copy the packet */
    bcopy(arphdr, ArpPacket + 1, sizeof(arphdr_fr_t));

    dlog((LOG_DEBUG, "%s: Sending InARP request on DLCI %d.\n", 
						card->devname, chan->dlci));
    err = fr_send(card, chan->dlci, 0, 
		  sizeof(arphdr_1490_t) + sizeof(arphdr_fr_t),
		  (void *)ArpPacket, 0);
    if (err)
        goto send_failed;

    /* see if it was InARP request - if so - flip the flag indicating that */
    /* we are awaiting InARP reply ...					 */
    if (arphdr->ar_op == htons(0x08)) 
    {
	dlog((LOG_DEBUG, "%s%d: - sent InARP request ...\n", 
					ifp->if_name, ifp->if_unit));
	chan->fr_inarp_reply_expected = 1;
	chan->fr_inarp_reply_timer    = INARP_REQUEST_TIMEOUT;
	/* schedule watchdog */
	ifp->if_timer = 1; 		/* call it in a second */
    } 
    else 
        dlog((LOG_DEBUG, "%s%d: - sent InARP reply ...\n", 
					ifp->if_name, ifp->if_unit));
 	
    /* free the packet and exit */
    if (ArpPacket)
    	free(ArpPacket, M_DEVBUF);

    /* done */
    return;

send_failed:

    /* ok, if the send failed for some reason - what we gonna do is to */
    /* copy the InARP packet into dlci_scp->fr_inarp_pkt and schedule  */
    /* watchdog which will try to resend it ...			     */
    log(LOG_ERR, "fr%d - failed to send InARP packet ..scheduling watchdog.\n",
						ifp->if_unit);

    /* copy the packet and set the flag for the watchdog */
    bcopy(arphdr, &chan->fr_inarp_pkt, sizeof(arphdr_fr_t));
    chan->fr_resend_inarp = 1;

    /* schedule watchdog */
    ifp->if_timer = 1; 		/* call it in a second */

    /*  free the pkt if needed and exit */
    if (ArpPacket) 
	free(ArpPacket, M_DEVBUF);

    /* done */
    return;
}

/* 
 * =============================================================================
 * Check packet for ARP Type
 */

int is_arp(void *buf)
{
    arphdr_1490_t *arphdr = (arphdr_1490_t *)buf;

    if (arphdr->pad   == 0x00  &&
        arphdr->NLPID == 0x80  &&
        arphdr->PID   == 0x0608) 
	return 1;
    else return 0;
}

/* 
 *=============================================================================
 * Process ARP Packet Type
 */

int process_ARP(arphdr_1490_t *ArpPacket, sdla_t *card, struct ifnet* ifp)
{
    struct ifaddr *ifa = NULL;
    struct sockaddr_in *local_addr = NULL, *dest_addr = NULL, *net_mask = NULL;
    arphdr_fr_t *arphdr = (arphdr_fr_t *)(ArpPacket + 1); /* Skip header */
    fr_rx_buf_ctl_t* frbuf = card->rxmb;
    fr_channel_t *chan = ifp->if_softc;

    /* get pointers to the IP addresses and masks */
    if ((ifa = TAILQ_FIRST(&ifp->if_addrhead)) == NULL) 
	return EFAULT;
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) == NULL)
	return EFAULT;	
    local_addr = (struct sockaddr_in *)ifa->ifa_addr;
    dest_addr  = (struct sockaddr_in *)ifa->ifa_dstaddr;
    net_mask   = (struct sockaddr_in *)ifa->ifa_netmask;

    switch (ntohs(arphdr->ar_op)) 
    {
    case 0x08:  /* Inverse ARP request  -- Send Reply, add route. */
	/* Check for valid Address */
        log(LOG_INFO, "%s: Recvd PtP addr %s -InArp Req\n", 
			chan->name, in_ntoa(arphdr->ar_sip));

        /* ok, let's see if the sender's IP is on the same subnet as ours 
         * if not - we don't reply and write a log message  	      
         */
        if ((local_addr->sin_addr.s_addr & net_mask->sin_addr.s_addr) != 
	    (arphdr->ar_sip & net_mask->sin_addr.s_addr)) 
        {
	    log(LOG_ERR, "%s: Invalid PtP address.  InARP ignored.\n", 
							card->devname);
            return -1;
        }

        if (local_addr->sin_addr.s_addr == arphdr->ar_sip) 
        {
	    log(LOG_ERR, "%s: Local addr = PtP addr.  InARP ignored.\n", 
							card->devname);
            return -1;
        } 
        arphdr->ar_op = htons(0x09);    /* InARP Reply */

        /* Set addresses */
        arphdr->ar_tip = arphdr->ar_sip;
        arphdr->ar_sip = local_addr->sin_addr.s_addr;

        fr_send(card, frbuf->dlci, 0, frbuf->length, (void *)ArpPacket, 0);

        /* Modify Point-to-Point Address */
        chan->dst_addr = arphdr->ar_tip;

        /* Add Route Flag */
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
         */
        chan->route_flag = ADD_ROUTE;
        /* now if the interface has InARP enabled - we would like to update 
         * routing and interface information to reflect new dest. ip   	
         */
	card->poll      = &process_route; 
        ifp->if_timer = 1;

        /* done */
	break;

    case 0x09:  /*  Inverse ARP reply */
	/* Check for valid Address */
        log(LOG_INFO, "%s: Recvd PtP addr %s -InArp Reply\n", 
				chan->name, in_ntoa(arphdr->ar_sip));

	/* reset the flag ... */
	chan->fr_inarp_reply_expected = 0;

        /* ok, let's see if the sender's IP is on the same subnet as ours
	 * if not - we don't reply and write a log message  	      
         */
        if ((net_mask->sin_addr.s_addr & arphdr->ar_sip) != 
	    (net_mask->sin_addr.s_addr & local_addr->sin_addr.s_addr)) 
	{
	    log(LOG_ERR, "%s: Invalid PtP address. InARP ignored.\n", 
							card->devname);
            return -1;
        }

        if (local_addr->sin_addr.s_addr == arphdr->ar_sip) 
	{
	    log(LOG_ERR, "%s: Local addr = PtP addr.  InARP ignored.\n", 
							card->devname);
            return -1;
        }

        /* Modify Point-to-Point Address */
 	dest_addr->sin_addr.s_addr = arphdr->ar_sip;

        /* Add Route Flag */
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
	 */
	chan->route_flag = ADD_ROUTE;
        chan->inarp = INARP_CONFIGURED;
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
         */
	card->poll      = &process_route; 
	ifp->if_timer = 1;

        /* Modify Point-to-Point Address */
	chan->dst_addr = arphdr->ar_sip;
        break;

    default:  /* ARP's and RARP's -- Shouldn't happen. */
        break;
    }

    return 0;
}


/* 
 * ============================================================================
 * Perform the Interrupt Test by running the READ_CODE_VERSION command 
 * MAX_INTR_TEST_COUNTER times.
 */
static int fr_intr_test( sdla_t* card )
{
    fr_mbox_t* mb = card->mbox;
    int err,i;

    err = fr_set_intr_mode(card, FR_INTR_READY, card->wandev.mtu, 0 );
	
    if (err == CMD_OK) 
    {
	for ( i = 0; i < MAX_INTR_TEST_COUNTER; i++ ) 
	{
 	    /* Run command READ_CODE_VERSION */
	    mb->cmd.length  = 0;
	    mb->cmd.command = FR_READ_CODE_VERSION;
	    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
	    if (err != CMD_OK) 
	        fr_event(card, err, mb);
        }
    } 
    else 
	return err;

    err = fr_set_intr_mode( card, 0, card->wandev.mtu, 0 );

    if (err != CMD_OK) 
	return err;

    return 0;
}

/* 
 * ============================================================================
 * Determine what type of UDP call it is. DRVSTATS or FPIPE8ND ?
 */
static int udp_pkt_type(struct mbuf *m, sdla_t* card)
{
    fr_udp_pkt_t *fr_udp_pkt = (fr_udp_pkt_t *)mtod(m, fr_udp_pkt_t *);

    if ((fr_udp_pkt->ip_pkt.protocol == UDPMGMT_UDP_PROTOCOL) &&
        (fr_udp_pkt->ip_pkt.ver_inet_hdr_length == 0x45) &&
        (fr_udp_pkt->udp_pkt.udp_dst_port == ntohs(card->wandev.udp_port)) &&
        (fr_udp_pkt->wp_mgmt.request_reply == UDPMGMT_REQUEST)) 
    {
        if (!strncmp(fr_udp_pkt->wp_mgmt.signature, UDPMGMT_FPIPE_SIGNATURE, 8))
	    return UDP_FPIPE_TYPE;
        if (!strncmp(fr_udp_pkt->wp_mgmt.signature, UDPMGMT_DRVRSTATS_SIGNATURE, 8))
	    return UDP_DRVSTATS_TYPE;
    }
    return UDP_INVALID_TYPE;
}

/*
 * =============================================================================
 * Initializes the Statistics values in the fr_channel structure.
 */
void init_chan_statistics( fr_channel_t* chan)
{
    bzero(&chan->drvstats_if_send.if_send_entry, 
				    sizeof(drvstats_if_send_t));
    bzero(&chan->drvstats_rx_intr.rx_intr_no_socket, 
				    sizeof(drvstats_rx_intr_t));
    bzero(&chan->drvstats_gen.UDP_PIPE_mgmt_malloc_err, 
			            sizeof(pipe_mgmt_stat_t));
}
	
/* 
 * =============================================================================
 * Initializes the Statistics values in the Sdla_t structure.
 */
void init_global_statistics( sdla_t* card )
{
    /* Intialize global statistics for a card */
    bzero(&card->statistics.isr_entry, sizeof(global_stats_t));
}

/* 
 * =============================================================================
 * 
 */
static void read_DLCI_IB_mapping( sdla_t* card, fr_channel_t* chan )
{
    fr_mbox_t* mbox = card->mbox;
    int retry = MAX_CMD_RETRY;	
    dlci_IB_mapping_t* result; 
    int err, counter, found;	

    do
    {
	mbox->cmd.command = FR_READ_DLCI_IB_MAPPING;
	mbox->cmd.length = 0;	
	err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
    } while (err && retry-- && fr_event(card, err, mbox));

    if (mbox->cmd.result != 0)
	log(LOG_ERR, "%s: Read DLCI IB Mapping failed\n", chan->name); 

    counter = mbox->cmd.length / sizeof(dlci_IB_mapping_t);
    result = (void *)mbox->data;

    found = 0;
    for (; counter; --counter, ++result) 
    {
	if (result->dlci == chan->dlci) 
	{
	    chan->IB_addr = result->addr_value;
	    if (card->hw.type == SDLA_S514)
	        chan->dlci_int_interface = 
		      (void*)((unsigned char*)card->hw.dpmbase + chan->IB_addr);
   	    else 
	        chan->dlci_int_interface = 
		      (void*)((unsigned char*)card->hw.dpmbase + 
		      (chan->IB_addr & 0x00001FFF));
	    found = 1;
	    break;
	} 
    }
    if (!found)
	log(LOG_ERR, "%s: DLCI %d not found by IB MAPPING cmd\n", 
					card->devname, chan->dlci);
}

/****** End *****************************************************************/
