/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_ppp.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
 * sdla_ppp.c	WANPIPE(tm) Multiprotocol WAN Link Driver. PPP module.
 *
 * Author: 	Alex Feldman <al.feldman@@sangoma.com>
 *
 * ============================================================================
 *
 * Nov 21, 1999  Alex Feldman    o Updated for FreeBSD 3.2
 * Jan 06, 1997	 Gene Kozin      o Initial version.
 *****************************************************************************
 */

#include "bpfilter.h"

#include <stddef.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <net/route.h>

#define _GNUC_
#include <net/wanpipe.h>		/* WANPIPE common user API definision */
#include <net/sdla_ppp.h>		/* PPP firmware API definitions */

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/if_ether.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <netipx/ipx.h>
#include <netipx/ipx_if.h>

#include <machine/clock.h>

/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#define WANDEBUG_PPP
#else
#define	STATIC		static
#endif

#define	PPP_DFLT_MTU	1500	/* default MTU */
#define	PPP_MAX_MTU	4000	/* maximum MTU */
#define PPP_HDR_LEN	1

#define SYSTEM_TICK ticks
#define HZ          hz

#define	CONNECT_TIMEOUT	(90*HZ)	/* link connection timeout */
#define	HOLD_DOWN_TIME	(5*HZ)	/* link hold down time : Changed from 30 to 5 */

/* For handle_IPXWAN() */
#define CVHexToAscii(b) (((unsigned char)(b) > (unsigned char)9) ? ((unsigned char)'A' + ((unsigned char)(b) - (unsigned char)10)) : ((unsigned char)'0' + (unsigned char)(b)))

/* Macro for enabling/disabling debugging comments */

#define DCD(a)   ( a & 0x08 ? "HIGH" : "LOW" )
#define CTS(a)   ( a & 0x20 ? "HIGH" : "LOW" ) 
#define LCP(a)   ( a == 0x09 ? "OPEN" : "CLOSED" )
#define IP(a)    ( a == 0x09 ? "ENABLED" : "DISABLED" )

#define TMR_INT_ENABLED_UPDATE  	1
#define TMR_INT_ENABLED_PPP_EVENT	2
#define TMR_INT_ENABLED_UDP		4

/* Set Configuraton Command Definitions */
#define PERCENT_TX_BUFF			60
#define TIME_BETWEEN_CONF_REQ  		30
#define TIME_BETWEEN_PAP_CHAP_REQ	30
#define WAIT_PAP_CHAP_WITHOUT_REPLY     300
#define WAIT_AFTER_DCD_CTS_LOW          5
#define TIME_DCD_CTS_LOW_AFTER_LNK_DOWN 10
#define WAIT_DCD_HIGH_AFTER_ENABLE_COMM 900
#define MAX_CONF_REQ_WITHOUT_REPLY      10
#define MAX_TERM_REQ_WITHOUT_REPLY      2
#define NUM_CONF_NAK_WITHOUT_REPLY      5
#define NUM_AUTH_REQ_WITHOUT_REPLY      10

#define END_OFFSET 0x1F0

/* For debugging convenience */
#ifdef WANDEBUG_PPP
#define dlog(x) log x
#else
#define dlog(x)
#endif

/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with PPP specific data
 */
  
typedef struct ppp_private_area
{
	sdla_t* card;	
	unsigned long router_start_time;	/*router start time in sec */
	unsigned long tick_counter;		/*used for 5 second counter*/
	unsigned mc;				/*multicast support on or off*/
	unsigned char enable_IPX;
	unsigned long network_number;
	unsigned char pap;
	unsigned char chap;
	unsigned char sysname[31];		/* system name for in-bnd auth*/
	unsigned char userid[511];		/* list of user ids */
	unsigned char passwd[511];		/* list of passwords */
	unsigned protocol;			/* Mbuff Protocol */
	uint32_t ip_local;			/* Local IP Address */
	uint32_t ip_remote;			/* remote IP Address */

	unsigned char timer_int_enabled;	/* Who enabled the timer inter*/
	unsigned char update_comms_stats;	/* Used by update function */
	unsigned long curr_trace_addr;		/* Trace information */
	unsigned long start_trace_addr;
	unsigned long end_trace_addr;

	unsigned short udp_pkt_lgth;
	char  udp_pkt_src;
      	char  udp_pkt_data[MAX_LGTH_UDP_MGNT_PKT];

	/* PPP specific statistics */
	if_send_stat_t if_send_stat;
	rx_intr_stat_t rx_intr_stat;
	pipe_mgmt_stat_t pipe_mgmt_stat;

	unsigned long router_up_time; 

} ppp_private_area_t;

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;

/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update(wan_device_t *wandev);
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf);
static int del_if(wan_device_t *wandev, struct ifnet *ifp);

/* WANPIPE-specific entry points */
static int wpp_exec (struct sdla *card, void *u_cmd, void *u_data);

/* Network device interface */
static int ppp_close  (struct ifnet *ifp);
/* ifnet entry points */
static 	int 	ppp_output		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		     struct rtentry *));
static 	void	ppp_start		/* initiate output routine */
		__P((struct ifnet *));
static	int	ppp_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static	void	ppp_watchdog		/* timer routine */
		__P((struct ifnet *));
static	int	ppp_init		/* init routine */
		__P((struct ifnet *, struct ifaddr *));

/* PPP firmware interface functions */
static int ppp_read_version(sdla_t *card, char *str);
static int ppp_set_outbnd_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_set_inbnd_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_configure(sdla_t *card, void *data);
static int ppp_set_intr_mode(sdla_t *card, unsigned char mode);
static int ppp_read_intr_mode(sdla_t *card);
static int ppp_comm_enable(sdla_t *card);
static int ppp_comm_disable(sdla_t *card);
static int ppp_get_err_stats(sdla_t *card);
static int ppp_send(sdla_t *card, void *data, unsigned len, unsigned proto);
static int ppp_error(sdla_t *card, int err, ppp_mbox_t *mb);

static void wpp_isr(sdla_t *card);
static void rx_intr(sdla_t *card);
static void event_intr(sdla_t *card);
static void timer_intr(sdla_t *card);

/* Background polling routines */
static void process_route(void *arg);
static void poll_disconnected(sdla_t *card);

/* Miscellaneous functions */
static int read_info( sdla_t *card );
static int read_connection_info (sdla_t *card);
static int remove_route( sdla_t *card );
static int config508(ppp_private_area_t *ppp_priv_area, struct ifaddr *ifa, sdla_t *card);
static void show_disc_cause(sdla_t * card, unsigned cause);
static int reply_udp( unsigned char *data, unsigned int mbox_len );
static int process_udp_mgmt_pkt(sdla_t *card, struct ifnet *ifp, 
				ppp_private_area_t *ppp_priv_area);
static void init_ppp_tx_rx_buff( sdla_t *card );
static int ppp_intr_test( sdla_t *card );
static int udp_pkt_type( struct mbuf *m0 , sdla_t *card);
static void init_ppp_priv_struct( ppp_private_area_t *ppp_priv_area);
static void init_global_statistics( sdla_t *card );
static int tokenize(char *str, char **tokens);
static char* strstrip(char *str, char *s);
static int chk_bcast_mcast_addr(sdla_t* card, struct ifnet* ifp,
				struct mbuf *m);
static int Read_connection_info;
static int Intr_test_counter;
static unsigned short available_buffer_space;

/* IPX functions */
#ifdef SUPPORT_IPX
static void switch_net_numbers(unsigned char *sendpacket, 
			       unsigned long network_number, 
			       unsigned char incoming);
static int handle_IPXWAN(unsigned char *sendpacket, char *devname, 
		         unsigned char enable_PX, unsigned long network_number, 
			 unsigned short proto);
#endif

static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, struct ifnet* ifp,
                                ppp_private_area_t* ppp_priv_area );

static unsigned short calc_checksum (char *data, int len);
static char *my_strtok(char *p, const char *tok);
static char *my_strchr(const char *p, char ch);



/****** Public Functions ****************************************************/

/*============================================================================
 * PPP protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpp_init(sdla_t *card, wandev_conf_t *conf)
{
    union
    {
	char str[80];
    } u;

    dlog((LOG_DEBUG, "%s: Starting PPP initialization\n", card->wandev.name));
    /* Verify configuration ID */
    if (conf->config_id != WANCONFIG_PPP) 
    {
	log(LOG_ERR, "%s: invalid configuration ID %u!\n",
					card->devname, conf->config_id);
	return -EINVAL;
    }

    /* Initialize miscellaneous pointers to structures on the adapter */
    switch (card->hw.type) 
    {
    case SDLA_S508:
    	card->mbox =(void*)((u_char*)card->hw.dpmbase + PPP508_MB_OFFS);
    	card->flags=(void*)((u_char*)card->hw.dpmbase + PPP508_FLG_OFFS);
    	break;

    case SDLA_S514:
    	card->mbox =(void*)((u_char*)card->hw.dpmbase + PPP514_MB_OFFS);
    	card->flags=(void*)((u_char*)card->hw.dpmbase + PPP514_FLG_OFFS);
    	break;

    default:
    	return -EINVAL;
    }

    /* Read firmware version.  Note that when adapter initializes, it
     * clears the mailbox, so it may appear that the first command was
     * executed successfully when in fact it was merely erased. To work
     * around this, we execute the first command twice.
     */
    if (ppp_read_version(card, NULL) || ppp_read_version(card, u.str))
    {
	return -EIO;
    }

    log(LOG_INFO, "%s: running PPP firmware v%s\n",card->devname, u.str); 
    /* Adjust configuration and set defaults */
    card->wandev.mtu = (conf->mtu) ? min(conf->mtu, PPP_MAX_MTU) : PPP_DFLT_MTU;

    card->wandev.bps		= conf->bps;
    card->wandev.interface	= conf->interface;
    card->wandev.clocking	= conf->clocking;
    card->wandev.station	= conf->station;
    card->isr			= &wpp_isr;
    card->poll			= NULL; 
    card->exec			= &wpp_exec;
    card->wandev.update		= &update;
    card->wandev.new_if		= &new_if;
    card->wandev.del_if		= &del_if;
    card->wandev.udp_port   	= conf->udp_port;
    card->wandev.ttl		= conf->ttl;
    card->irq_dis_if_send_count = 0;
    card->irq_dis_poll_count 	= 0;
    card->u.p.authenticator 	= conf->u.ppp.authenticator;
    card->u.p.ip_mode 		= 
		conf->u.ppp.ip_mode ? conf->u.ppp.ip_mode : WANOPT_PPP_STATIC;
    card->TracingEnabled    	= 0;
    Read_connection_info    	= 1;

    /* initialize global statistics */
    init_global_statistics(card);

    /* Update card state */
    card->wandev.state	= WAN_DISCONNECTED;

    return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/*============================================================================
 * Update device status & statistics.
 */
static int update(wan_device_t *wandev)
{
    sdla_t* card = wandev->private;
    struct ifnet* ifp = card->wandev.if_head->ifp;
    volatile ppp_private_area_t *ppp_priv_area = ifp->if_softc;
    ppp_flags_t *flags = card->flags;
    unsigned long timeout;

    /* sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL))
	return -EFAULT;

    if (wandev->state == WAN_UNCONFIGURED)
	return -ENODEV;

    ppp_priv_area->update_comms_stats = 2;
    ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_UPDATE;
    flags->imask |= PPP_INTR_TIMER;	

    /* wait a maximum of 1 second for the statistics to be updated */ 
    timeout = SYSTEM_TICK;
    for(;;) 
    {
	if (ppp_priv_area->update_comms_stats == 0)
	    break;
        if ((SYSTEM_TICK - timeout) > (1 * HZ))
	{
    	    ppp_priv_area->update_comms_stats = 0;
 	    ppp_priv_area->timer_int_enabled &=
				~TMR_INT_ENABLED_UPDATE; 
 	    return -EAGAIN;
	}
    }

    return 0;
}

/*============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf)
{
    sdla_t *card = wandev->private;
    ppp_private_area_t *ppp_priv_area;
	
    if (wandev->ndev)
	return -EEXIST;

    if ((conf->name[0] == '\0') || (strlen(conf->name) > WAN_IFNAME_SZ)) 
    {
	log(LOG_ERR, "%s: invalid interface name!\n", card->devname);
	return -EINVAL;
    }

    /* allocate and initialize private data */
    ppp_priv_area = malloc(sizeof(ppp_private_area_t), M_DEVBUF, M_NOWAIT);

    if (ppp_priv_area == NULL)
	return	-ENOMEM;

    bzero(ppp_priv_area, sizeof(ppp_private_area_t));

    ppp_priv_area->card = card; 

    if (wandev->new_if_cnt > 0)
    {
	free(ppp_priv_area, M_DEVBUF);
	return -EEXIST;
    }

    /* initialize data */
    strcpy(card->u.p.if_name, conf->name); 

    /* increment number of interfaces */
    wandev->new_if_cnt++;

    /* initialize data in ppp_private_area structure */
    init_ppp_priv_struct( ppp_priv_area );

    ppp_priv_area->mc = conf->mc;
    ppp_priv_area->pap = conf->pap;
    ppp_priv_area->chap = conf->chap;

    /* If no user ids are specified */
    if (!strlen(conf->userid) && (ppp_priv_area->pap||ppp_priv_area->chap))
    {
	free(ppp_priv_area, M_DEVBUF);
	return -EINVAL;
    }

    /* If no passwords are specified */
    if (!strlen(conf->passwd) && (ppp_priv_area->pap||ppp_priv_area->chap))
    {
	free(ppp_priv_area, M_DEVBUF);
	return -EINVAL;
    }

    if (strlen(conf->sysname) > 31)
    {
	free(ppp_priv_area, M_DEVBUF);
	return -EINVAL;
    }

    /* If no system name is specified */
    if (!strlen(conf->sysname) && (card->u.p.authenticator))
    {
	free(ppp_priv_area, M_DEVBUF);
	return -EINVAL;
    }

    /* copy the data into the ppp private structure */
    bcopy(conf->userid, ppp_priv_area->userid, strlen(conf->userid));
    bcopy(conf->passwd, ppp_priv_area->passwd, strlen(conf->passwd));
    bcopy(conf->sysname, ppp_priv_area->sysname, strlen(conf->sysname));

    ppp_priv_area->enable_IPX = conf->enable_IPX;
    if (conf->network_number)
	ppp_priv_area->network_number = conf->network_number;
    else
	ppp_priv_area->network_number = 0xDEADBEEF;

    /* prepare network device data space for registration */
    ifp->if_softc          = ppp_priv_area;
    ifp->if_unit           = 0; /* description in sdlamain.c */
    ifp->if_name           = malloc(strlen(conf->name)+1, M_DEVBUF, M_NOWAIT);
    if (ifp->if_name == NULL)
    {
	log(LOG_ERR, "%s: Cannot allocate memory!\n", conf->name);
	ifp->if_softc = NULL;
	free(ppp_priv_area, M_DEVBUF);
	return -ENOMEM;
    }
    strcpy(ifp->if_name, conf->name);
    ifp->if_mtu            = wandev->mtu; /* ETHERMTU */
    ifp->if_type           = IFT_OTHER;
    ifp->if_flags          = IFF_POINTOPOINT; 
    ifp->if_output   	   = ppp_output;
    ifp->if_start          = ppp_start;
    ifp->if_watchdog       = ppp_watchdog;
    ifp->if_init           = (void*)ppp_init;
    ifp->if_ioctl	   = ppp_ioctl;
    ifp->if_snd.ifq_len    = 0;
    ifp->if_snd.ifq_maxlen = ifqmaxlen;
	
    /* ???? This members was not initialize in 3.1 */
    ifp->if_hdrlen = PPP_HDR_LEN;

    return 0;
}

/* 
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
    sdla_t                  *card = wandev->private;

    dlog((LOG_DEBUG, "%s: Deleting PPP interface %s...\n", 
					wandev->name, ifp->if_name));
    if (card == NULL)
	return -ENODEV;

    /* Initialize interrupt handler pointer */
    card->isr = NULL;

    /* Close current interface, set interrupt permission byte to 0 */ 
    ppp_close(ifp);

    /* Initialize network interface */
    ifp->if_output = NULL;
    ifp->if_start = NULL;
    ifp->if_ioctl = NULL;
    ifp->if_watchdog = NULL;
    ifp->if_init = NULL;
    ifp->if_flags &= ~(IFF_RUNNING | IFF_UP | IFF_OACTIVE);
    if (ifp->if_softc)
    {
        free(ifp->if_softc, M_DEVBUF);
        ifp->if_softc = NULL;
    }
    dlog((LOG_DEBUG, "%s: Interface %s is deleted\n", wandev->name, ifp->if_name));
    return 0;
}

/****** WANPIPE-specific entry points ***************************************/

/*============================================================================
 * Execute adapter interface command.
 */

/* FIXME: Why do we need this ???? */
static int wpp_exec(struct sdla *card, void *u_cmd, void *u_data)
{
    ppp_mbox_t *mbox = card->mbox;
    int len;

    if (!u_cmd)
	return -EFAULT;

    if (copyin(u_cmd, (void*)&mbox->cmd, sizeof(ppp_cmd_t)))
	return -EFAULT;

    len = mbox->cmd.length;

    if (len)
        if (!u_data || copyin(u_data, (void*)&mbox->data, len)) 
	    return -EFAULT;

    /* execute command */
    if (!sdla_exec(mbox))
	return -EIO;

    /* return result */
    if (copyout((void*)&mbox->cmd, u_cmd, sizeof(ppp_cmd_t)))
	return -EFAULT;   
    len = mbox->cmd.length;

    if (len && u_data &&
		copyout((void*)&mbox->data, u_data, len))
	return -EFAULT;	

    return 0;
}

/****** Network Device Interface ********************************************/

/*
 * ptpipe_ioctl() -
 *
 */
static int
ppp_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
    ppp_private_area_t *ppp_priv_area = ifp->if_softc; 
    sdla_t             *card = ppp_priv_area->card;
    struct ifaddr      *ifa = (struct ifaddr *)data; /* address struct */
    int 	       s;	    	    /* saved interrupt state */
    int		       error = 0;    	    /* resulting status */

    s = splimp();

    switch (command) 
    {
    case SIOCSIFADDR:
	dlog((LOG_DEBUG, "%s - ioctl(SIOCSIFADDR) called.\n", ifp->if_name));
	ifp->if_flags |= IFF_UP;
	/* check the protocol family - we currently support IP only */
	switch(ifa->ifa_addr->sa_family) 
	{
	case AF_INET:
	    ppp_init(ifp, ifa);
	    break;
	default:
	    error = EAFNOSUPPORT;
	    break;
	}
	break;

    case SIOCGIFADDR:
	dlog((LOG_DEBUG, "%s - ioctl(SIOCGIFADDR) called.\n", ifp->if_name));
	break;

    case SIOCSIFFLAGS:
	dlog((LOG_DEBUG, "%s - ioctl(SIOCSIFFLAGS) called.\n", ifp->if_name));
	/*
	 * If the interface is marked up - enable communications. If down - 
	 * disable communications.  IFF_UP is taken care of before entering 
	 * this function.
	 */
	if (((ifp->if_flags & IFF_UP)==0)/*&& (ifp->if_flags & IFF_RUNNING)*/) 
	{
	    /* bring it down */
	    log(LOG_INFO, "%s - bringing interface down.\n", ifp->if_name);
	    ifp->if_flags &= ~IFF_RUNNING;
	    ppp_comm_disable(card);
	    /* set up the disconnect cause */
	} 
	else 
	{
	    struct timeval tv;
	    /* bring it up */
	    log(LOG_INFO, "%s - bringing interface up.\n", ifp->if_name);
	    
	    /*
	     * If flag configured is not set, run interrupt test. This test 
	     * will be run each time when specific wanpipe will start.
	     */
	    splx(s);
	    if (!card->configured)
		if (ppp_intr_test(card))
		    return -EIO;
 	    s = splimp();

	    /* 
	     * ppp_enable_comms() does not mark the interface IFF_RUNNING - this
	     * is done when the protocol-open interrupt is received. If the 
	     * IFF_RUNNING is already set - this means that the communications
	     * are aldready enabled, actually that the interface is already up.
	     * In this case we dont wanna enable communications again.
	     */
	    if ((ifp->if_flags & IFF_RUNNING) == 0) 
		ppp_comm_enable(card);
	    /* update the up time ... */
	    microtime(&tv);
	    ppp_priv_area->router_up_time = tv.tv_sec; 	/* BF-0003 */
	}
	break;

    case SIOCGIFFLAGS:
    	dlog((LOG_DEBUG, "%s - ioctl(SIOCGIFFLAGS) called.\n", ifp->if_name));
	break;

    case SIOCSIFMTU:
	dlog((LOG_DEBUG, "%s - ioctl(SIOCSIFMTU) called.\n", ifp->if_name));
	error = EINVAL;
	break;

    default:
	dlog((LOG_DEBUG, "%s - ioctl(default) called.\n", ifp->if_name));
	error = EINVAL;
	break;
    }

    splx(s);
    return error;
}

/*
 * ppp_watchdog() - this function will be called for the first time
 *		    a second after first protocol-open interrupt. 
 *		    After that it will reschedule itself to be called
 *		    once every second. 
 *		    The routine will check the protocol state and 
 *		    update the interface state if it changes. Actually
 *		    the only change we intereseted in is when the protocol
 *		    goes down ( coz when it goes up we will receive the 
 *		    protocol-open interrupt ).
 */
static void
ppp_watchdog(struct ifnet * ifp)
{
    ppp_private_area_t *ppp_priv_area = ifp->if_softc;
    sdla_t *card = ppp_priv_area->card;
    
    if (card->poll)
	card->poll(card); 
    return;
}


/* 
 * ============================================================================
 * Open network interface.
 * o enable communications and interrupts.
 * o prevent module from unloading by incrementing use count
 *
 * Return 0 if O.k. or errno.
 */
static int ppp_init (struct ifnet *ifp, struct ifaddr *ifa)
{
    ppp_private_area_t *ppp_priv_area = ifp->if_softc;
    sdla_t *card = ppp_priv_area->card;
    ppp_flags_t *flags = card->flags;
    struct timeval tv;
    int s = 0;
    int err = 0;

    dlog((LOG_DEBUG, "%s: Initialize interface...\n", ifp->if_name));
    s = splimp();

    if (!card->configured)
    {
	if (config508(ppp_priv_area, ifa, card))
	{
	    splx(s);
	    return -EIO;
	}

    }

    /* Initialize Rx/Tx buffer control fields */
    init_ppp_tx_rx_buff( card );

    if (ppp_set_intr_mode(card, PPP_INTR_RXRDY|
    				PPP_INTR_TXRDY|
	    			PPP_INTR_MODEM|
	    			PPP_INTR_DISC |
	    			PPP_INTR_OPEN |
	    			PPP_INTR_DROP_DTR |
				PPP_INTR_TIMER)) 
    {
	splx(s);
	return -EIO;
    }

    /* Turn off the transmit and timer interrupt */
    flags->imask &= ~(PPP_INTR_TXRDY | PPP_INTR_TIMER);

    /* If you are not the authenticator and any one of the protocol is 
     * enabled then we call the set_out_bound_authentication.
     */
    if (!card->u.p.authenticator  && 
        (ppp_priv_area->pap || ppp_priv_area->chap)) 
    {
	if ( ppp_set_outbnd_auth(card, ppp_priv_area) )
	{
	    splx(s);
	    return -EIO;
	}
    } 

    /* If you are the authenticator and any one of the protocol is enabled
     * then we call the set_in_bound_authentication.
     */
    if ( card->u.p.authenticator && (ppp_priv_area->pap || ppp_priv_area->chap))
    {
	if ( ppp_set_inbnd_auth(card, ppp_priv_area) )
	{
	    splx(s);
	    return -EIO;
	}
    }

    if (ppp_comm_enable(card)) 
    {
	splx(s);
	return -EIO;
    }
	
    wanpipe_set_state(card, WAN_CONNECTING);
    wanpipe_open(card);
    microtime(&tv);
    ppp_priv_area->router_start_time = tv.tv_sec;
    splx(s);
    return err;
}

/*============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int ppp_close(struct ifnet *ifp)
{
    ppp_private_area_t *ppp_priv_area = ifp->if_softc;
    sdla_t *card = ppp_priv_area->card;
    int s = 0;

    s = splimp();
	
    wanpipe_close(card);
    wanpipe_set_state(card, WAN_DISCONNECTED);
    ppp_set_intr_mode(card, 0);
    ppp_comm_disable(card);

    splx(s);
    return 0;
}


/*============================================================================
 * ptpipe_output() - output routine. The function tries to enqueue the 
 *		     packet and then if the device is not IFF_OATIVE
 *		     calls ptpipe_start(). The function returns an error
 *		     value or 0 on success.
 */
static int
ppp_output(struct ifnet *ifp, struct mbuf *m0, struct sockaddr *dst, struct rtentry *rt0)
{
    ppp_private_area_t *ppp_priv_area = ifp->if_softc;
    sdla_t *card = ppp_priv_area->card;
    ppp_flags_t *flags = card->flags;
    int udp_type = 0;
    int error = 0;
    struct ip_packet	 *ip_pckt;		/* used to interpret the data*/

    /* we used to check here for device being UP and RUNNING - but since  */
    /* we would like to have a monitor running even if the device is down */
    /* we have to move this check till after the UDP check 		*/

    /* check the protocol family - we currently support only IP */
    if (dst->sa_family != AF_INET) 
    {
        /* protocol family is not supported */
        log(LOG_ERR, "Protocol family is not supported ...\n");
        return EAFNOSUPPORT;
    }
    ++ppp_priv_area->if_send_stat.if_send_entry;

    if (m0 == NULL) 
    {
	/* If we get here, some higher layer thinks we've missed an
	 * tx-done interrupt.
	 */
	log(LOG_ERR, "%s: interface %s got kicked!\n", card->devname, ifp->if_name);
	++ppp_priv_area->if_send_stat.if_send_mbuf_null;
	return 0;
    }

    if ((ifp->if_flags & IFF_OACTIVE) != 0) 
    {
	/* If our device stays busy for at least 5 seconds then we will
	 * increment statistics. We expect that device never stays busy more than
	 * 5 seconds.
	 */
	if ((SYSTEM_TICK - ppp_priv_area->tick_counter) > (5 * HZ)) 
	{
	    card->wandev.stats.collisions++;
	    ppp_priv_area->if_send_stat.if_send_tbusy++;
            log(LOG_INFO, "%s: Transmit times out \n", card->devname);
	}
    }

    /* 
     * Ok, UDP management additions - we need to check a few things in order 
     * to be sure that this is our UDP management packet :
     *
     *  o this is AF_INET packet
     *  o this is UDP packet ( IPPCKT_IP_PROTO byte set to IPPROTO_UDP )
     *  o the UDP port corresponds to the one set for the driver
     *  o the UDP signature matches ours 
     *
     * To improve performance we will start checking with less-likely possible
     * event ( i.e. bottom-up in the above list ). Also since mbuf is at least
     * 128 bytes length - i don't even check for the possibility of a very short
     * packet ( like raw IP with no data at all ) - even if this happens we wont
     * get an exception since there's at least 128 bytes of allocated memery
     *
     * Oops, also i need to make sure its a request packet !!!
     *
     */
    ip_pckt = mtod(m0, struct ip_packet *);

    udp_type = udp_pkt_type( m0, card );
    if (udp_type == UDP_PTPIPE_TYPE)
    {
	if (store_udp_mgmt_pkt(UDP_PKT_FRM_STACK, card, m0, ifp, ppp_priv_area))
	    flags->imask |= PPP_INTR_TIMER;
	++ppp_priv_area->if_send_stat.if_send_PIPE_request;
	return 0;
    }

    /* Check for broadcast and multicast addresses 
     * If found, drop (deallocate) a packet and return.
     */
    if (chk_bcast_mcast_addr(card, ifp, m0))
	return 0;


    /*
     * Regular packet - check if the device is up and running, queue message
     * on interface, and start output if interface not yet active.
     */
    if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) 
    {
	/* device is not ready */
	log(LOG_ERR, "%s: Device is not ready ...\n", ifp->if_name);
	return ENETDOWN;
    }

    if (card->wandev.state != WAN_CONNECTED)
    {
        ++ card->wandev.stats.tx_dropped;
	log(LOG_ERR, "%s: Device is not connected.\n", ifp->if_name);
        error = ENETDOWN;
    }
    else
    {
	int s = 0;
	/* Add support IPX here */
#ifdef SUPPORT_IPX
#endif
        s =splimp();
        if (IF_QFULL(&ifp->if_snd)) 
        {
    	    /* update some statistics */
    	    /* send queue is full - drop the packet and exit */
    	    IF_DROP(&ifp->if_snd);
    	    splx(s);
    	    log(LOG_ERR, "%s: Send queue is full ...\n", ifp->if_name);
    	    log(LOG_INFO, "%s: Send queue is full ...\n", ifp->if_name);
    	    return ENOBUFS;
    	}

    	IF_ENQUEUE(&ifp->if_snd, m0);

    	if ((ifp->if_flags & IFF_OACTIVE) == 0) 
    	    ppp_start(ifp);
        splx(s);
    }

    /* update statistics */
    ifp->if_obytes += m0->m_pkthdr.len;
    return error;
}

/*
 * ptpipe_start() - called by the system when a packet is added to the 
 *		    output queue for the interface and the device flags
 *		    indicate the device is not currently busy ( IFF_OACTIVE
 *                  is clear ).
 *                  We will try to transmit the packet. In case there's no
 *  		    available buffer - we will set IFF_OACTIVE, enable 
 *		    transmit interrupt and exit. And then the tx interrupt 
 *		    handler will call ptpipe_start() and once its done - it 
 *		    will reset the IFF_OACTIVE.
 */
static void
ppp_start(struct ifnet * ifp)
{
    ppp_private_area_t   *ppp_priv_area = ifp->if_softc;
    sdla_t	         *card = ppp_priv_area->card;
    ppp_flags_t          *flags = card->flags;
    ppp_buf_ctl_t        *txbuf = card->u.p.txbuf;
    struct mbuf		 *opkt = NULL;			/* the packet */
    register struct mbuf *m = NULL;			/* current mbuf*/
    unsigned long 	 data_addr;			/* where to copy data*/
    unsigned short	 data_length;			/* data length */
    int 		 s = 0;			        /* saved int. state */

    s = splimp();

    /* We're trying to drain the queue as much as possible here because */
    /* if we manage to get more than one packet into transmit queue     */
    /* ( line-down/line-up scenario, or heavy load ) and take only one  */
    /* at a time out we get an effective delay ...		      */
    /* 							      BF-0001 */ 

    while (NULL != (opkt = ifp->if_snd.ifq_head)) 
    {	   /* BF-0001 */
	/* calculate tx element address */
        txbuf = card->u.p.txbuf;
	/* check if the buffer is available */
	if (txbuf->flag == 0x01) 
	{
	    /* no buffers */
	    dlog((LOG_DEBUG, "%s: TR buffers are full!!! Enabling TX INTR\n",
							ifp->if_name));
	    /* mark an interface active */
	    ifp->if_flags |= IFF_OACTIVE;

	    ppp_priv_area->tick_counter = SYSTEM_TICK;
	    card->wandev.stats.tx_errors++;
	    ppp_priv_area->if_send_stat.if_send_adptr_bfrs_full++;
	    ppp_priv_area->if_send_stat.if_send_tx_int_enabled++;

	    /* enable transmit interrupts */
	    flags->imask |= PPP_INTR_TXRDY;  /* unmask TX interrupt */

	    /* restore priority level and exit */
	    splx(s);
            return;
        }
  
	/* 
	 * ok, now when we've got an available buffer we can transmit something
	 */
	/* initialize some of our stuff */
	data_addr 	= txbuf->buf.ptr;
	data_length	= 0;

	/* go thru mbuf chain and copy data to the board */	
	for( m = opkt; m != NULL; m = m->m_next ) 
	{
	    /* Copy data to board */
	    sdla_poke(&card->hw, data_addr, mtod(m, caddr_t), m->m_len);
	    data_length += m->m_len;
	    data_addr += m->m_len;
 	}

	/* write data length to the transmit element - offset 1, length 2 */
	txbuf->length = data_length;

	/* start transmission */
	txbuf->flag = 1;

	/* dequeue packet from IP queue */
	IF_DEQUEUE(&ifp->if_snd, opkt);

	/* report the packet to BPF if present and attached */
#if NBPFILTER > 0
	if (ifp->if_bpf != NULL) 
	{	/* BF-0002 */
	    struct mbuf m0;
	    u_int af = AF_INET;

	    m0.m_next = opkt;
	    m0.m_len = 4;
	    m0.m_data = (char *)&af;

            bpf_mtap(ifp, &m0);
	}
#endif

	/* update statistics and free the packet */
	ifp->if_opackets++;
	m_freem(opkt);

	/* Update transmit buffer control fields */
	card->u.p.txbuf = ++txbuf;

	ppp_priv_area->if_send_stat.if_send_bfr_passed_to_adptr++;
	card->wandev.stats.tx_packets++;

	if ((void*)txbuf > card->u.p.txbuf_last)
	    card->u.p.txbuf = card->u.p.txbuf_base;

    } /* while there're packets in the queue */

    splx(s);		/* restore interrupt level and exit */
    return;
}


/*=============================================================================
 * Store a UDP management packet for later processing.
 */
static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                              struct mbuf *m, struct ifnet* ifp,
                              ppp_private_area_t* ppp_priv_area)
{
    int udp_pkt_stored = 0;

    if (!ppp_priv_area->udp_pkt_lgth && (m->m_len <= MAX_LGTH_UDP_MGNT_PKT))
    {
	ppp_priv_area->udp_pkt_lgth = m->m_len;
	ppp_priv_area->udp_pkt_src = udp_pkt_src;
       	bcopy(mtod(m, caddr_t), ppp_priv_area->udp_pkt_data, m->m_len);
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_UDP;
	ppp_priv_area->protocol = htons(ETHERTYPE_IP); 
	udp_pkt_stored = 1;
    }
    else
    {
	if (m->m_len > MAX_LGTH_UDP_MGNT_PKT)
	    log(LOG_ERR, "%s: PIPEMON UDP request too long : %d\n",
						card->devname, m->m_len);
	else
	    log(LOG_ERR, "%s: PIPEMON UDP request already pending\n",
						card->devname);
	ppp_priv_area->udp_pkt_lgth = 0;
    }

    m_freem(m);
    return(udp_pkt_stored);
}


/*============================================================================
 * Reply to UDP Management system.
 * Return length of reply.
 */
static int reply_udp( unsigned char *data, unsigned int mbox_len ) 
{
    unsigned short len, udp_length, temp, ip_length;
    unsigned long ip_temp;
    int even_bound = 0;
    ppp_udp_pkt_t *p_udp_pkt = (ppp_udp_pkt_t *)data;
 
    /* Set length of packet */
    len = sizeof(ip_pkt_t)+ 
    	  sizeof(udp_pkt_t)+
      	  sizeof(wp_mgmt_t)+
      	  sizeof(cblock_t)+
          mbox_len;

    /* fill in UDP reply */
    p_udp_pkt->wp_mgmt.request_reply = UDPMGMT_REPLY; 

    /* fill in UDP length */
    udp_length = sizeof(udp_pkt_t)+ 
		 sizeof(wp_mgmt_t)+
		 sizeof(cblock_t)+
		 mbox_len; 
 
    /* put it on an even boundary */
    if ( udp_length & 0x0001 ) 
    {
	udp_length += 1;
	len += 1;
	even_bound=1;
    } 

    temp = (udp_length<<8)|(udp_length>>8);
    p_udp_pkt->udp_pkt.udp_length = temp;

    /* swap UDP ports */
    temp = p_udp_pkt->udp_pkt.udp_src_port;
    p_udp_pkt->udp_pkt.udp_src_port = p_udp_pkt->udp_pkt.udp_dst_port; 
    p_udp_pkt->udp_pkt.udp_dst_port = temp;

    /* add UDP pseudo header */
    temp = 0x1100;
    *((unsigned short *)(p_udp_pkt->data+mbox_len+even_bound)) = temp;
    temp = (udp_length<<8)|(udp_length>>8);
    *((unsigned short *)(p_udp_pkt->data+mbox_len+even_bound+2)) = temp;

    /* calculate UDP checksum */
    p_udp_pkt->udp_pkt.udp_checksum = 0;
    p_udp_pkt->udp_pkt.udp_checksum = 
			calc_checksum(&data[UDP_OFFSET],udp_length+UDP_OFFSET);

    /* fill in IP length */
    ip_length = udp_length + sizeof(ip_pkt_t);
    temp = (ip_length<<8)|(ip_length>>8);
    p_udp_pkt->ip_pkt.total_length = temp;

    /* swap IP addresses */
    ip_temp = p_udp_pkt->ip_pkt.ip_src_address;
    p_udp_pkt->ip_pkt.ip_src_address = p_udp_pkt->ip_pkt.ip_dst_address;
    p_udp_pkt->ip_pkt.ip_dst_address = ip_temp;

    /* fill in IP checksum */
    p_udp_pkt->ip_pkt.hdr_checksum = 0;
    p_udp_pkt->ip_pkt.hdr_checksum = calc_checksum(data,sizeof(ip_pkt_t));

    return len;
} /* reply_udp */

unsigned short calc_checksum (char *data, int len)
{
    unsigned short temp; 
    unsigned long sum=0;
    int i;

    for( i = 0; i <len; i+=2 ) 
    {
	bcopy(&data[i], &temp, 2);
	sum += (unsigned long)temp;
    }

    while (sum >> 16 ) 
    {
	sum = (sum & 0xffffUL) + (sum >> 16);
    }

    temp = (unsigned short)sum;
    temp = ~temp;

    if (temp == 0) 
	temp = 0xffff;

    return temp;
}

/*
 ****************************************************************************
 *  If incoming is 0 (outgoing)- if the net numbers is ours make it 0
 *  if incoming is 1 - if the net number is 0 make it ours 
*/
#ifdef SUPPORT_IPX
static void 
switch_net_numbers(unsigned char *sendpacket, unsigned long network_number, unsigned char incoming)
{
    unsigned long pnetwork_number;

    pnetwork_number = (unsigned long)((sendpacket[6] << 24) + 
		      (sendpacket[7] << 16) + (sendpacket[8] << 8) + 
		      sendpacket[9]);

    if (!incoming) 
    {
	/* If the destination network number is ours, make it 0 */
	if (pnetwork_number == network_number) 
  	{
	    sendpacket[6] = sendpacket[7] = sendpacket[8] = 
				 sendpacket[9] = 0x00;
	}
    } 
    else 
    {
	/* If the incoming network is 0, make it ours */
	if (pnetwork_number == 0) 
	{
	    sendpacket[6] = (unsigned char)(network_number >> 24);
  	    sendpacket[7] = (unsigned char)((network_number & 
						 0x00FF0000) >> 16);
	    sendpacket[8] = (unsigned char)((network_number & 
						 0x0000FF00) >> 8);
	    sendpacket[9] = (unsigned char)(network_number & 
						 0x000000FF);
	}
    }

    pnetwork_number = (unsigned long)((sendpacket[18] << 24) + 
		      (sendpacket[19] << 16) + (sendpacket[20] << 8) + 
		      sendpacket[21]);

    if (!incoming) 
    {
	/* If the source network is ours, make it 0 */
	if (pnetwork_number == network_number) 
	{
	    sendpacket[18] = sendpacket[19] = sendpacket[20] = 
					 sendpacket[21] = 0x00;
	}
    } 
    else 
    {
	/* If the source network is 0, make it ours */
	if (pnetwork_number == 0) 
	{
	    sendpacket[18] = (unsigned char)(network_number >> 24);
	    sendpacket[19] = (unsigned char)((network_number & 
						 0x00FF0000) >> 16);
	    sendpacket[20] = (unsigned char)((network_number & 
						 0x0000FF00) >> 8);
	    sendpacket[21] = (unsigned char)(network_number & 
						 0x000000FF);
	}
    }
} /* switch_net_numbers */
#endif


/****** Interrupt Handlers **************************************************/

/*
 * ============================================================================
 * PPP interrupt service routine.
 */
STATIC void wpp_isr(sdla_t *card)
{
    ppp_flags_t *flags = card->flags;
    struct ifnet *ifp = card->wandev.if_head->ifp;
    ppp_private_area_t   *ppp_priv_area = ifp->if_softc;
    int i;
    char *ptr = &flags->iflag;

    card->in_isr = 1;
	
    ++card->statistics.isr_entry;

    /* FIXME: Do we need this */
    card->force_enable_irq = 0;

    card->buff_int_mode_unbusy = 0;

    switch (flags->iflag) 
    {
    case PPP_INTR_RXRDY:	/* receive interrupt  0x01  (bit 0)*/
	++card->statistics.isr_rx;
	rx_intr(card);
	break;

    case PPP_INTR_TXRDY:	/* transmit interrupt  0x02 (bit 1)*/
        /* 
	 * 1. Disable further tranmit interrupts.
	 * 2. Check the interface queue - if its not empty call ptpipe_start().
	 * 3. Reset the IFF_OACTIVE.
	 */
	++card->statistics.isr_tx;
	flags->imask &= ~PPP_INTR_TXRDY;
	/* interface is not active */
	ifp->if_flags &= ~IFF_OACTIVE;
	/* Update tick_counter */
	ppp_priv_area->tick_counter = SYSTEM_TICK;
	if (ifp->if_snd.ifq_head != NULL)
	    ppp_start(ifp);
	card->buff_int_mode_unbusy = 1;
	break;

    case PPP_INTR_CMD:      /* interface command completed */
	++Intr_test_counter;
	++card->statistics.isr_intr_test;
	break;

    case PPP_INTR_MODEM:    /* modem status change (DCD, CTS) 0x04 (bit 2)*/
    case PPP_INTR_DISC:  	/* Data link disconnected 0x10  (bit 4)*/
    case PPP_INTR_OPEN:   	/* Data link open 0x20  (bit 5)*/
    case PPP_INTR_DROP_DTR:	/* DTR drop timeout expired  0x40 bit 6 */
	event_intr(card);
	break;

    case PPP_INTR_TIMER:
	timer_intr(card);
	break;	 

    default:	/* unexpected interrupt */
	++card->statistics.isr_spurious;
	log(LOG_INFO, "%s: ID Bytes = ",card->devname);
 	for(i = 0; i < 8; i ++)
	    log(LOG_INFO, "0x%02x ", *(ptr + 0x28 + i));
	log(LOG_INFO, "\n");	
    }
	
    card->in_isr = 0;
    flags->iflag = 0;

}

/*============================================================================
 * Receive interrupt handler.
 */
static void rx_intr(sdla_t *card)
{
    ppp_buf_ctl_t *rxbuf = card->rxmb;
    struct ifnet *ifp = NULL;
    ppp_private_area_t *ppp_priv_area = NULL;
    struct mbuf *m = NULL, *ipkt = NULL;
    unsigned len;
    int i;
    ppp_flags_t *flags = card->flags;
    int protocol = 0, udp_type = 0;
    int s = 0;
    char *ptr = &flags->iflag;

    if (rxbuf->flag != 0x01) 
    {
	log(LOG_ERR, "%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
			card->devname, (unsigned)rxbuf, rxbuf->flag);
	log(LOG_INFO, "%s: ID Bytes = ",card->devname);
	for(i = 0; i < 8; i ++)
	    log(LOG_INFO, "0x%02x ", *(ptr + 0x28 + i));
	log(LOG_INFO, "\n");	
		
	++card->statistics.rx_intr_corrupt_rx_bfr;
	return;
    }

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", card->devname, __LINE__);
	return;
    } 
    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;
	
    if (ifp) 
    {
	len  = rxbuf->length;
	ppp_priv_area = ifp->if_softc;

	/* get the first ( header ) mbuf */
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m != NULL) 
	{
	    /* Copy data to the socket buffer */
	    unsigned addr = rxbuf->buf.ptr;

	    /* m will be running through the chain. ipkt always point to the
	     * header. 
	     */
	    ipkt  		 = m; 
	    ipkt->m_pkthdr.rcvif = ifp;	    /* our interface */
	    ipkt->m_pkthdr.len	 = len;	    /* length of the data */
	    ipkt->m_len		 = MHLEN;   /* ... well, at leasts as of now */

	    /* let's start filling in the chain */
            while(len > 0) 
	    {
		/* check if we can use cluster instead ... */
		if (len > MINCLSIZE) 
		{
		    /* all right - get the cluster */
		    MCLGET(m, M_DONTWAIT);
		    if (m->m_flags & M_EXT) 
		    {
			/* all right - it worked - set the new length */
			m->m_len = MCLBYTES;
		    } 
		    else 
		    {
			/* it failed - well, screw it - we'll try again later */
			m_freem(ipkt);
			log(LOG_ERR, "%s: Failed to allocate cluster !!!\n",
							ifp->if_name);
			ifp->if_iqdrops++;	
			return;
		    }
		}

	        /* 
		 * ok, at this point we've got mbuf pointed by 'm' which is 
		 * either a regular one, or a cluster depending on how much 
		 * data do we have left. The beauty is that from here we dont
		 * care what kind of mbuf it is since all the macros work 
		 * correctly on both.
		 */
		m->m_len = min( m->m_len, len );
    
		/* 
	 	 * Now, we can start copying data into mbuf data area given 
		 * by mtod() macro. We copy m->m_len bytes. Note that since 
		 * the S508s buffers are cyclic we might have to copy the 
		 * packet in two pieces
		 */
		if (addr + m->m_len <= card->u.p.rx_top + 1) 
		{
		    /* copy iRn one piece */
		    sdla_peek(&card->hw, addr, mtod(m, caddr_t), m->m_len);
		    addr += m->m_len;
		    len -= m->m_len;
		} 
		else 
		{
		    /* copy in two pieces */
		    int piece1_len = card->u.p.rx_top + 1 - addr;
		    int piece2_len = m->m_len - piece1_len;
	   
	            sdla_peek(&card->hw, addr, mtod(m, caddr_t), piece1_len);
		    sdla_peek(&card->hw, card->u.p.rx_top, 
		    		mtod(m, caddr_t) + piece1_len, piece2_len);
		    addr = card->u.p.rx_top + piece2_len;
		    len -= m->m_len;
		}
    
	        /* done with this mbuf - allocate the next one if needed */
		if (len > 0) 
		{
		    MGET(m->m_next, M_DONTWAIT, MT_DATA);
		    if (m->m_next == NULL) 
	 	    {
			/* it failed - well, screw it - we'll try again later */
			m_freem(ipkt);
			log(LOG_ERR, "%s: Failed to allocate mbuf !!!\n", 
								ifp->if_name);
			ifp->if_iqdrops++;	
			return;
		    }
		    /* set m to point to the new one, and set up initial 
		     * length
		     */
		    m = m->m_next;
		    m->m_len = MLEN;
		}
            } /* while ... */

	    /* Decapsulate packet */
       	    switch (rxbuf->proto) 
	    {
	    case 0x00:
		protocol = htons(ETHERTYPE_IP);
		break;
	    case 0x01:
		protocol = htons(ETHERTYPE_IPX);
		break;
	    }

	    udp_type = udp_pkt_type( m, card );
	    if (udp_type == UDP_PTPIPE_TYPE)
	    {
		/* Handle a UDP Request in Timer Interrupt */
		if (store_udp_mgmt_pkt(UDP_PKT_FRM_NETWORK, card, m, ifp,
                	              			ppp_priv_area))
           	    flags->imask |= PPP_INTR_TIMER;
		++ppp_priv_area->rx_intr_stat.rx_intr_PIPE_request;
	    }
	    else
#ifdef SUPPORT_IPX
	        if (handle_IPXWAN(mtod(ipkt, caddr_t),card->devname, 
				 ppp_priv_area->enable_IPX, 
				 ppp_priv_area->network_number,
				 protocol)) 
	        {
	            /* Handle an IPXWAN packet */
	            if (ppp_priv_area->enable_IPX) 
	            {
	    	        ppp_send(card, mtod(ipkt, caddr_t), m->m_len, 
					htons(ETHERTYPE_IPX));
		        m_freem(m);
	            } 
	            else
		        ++card->wandev.stats.rx_dropped;
	        }
		else
#endif
                {
		    /* Ok, this is a regular packet - do what's said above */
#if NBPFILTER > 0
		    if (ifp->if_bpf != NULL)
		    {	/* BF-0002 */
			struct mbuf m0;
			u_int af = AF_INET;

			m0.m_next = ipkt;
			m0.m_len = 4;
			m0.m_data = (char *)&af;       	

			bpf_mtap(ifp, &m0);
		    }
#endif

		    /* 
		     * Schedule IP input routine. Note that it makes much more
		     * sence to check first if the protocol queue is full - and
		     * then to schedule the input routine, but i took a look at
		     * kernel sources for ether_input() 
		     * [ /usr/src/sys/net/if_ethersubr.c ] and they do it in 
		     * this order - so i figured they know better.
		     */
		    s = splimp();
		    schednetisr(NETISR_IP);
		    if (IF_QFULL(&ipintrq)) 
		    {
			/* oh, no - IP queue is full - well - we'll try 
			 * again later 
			 */ 
			log(LOG_ERR, "%s: IP queue is full, drop packet !!!\n", 
								ifp->if_name);
			IF_DROP(&ipintrq);
			m_freem(ipkt);
			splx(s);
			return;
		    }
  
		    /* ok enqueue the packet */
		    IF_ENQUEUE(&ipintrq, ipkt);
		    splx(s);
	 
		    /* update interface statistics */
		    ifp->if_ipackets++;
	    	    card->wandev.stats.rx_packets++;
		    ppp_priv_area->rx_intr_stat.rx_intr_bfr_passed_to_stack++;
		}
	}
    	else
    	{
	    log(LOG_ERR, "%s: no socket buffers available!\n", card->devname);
	    ++card->wandev.stats.rx_dropped;
	    ++ppp_priv_area->rx_intr_stat.rx_intr_no_socket;
	    ifp->if_iqdrops++;
	    return;
        }
    } 
    else
    	++card->statistics.rx_intr_dev_not_started;

    /* Release buffer element and calculate a pointer to the next one */
    rxbuf->flag = 0x00;
    card->rxmb = ++rxbuf;
    if ((void*)rxbuf > card->u.p.rxbuf_last)
	card->rxmb = card->u.p.rxbuf_base;
}


/*
 * ============================================================================
 * Receive event interrupt.
 */
void event_intr (sdla_t *card)
{
    struct ifnet* ifp = NULL;
    ppp_private_area_t* ppp_priv_area = NULL;
    volatile ppp_flags_t *flags = card->flags;
    struct timeval tv;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", card->devname, __LINE__);
	return;
    } 
    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;

    switch (flags->iflag)
    {
    case PPP_INTR_MODEM:    /* modem status change (DCD, CTS) 0x04  (bit 2)*/
	log(LOG_INFO, "%s: Modem status: DCD=%s CTS=%s\n",
		     card->devname, DCD(flags->mstatus), CTS(flags->mstatus));
	break;

    case PPP_INTR_DISC:  	/* Data link disconnected 0x10  (bit 4) */
	log(LOG_INFO, "%s: Data link disconnected intr Cause %x\n",
				  card->devname, flags->disc_cause);
	if (flags->disc_cause &
	    (PPP_LOCAL_TERMINATION | PPP_DCD_CTS_DROP | PPP_REMOTE_TERMINATION))
	{
	    if (card->u.p.ip_mode == WANOPT_PPP_PEER) 
	    { 
		Read_connection_info = 1;
		remove_route (card);
	    }
	    wanpipe_set_state(card, WAN_DISCONNECTED);
	    show_disc_cause(card, flags->disc_cause);
	    ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
	    flags->imask |= PPP_INTR_TIMER;
	}
	ifp->if_flags &= ~IFF_UP;
	break;

    case PPP_INTR_OPEN:   	/* Data link open 0x20  (bit 5)*/
	log(LOG_INFO, "%s: PPP Link Open, LCP=%s IP=%s\n",
		   card->devname, LCP(flags->lcp_state), IP(flags->ip_state));
	/* Protocol is opened - mark interface*/
	ifp->if_flags |= IFF_RUNNING;
	/* reset disconnect cause */
        /*
         * ok, this is a hack - what happens is that sometimes when we bring 
	 * the interface up a disconnect interrupt occures which resets 
	 * the IFF_UP flag. The protocol-open interrupt always occures later on
	 * so we can fix it here. On second thought - the only way we can bring 
	 * up the interface which was down is through here - so i'll just be 
	 * always setting IFF_UPO flag here ;-)
	 */
	ifp->if_flags |= IFF_UP;

	/* update the up time ... */
	microtime(&tv);
	ppp_priv_area->router_up_time = tv.tv_sec; 	/* BF-0003 */

        /* 
         * also from this call the interface begins "intelligent" existance so
	 * we would like to activate the watchdog routine that will monitor the
	 * protocol state and update the interface state accordingly.
	 */
	ifp->if_timer = 1; 	/* call it in a second */

	if (flags->lcp_state == 0x09 && 
            (flags->ip_state == 0x09 || flags->ipx_state == 0x09))
	{
	    /* Initialize the polling timer and set the state
             * to WAN_CONNNECTED 
             */
            card->state_tick = SYSTEM_TICK;
  	    wanpipe_set_state(card, WAN_CONNECTED);
	    ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
	    flags->imask |= PPP_INTR_TIMER;
	}
        break;

    case PPP_INTR_DROP_DTR:	/* DTR drop timeout expired  0x40 bit 6 */
	log(LOG_INFO, "%s: DTR Drop Timeout Interrrupt \n", card->devname); 
	if (card->u.p.ip_mode == WANOPT_PPP_PEER) 
	{ 
	    Read_connection_info = 1;
	    remove_route (card);
	}
	wanpipe_set_state(card, WAN_DISCONNECTED);
	show_disc_cause(card, flags->disc_cause);
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
	flags->imask |= PPP_INTR_TIMER;
	break;

    default:
	log(LOG_ERR, "%s: Error, Invalid PPP Event\n",card->devname);
    }
}


/*
 * ============================================================================
 * Timer interrupt.
 */
void timer_intr (sdla_t *card)
{
    struct ifnet* ifp = NULL;
    ppp_private_area_t* ppp_priv_area = NULL;
    ppp_flags_t *flags = card->flags;

    if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL)
    {
	log(LOG_EMERG, "%s: Internal Error (line no. %d)!\n", card->devname, __LINE__);
	return;
    } 
    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;

    /* Update statistics */
    if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_UPDATE)
    {
	ppp_get_err_stats(card);
        if (!(--ppp_priv_area->update_comms_stats))
	{
	    ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
	}
    }

    /* PPIPEMON UDP request */
    if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_UDP)
    {
	process_udp_mgmt_pkt(card,ifp, ppp_priv_area);
	ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
    }

    /* PPP Event */
    if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_PPP_EVENT)
    {
	if (card->wandev.state == WAN_DISCONNECTED)
	{
	    poll_disconnected(card);
	}
	/* If the state is CONNECTING, it means that communicatins were
 	 * enabled. When the remote side enables its comminication we
 	 * should get an interrupt PPP_INTR_OPEN, thus turn off polling 
	 */
	else 
	    if (card->wandev.state == WAN_CONNECTING)
	    {
		/* Turn off the timer interrupt */
		ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PPP_EVENT;
	    }
	    /* If state is connected and we are in PEER mode 
	     * poll for an IP address which will be provided by remote end.
	     */
	    else 
	        if ((card->wandev.state == WAN_CONNECTED && 
		     card->u.p.ip_mode == WANOPT_PPP_PEER) && 
		    Read_connection_info)
		{
		    card->state_tick = SYSTEM_TICK;
		    if (!read_connection_info (card))
		    {
			card->poll = &process_route;
			ifp->if_timer = 1; 
		    }
		}
		else
		{
		    /* If we are using Static IP,no need to poll for
		     * an IP address. 
		     */
		    ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PPP_EVENT;
		}	

    }/* End of PPP_EVENT */

    /* Only disable the timer interrupt if there are no udp, statistic */
    /* updates or events pending */
    if (!ppp_priv_area->timer_int_enabled) 
	flags->imask &= ~PPP_INTR_TIMER;
}


#ifdef SUPPPORT_IPX
static int 
handle_IPXWAN(unsigned char *sendpacket, char *devname, unsigned char enable_IPX, unsigned long network_number, unsigned short proto)
{
    int i;

    if (proto == htons(ETHERTYPE_IPX)) 
    {
	/* It's an IPX packet */
	if (!enable_IPX) 
	{
	    /* Return 1 so we don't pass it up the stack. */
	    return 1;
	}
    } 
    else 
    {
	/* It's not IPX so pass it up the stack. */
	return 0;
    }

    if (sendpacket[16] == 0x90 && sendpacket[17] == 0x04)
    {
	/* It's IPXWAN */
	if (sendpacket[2] == 0x02 && sendpacket[34] == 0x00)
	{
	    /* It's a timer request packet */ 
	    log(LOG_INFO, "%s: Received IPXWAN Timer Request packet\n", devname);

	    /* Go through the routing options and answer no to every */
	    /* option except Unnumbered RIP/SAP */
	    for(i = 41; sendpacket[i] == 0x00; i += 5)
	    {
		/* 0x02 is the option for Unnumbered RIP/SAP */
		if (sendpacket[i + 4] != 0x02)
		{
		    sendpacket[i + 1] = 0;
		}
	    }

	    /* Skip over the extended Node ID option */
	    if (sendpacket[i] == 0x04)
	    {
		i += 8;
	    }

	    /* We also want to turn off all header compression opt. */
	    for(; sendpacket[i] == 0x80 ;)
	    {
		sendpacket[i + 1] = 0;
		i += (sendpacket[i + 2] << 8) + (sendpacket[i + 3]) + 4;
	    }

	    /* Set the packet type to timer response */
	    sendpacket[34] = 0x01;

	    log(LOG_INFO, "%s: Sending IPXWAN Timer Response\n", devname);
	}
	else 
	    if (sendpacket[34] == 0x02)
	    {
		/* This is an information request packet */
		log(LOG_INFO, "%s: Received IPXWAN Information Request packet\n",
								devname);

		/* Set the packet type to information response */
		sendpacket[34] = 0x03;

		/* Set the router name */
		sendpacket[51] = 'P';
		sendpacket[52] = 'T';
		sendpacket[53] = 'P';
		sendpacket[54] = 'I';
		sendpacket[55] = 'P';
		sendpacket[56] = 'E';
		sendpacket[57] = '-';
		sendpacket[58] = CVHexToAscii(network_number >> 28);
		sendpacket[59] = 
			CVHexToAscii((network_number & 0x0F000000)>> 24);
		sendpacket[60] = 
			CVHexToAscii((network_number & 0x00F00000)>> 20);
		sendpacket[61] = 
			CVHexToAscii((network_number & 0x000F0000)>> 16);
		sendpacket[62] = 
			CVHexToAscii((network_number & 0x0000F000)>> 12);
		sendpacket[63] = 
			CVHexToAscii((network_number & 0x00000F00)>> 8);
		sendpacket[64] = 
			CVHexToAscii((network_number & 0x000000F0)>> 4);
		sendpacket[65] = CVHexToAscii(network_number & 0x0000000F);
		for(i = 66; i < 99; i+= 1)
		    sendpacket[i] = 0;

		log(LOG_INFO, "%s: Sending IPXWAN Information Response packet\n",
								devname);
	    }
	    else
	    {
		log(LOG_ERR, "%s: Unknown IPXWAN packet!\n",devname);
		return 0;
	    }

        /* Set the WNodeID to our network address */
        sendpacket[35] = (unsigned char)(network_number >> 24);
	sendpacket[36] = (unsigned char)((network_number & 0x00FF0000) >> 16);
    	sendpacket[38] = (unsigned char)((network_number & 0x0000FF00) >> 8);
        sendpacket[38] = (unsigned char)(network_number & 0x000000FF);

    	return 1;
    } 
    else 
    {
        /* If we get here's its an IPX-data packet, so it'll get passed 
         * up the stack. 
         */

	/* switch the network numbers */
	switch_net_numbers(sendpacket, network_number, 1);	
	return 0;
    }
}
#endif

/****** Background Polling Routines  ****************************************/

/* All polling functions are invoked by the TIMER interrupt in the wpp_isr 
 * routine.  
 */

/*============================================================================
 * Monitor active link phase.
 */
static void process_route (void *arg)
{
    sdla_t       *card = (sdla_t*)arg;
    ppp_flags_t *flags = card->flags;

    if ((card->u.p.ip_mode == WANOPT_PPP_PEER) &&
        (Read_connection_info && flags->ip_state == 0x09))
    { 
        log(LOG_INFO, "%s: IPCP State Opened.\n", card->devname);
        read_info(card);
        Read_connection_info = 0;
    }
    card->poll = NULL;
}


/*============================================================================
 * Monitor physical link disconnected phase.
 *  o if interface is up and the hold-down timeout has expired, then retry
 *    connection.
 */
static void poll_disconnected(sdla_t *card)
{
    struct ifnet *ifp = NULL;

    ifp = card->wandev.if_head->ifp;

    if (ifp && ((SYSTEM_TICK - card->state_tick) > HOLD_DOWN_TIME)) 
    {
	wanpipe_set_state(card, WAN_CONNECTING);
	if (ppp_comm_enable(card) == CMD_OK)
	    init_ppp_tx_rx_buff(card);
    }
}

/****** Miscellaneous Functions *********************************************/

/*============================================================================
 * Configure S508 adapter.
 */
static int 
config508(ppp_private_area_t *ppp_priv_area, struct ifaddr *ifa, sdla_t *card)
{
    ppp508_conf_t cfg;
    struct ifnet *ifp = NULL;
    struct sockaddr_in *saddr = NULL;

    ifp = card->wandev.if_head->ifp;

    /* Prepare PPP configuration structure */
    bzero(&cfg, sizeof(ppp508_conf_t));

    if (card->wandev.clocking)
	cfg.line_speed = card->wandev.bps;

    if (card->wandev.interface == WANOPT_RS232)
	cfg.conf_flags |= INTERFACE_LEVEL_RS232;

    cfg.conf_flags 	|= DONT_TERMINATE_LNK_MAX_CONFIG; /*send Configure-Request packets forever*/
    cfg.txbuf_percent	= PERCENT_TX_BUFF;	/* % of Tx bufs */
    cfg.mtu_local	= card->wandev.mtu;
    cfg.mtu_remote	= card->wandev.mtu;                  /*    Default   */
    cfg.restart_tmr	= TIME_BETWEEN_CONF_REQ;  	     /*    30 = 3sec */
    cfg.auth_rsrt_tmr	= TIME_BETWEEN_PAP_CHAP_REQ;         /*    30 = 3sec */
    cfg.auth_wait_tmr	= WAIT_PAP_CHAP_WITHOUT_REPLY;       /*   300 = 30s  */
    cfg.mdm_fail_tmr	= WAIT_AFTER_DCD_CTS_LOW;            /*     5 = 0.5s */
    cfg.dtr_drop_tmr	= TIME_DCD_CTS_LOW_AFTER_LNK_DOWN;   /*    10 = 1s   */
    cfg.connect_tmout	= WAIT_DCD_HIGH_AFTER_ENABLE_COMM;   /*   900 = 90s  */
    cfg.conf_retry	= MAX_CONF_REQ_WITHOUT_REPLY;        /*    10 = 1s   */
    cfg.term_retry	= MAX_TERM_REQ_WITHOUT_REPLY;	     /*     2 times  */
    cfg.fail_retry	= NUM_CONF_NAK_WITHOUT_REPLY;        /*     5 times  */
    cfg.auth_retry	= NUM_AUTH_REQ_WITHOUT_REPLY;        /*     10 times */  
    if (!card->u.p.authenticator) 
    {
	log(LOG_INFO, "%s: Device is not configured as an authenticator\n",
						card->devname);
	cfg.auth_options = NO_AUTHENTICATION;
    }
    else
    {
	log(LOG_INFO, "%s: Device is configured as an authenticator\n", 
						card->devname);
	cfg.auth_options = INBOUND_AUTH;
    }
    if (ppp_priv_area->pap == WANOPT_YES)
    {
	cfg.auth_options |=PAP_AUTH;
	log(LOG_INFO, "%s: Pap enabled\n", card->devname);
    }
    if (ppp_priv_area->chap == WANOPT_YES)
    {
	cfg.auth_options |= CHAP_AUTH;
	log(LOG_INFO, "%s: Chap enabled\n", card->devname);
    }

    if (ppp_priv_area->enable_IPX == WANOPT_YES)
	cfg.ipx_options		= ENABLE_IPX | ROUTING_PROT_DEFAULT;
    else
	cfg.ipx_options 	= DISABLE_IPX;

    switch (card->u.p.ip_mode) 
    {
    case WANOPT_PPP_STATIC:
	cfg.ip_options		= L_AND_R_IP_NO_ASSIG | ENABLE_IP;
	/* Local IP address */
	saddr = (struct sockaddr_in*)ifa->ifa_addr;
	if (saddr)
	    cfg.ip_local = saddr->sin_addr.s_addr;
	/* Destination IP address */
	saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
	if (saddr)
	    cfg.ip_remote = saddr->sin_addr.s_addr;

	log(LOG_INFO, "%s: Local  %s\n", ifp->if_name, in_ntoa(cfg.ip_local));
	log(LOG_INFO, "%s: Remote %s\n", ifp->if_name, in_ntoa(cfg.ip_remote));
	break;

    case WANOPT_PPP_PEER:
	cfg.ip_options		= L_IP_REMOTE_ASSIG | 
				  R_IP_REMOTE_ASSIG | 
				  ENABLE_IP;
	cfg.ip_local		= 0x00;
	cfg.ip_remote		= 0x00;
	break;
    }
    return ppp_configure(card, &cfg);
}

/*============================================================================
 * Show disconnection cause.
 */
static void show_disc_cause(sdla_t *card, unsigned cause)
{
    if (cause & 0x0802)
	log(LOG_INFO, "%s: link terminated by peer\n", card->devname);
    else if (cause & 0x0004)
        log(LOG_INFO, "%s: link terminated by user\n", card->devname);
    else if (cause & 0x0008) 
	log(LOG_INFO, "%s: authentication failed\n", card->devname);
    else if (cause & 0x0010) 
	log(LOG_INFO, "%s: authentication protocol negotiation failed\n", 
							card->devname);
    else if (cause & 0x0020) 
	log(LOG_INFO, "%s: peer's request for authentication rejected\n",
							card->devname);
    else if (cause & 0x0040) 
	log(LOG_INFO, "%s: MRU option rejected by peer\n", 
							card->devname);
    else if (cause & 0x0080) 
	log(LOG_INFO, "%s: peer's MRU was too small\n", 
							card->devname);
    else if (cause & 0x0100) 
	log(LOG_INFO, "%s: failed to negotiate peer's LCP options\n",
							card->devname);
    else if (cause & 0x0200) 
	log(LOG_INFO, "%s: failed to negotiate peer's IPCP options\n",
							card->devname);
    else if (cause & 0x0400) 
	log(LOG_INFO, "%s: failed to negotiate peer's IPXCP options\n",
							card->devname);
}

/*=============================================================================
 * Process UDP call of type PTPIPEAB.
 */
static int process_udp_mgmt_pkt(sdla_t *card, struct ifnet *ifp, 
				 ppp_private_area_t *ppp_priv_area) 
{
    unsigned char buf2[5];
    unsigned int frames;
    struct mbuf *m = NULL, *m_curr = NULL;
    unsigned short data_length = 0, buffer_length = 0, real_len = 0;
    unsigned char *data_addr = NULL;
    unsigned long data_ptr;
    int udp_mgmt_req_valid = 1;
    ppp_mbox_t *mbox = card->mbox;
    struct timeval tv;
    int err = 0;
    int s = 0;
    ppp_udp_pkt_t *ppp_udp_pkt = (ppp_udp_pkt_t*)&ppp_priv_area->udp_pkt_data;

    bcopy(&card->wandev.udp_port, &buf2, 2 );

    switch(ppp_udp_pkt->cblock.command) 
    {
    case FT1_MONITOR_STATUS_CTRL:
    case PPIPE_ENABLE_TRACING:
    case PPIPE_DISABLE_TRACING:
    case PPIPE_GET_TRACE_INFO:
    case SET_FT1_MODE:
	if (ppp_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) 
	{
	    ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_direction_err;
	    udp_mgmt_req_valid = 0;
	}	 
	break;

    default:
	break;
    } 

    if (!udp_mgmt_req_valid) 
    {
	/* set length to 0 */
  	ppp_udp_pkt->cblock.length = 0x00;

	/* set return code */
        ppp_udp_pkt->cblock.result = 0xCD; 
    } 
    else 
    {
	/* Initialize the trace element */
	trace_element_t trace_element;

	switch (ppp_udp_pkt->cblock.command)
	{
   	case PPIPE_ENABLE_TRACING:
	    if (!card->TracingEnabled) 
	    {
    		/* OPERATE_DATALINE_MONITOR */
    		mbox->cmd.command = PPP_DATALINE_MONITOR;
    		mbox->cmd.length = 0x01;
   		mbox->data[0] = ppp_udp_pkt->data[0];
		err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT; 
		if (err != CMD_OK) 
		{ 
		    ppp_error(card, err, mbox);
        	    card->TracingEnabled = 0;
	       
		    /* set the return code */
	       	    ppp_udp_pkt->cblock.result = mbox->cmd.result;
        	    mbox->cmd.length = 0;
        	    break;
    		} 
		sdla_peek(&card->hw, 0xC000, &buf2, 2);

		ppp_priv_area->curr_trace_addr = 0;
   		bcopy(&buf2, &ppp_priv_area->curr_trace_addr, 2);
    		ppp_priv_area->start_trace_addr = 
					ppp_priv_area->curr_trace_addr;
		ppp_priv_area->end_trace_addr = 
				ppp_priv_area->start_trace_addr + END_OFFSET;
		    	
		/* MAX_SEND_BUFFER_SIZE - 28 (IP header) 
		 *  - 32 (ppipemon CBLOCK) 
		 */
	    	available_buffer_space = MAX_LGTH_UDP_MGNT_PKT - 
					 sizeof(ip_pkt_t)-
					 sizeof(udp_pkt_t)-
					 sizeof(wp_mgmt_t)-
					 sizeof(cblock_t);
       	    }
            ppp_udp_pkt->cblock.result = 0;
	    mbox->cmd.length = 0;
	    card->TracingEnabled = 1;
	    break;
	  
	case PPIPE_DISABLE_TRACING:
	    if (card->TracingEnabled) 
	    {
		/* OPERATE_DATALINE_MONITOR */
  		mbox->cmd.command = 0x33;
 		mbox->cmd.length = 1;
 		mbox->data[0] = 0x00;
  		err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT; 
	    } 

	    /*set return code*/
	    ppp_udp_pkt->cblock.result = 0;
	    mbox->cmd.length = 0;
	    card->TracingEnabled = 0;
	    break;

	case PPIPE_GET_TRACE_INFO:
	    if (!card->TracingEnabled) 
	    {
	  	/* set return code */
		ppp_udp_pkt->cblock.result = 1;
		mbox->cmd.length = 0;
	    }		    

	    buffer_length = 0;

	    /* frames < 62, where 62 is the number of trace
	     * information elements.  There is in total 496
	     * bytes of space and each trace information
	     * element is 8 bytes. 
	     */
	    for ( frames=0; frames<62; frames++) 
	    { 
		trace_pkt_t *trace_pkt = 
			(trace_pkt_t*)&ppp_udp_pkt->data[buffer_length];

		/* Read the whole trace packet */
		sdla_peek(&card->hw, ppp_priv_area->curr_trace_addr, 
				  &trace_element, sizeof(trace_element_t));
	
		/* no data on board so exit */
		if (trace_element.opp_flag == 0x00) 
		    break;
	      
		data_ptr = trace_element.trace_data_ptr;

		/* See if there is actual data on the trace buffer */
		if (data_ptr)
		    data_length = trace_element.trace_length;
		else
		{
		    data_length = 0;
		    ppp_udp_pkt->data[0] |= 0x02;
		}

		/* FIXME: Do we need this check */
		if ((available_buffer_space - buffer_length) < 
		    (sizeof(trace_element_t)+1))
		{
		    /* indicate we have more frames 
		     * on board and exit 
		     */
		    ppp_udp_pkt->data[0] |= 0x02;
		    break;
		}
	
		trace_pkt->status = trace_element.trace_type;
		trace_pkt->time_stamp = trace_element.trace_time_stamp;
		trace_pkt->real_length = trace_element.trace_length;

		real_len = trace_element.trace_length;	
	
		if (data_ptr == 0)
		    trace_pkt->data_avail = 0x00;
		else
		{
		    /* we can take it next time */
		    if ((available_buffer_space - buffer_length) <
				(real_len + sizeof(trace_pkt_t)))
		    {
			ppp_udp_pkt->data[0] |= 0x02;
			break;
		    } 
		    trace_pkt->data_avail = 0x01;
	
		    /* get the data */
		    sdla_peek(&card->hw, data_ptr, &trace_pkt->data, real_len);
		}	
		/* zero the opp flag to 
		 * show we got the frame 
		 */
		buf2[0] = 0x00;
		sdla_poke(&card->hw, ppp_priv_area->curr_trace_addr, &buf2, 1);

		/* now move onto the next 
		 * frame 
		 */
		ppp_priv_area->curr_trace_addr += 8;

		/* check if we passed the last address */
		if (ppp_priv_area->curr_trace_addr >= 
		  				  ppp_priv_area->end_trace_addr)
		    ppp_priv_area->curr_trace_addr = 
				ppp_priv_area->start_trace_addr;
 
		/* update buffer length and make sure its even */ 
		if (trace_pkt->data_avail == 0x01) 
		    buffer_length += real_len - 1;
 
		/* for the header */
		buffer_length += 8;

		if (buffer_length & 0x0001)
		    buffer_length += 1;
	    }

	    /* ok now set the total number of frames passed
	     * in the high 5 bits 
	     */
	    ppp_udp_pkt->data[0] |= (frames << 2);

	    /* set the data length */
	    mbox->cmd.length = buffer_length;
	    ppp_udp_pkt->cblock.length = buffer_length;

	    /* set return code */
	    ppp_udp_pkt->cblock.result = 0;
	    break;

	case PPIPE_GET_IBA_DATA:
	    mbox->cmd.length = 0x09;
	    sdla_peek(&card->hw, 0xF003, &ppp_udp_pkt->data, mbox->cmd.length);
	        
	    /* set the length of the data */
	    ppp_udp_pkt->cblock.length = 0x09;

	    /* set return code */
	    ppp_udp_pkt->cblock.result = 0x00;
	    break;

	case PPIPE_KILL_BOARD:
	    break;

	case PPIPE_FT1_READ_STATUS:
	    sdla_peek(&card->hw, 0xF020, &ppp_udp_pkt->data, 2);
	    ppp_udp_pkt->cblock.length = 2;
	    ppp_udp_pkt->cblock.result = 0;
	    mbox->cmd.length = 2;
	    break;

	case PPIPE_FLUSH_DRIVER_STATS:   
	    init_ppp_priv_struct(ppp_priv_area);
	    init_global_statistics(card);
	    mbox->cmd.length = 0;
	    break;

	case PPIPE_ROUTER_UP_TIME:
	    microtime(&tv);
	    ppp_priv_area->router_up_time = tv.tv_sec - 
					ppp_priv_area->router_start_time;
	    *(unsigned long *)&ppp_udp_pkt->data =
					ppp_priv_area->router_up_time;
	    mbox->cmd.length = 4;
	    break;

   	case FT1_MONITOR_STATUS_CTRL:
	    /* Enable FT1 MONITOR STATUS */
            if (ppp_udp_pkt->data[0] == 1) 
	    {
		if( rCount++ != 0 ) 
		{
		    ppp_udp_pkt->cblock.result = 0;
       		    mbox->cmd.length = 1;
		    break;
    		}	
	    }

      	    /* Disable FT1 MONITOR STATUS */
      	    if (ppp_udp_pkt->data[0] == 0) 
	    {
		if( --rCount != 0) 
		{
		    ppp_udp_pkt->cblock.result = 0;
		    mbox->cmd.length = 1;
		    break;
   	    	} 
      	    } 
	    break;	

	case PPIPE_DRIVER_STAT_IFSEND:
	    log(LOG_INFO, "Getting IF_SEND Drivers Statistics\n");
	    bcopy(&ppp_udp_pkt->data, &ppp_priv_area->if_send_stat,   
						sizeof(if_send_stat_t));

	    ppp_udp_pkt->cblock.result = 0;
	    ppp_udp_pkt->cblock.length = sizeof(if_send_stat_t);
	    mbox->cmd.length = sizeof(if_send_stat_t);	
	    break;

	case PPIPE_DRIVER_STAT_INTR:
	    bcopy(&card->statistics, &ppp_udp_pkt->data,  
				sizeof(global_stats_t));
            bcopy(&ppp_priv_area->rx_intr_stat,
	    	  		&ppp_udp_pkt->data+sizeof(global_stats_t),
				sizeof(rx_intr_stat_t));
	    ppp_udp_pkt->cblock.result = 0;
	    ppp_udp_pkt->cblock.length = sizeof(global_stats_t)+
				         sizeof(rx_intr_stat_t);
	    mbox->cmd.length = ppp_udp_pkt->cblock.length;
	    break;

	case PPIPE_DRIVER_STAT_GEN:
	    bcopy(&ppp_priv_area->pipe_mgmt_stat, &ppp_udp_pkt->data, 
					sizeof(pipe_mgmt_stat_t));
	    bcopy(&card->statistics,
	          &ppp_udp_pkt->data+sizeof(pipe_mgmt_stat_t),
		  sizeof(global_stats_t)); 
	    ppp_udp_pkt->cblock.result = 0;
	    ppp_udp_pkt->cblock.length = sizeof(global_stats_t)+
				     sizeof(rx_intr_stat_t);
	    mbox->cmd.length = ppp_udp_pkt->cblock.length;
	    break;

	default:
	    /* it's a board command */
	    mbox->cmd.command = ppp_udp_pkt->cblock.command;
	    mbox->cmd.length = ppp_udp_pkt->cblock.length;

	    if (mbox->cmd.length) 
		bcopy((unsigned char *)ppp_udp_pkt->data, &mbox->data,
						       mbox->cmd.length);

	    /* run the command on the board */
	    err = sdla_exec(mbox) ? mbox->cmd.result : CMD_TIMEOUT;
	    if (err != CMD_OK) 
	    {
		ppp_error(card, err, mbox);
	      ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_adptr_cmnd_timeout;
		break;
	    }

  	    ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_adptr_cmnd_OK;

	    /* copy the result back to our buffer */
	    bcopy(mbox, &ppp_udp_pkt->cblock, sizeof(cblock_t));

	    if (mbox->cmd.length)
		bcopy(&mbox->data, &ppp_udp_pkt->data, mbox->cmd.length);

	} /* end of switch */
    } /* end of else */

    /* Fill UDP TTL */
    ppp_udp_pkt->ip_pkt.ttl = card->wandev.ttl; 
    data_length = reply_udp(ppp_priv_area->udp_pkt_data, mbox->cmd.length);

    if (ppp_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) 
    {
	++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_passed_to_adptr;
	ppp_send(card, ppp_priv_area->udp_pkt_data, data_length, ppp_priv_area->protocol);
    } 
    else 
    {	/* Pass it up the stack */
	/*
	 * Ok, what we've got now in scp->umg_top_buff is a fully formatted IP 
	 * packet which consists of responce to the UDP management request. The 
	 * length of the packet is 'data_length'. We have to free the original 
	 * mbuf and allocate the new one. ( when i tried to re-use the same 
	 * mbuf with m_pullup i ended up crashing the system somewhere in IP
	 * stack after i passed the frame up - looks like they free the buffer
	 * anyways if the driver does not ).
	 */  
        data_addr  = ppp_priv_area->udp_pkt_data;	

        /*
         * Build a new mbuf chain 
         */

        /* get the first ( header ) mbuf */
        MGETHDR(m, M_DONTWAIT, MT_DATA);
        if (m == NULL) 
        {
	    ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_no_socket;
	    log(LOG_ERR, "%s: no socket buffers available!\n", card->devname);
            /* i guess nothing can be done ... the packet is gone */
            return ENOMEM;
        }

        /* fill in the header */
        m->m_pkthdr.rcvif 	= ifp;		/* our interface */
        m->m_pkthdr.len	= data_length;	/* length of the data */
        m->m_len		= MHLEN;	/* well, at least as of now */

        /* m_curr will be running through the chain. m points to 
         * the header 
         */
        m_curr = m; 

        /* let's start filling in the chain */
        while(data_length > 0) 
        {
            /* check if we can use cluster instead ... */
            if (data_length > MINCLSIZE) 
            {
                /* all right - get the cluster */
                MCLGET(m_curr, M_DONTWAIT);
                if (m_curr->m_flags & M_EXT) 
                {
                    /* all right - it worked - set the new length */
                    m_curr->m_len = MCLBYTES;
                } 
	        else 
                {
                    /* it failed - well, screw it - what can we do ? */
                    m_freem(m);
                    log(LOG_ERR, "%s: Failed to allocate cluster !!!\n", 
							ifp->if_name);
                    return ENOMEM;
                }
            }

            /* 
             * ok, at this point we've got mbuf pointed by 'm_curr' which 
             * is either a regular one, or a cluster depending on how much
             * data do we have left. The beauty is that from here we dont 
             * care what kind of mbuf it is since all the macros work 
             * correctly on both.
             */
            m_curr->m_len = min(m_curr->m_len, data_length);
  
            /* 
             * Now, we can start copying data into mbuf data area given
             * by mtod() macro. We copy m_curr->m_len bytes.
             */
            bcopy(data_addr, mtod(m_curr, caddr_t), m_curr->m_len);
  
            /* adjust data pointer and remaining length */
            data_addr += m_curr->m_len;
            data_length -= m_curr->m_len;
  
            /* done with this mbuf - allocate the next one if needed */
            if (data_length > 0) 
            {
                MGET(m_curr->m_next, M_DONTWAIT, MT_DATA);
                if (m_curr->m_next == NULL) 
                {
                    /* it failed - well, screw it - we cant do much here */
                    m_freem(m);
                    log(LOG_ERR, "%s: Failed to allocate mbuf !!!\n", ifp->if_name);
                    return ENOMEM;
                }
                /* set m_curr to point to the new one, and set up initial length
                 */
		m_curr = m_curr->m_next;
                m_curr->m_len = MLEN;
            }
        } /* while ... */

        /* 
         * Ok, at this point we've got an mbuf chain pointed by m which contains
         * a UDP management reply - now we just have to pass it up the stack -
         * and that's it !!!
         * Note that we neither report those packets to BPF nore update any 
         * stats. Those packets are meant to be invisible ;-)
         * Also we run inside splimp already - do there's no need to call it 
         * again.
	 */
	s = splimp();
        schednetisr(NETISR_IP);
        if (IF_QFULL(&ipintrq)) 
        {
            /* oh, no - IP queue is full - well - screw it. We cant do much 
             * here. 
	     */ 
            log(LOG_ERR, "%s: IP queue is full, drop the packet!!!\n", ifp->if_name);
            IF_DROP(&ipintrq);
            m_freem(m);
            return -EINVAL;
        }

        /* ok enqueue the packet */
        IF_ENQUEUE(&ipintrq, m);
	splx(s);
	dlog((LOG_DEBUG, "%s: UDP packet passed back on IP's queue\n", 
								card->devname));
	++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_passed_to_stack;
    }

    ppp_priv_area->udp_pkt_lgth = 0;
    return 0; 
}

/*=============================================================================
 * Initial the ppp_private_area structure.
 */
static void init_ppp_priv_struct( ppp_private_area_t *ppp_priv_area )
{
    bzero(&ppp_priv_area->if_send_stat, sizeof(if_send_stat_t));
    bzero(&ppp_priv_area->rx_intr_stat, sizeof(rx_intr_stat_t));
    bzero(&ppp_priv_area->pipe_mgmt_stat, sizeof(pipe_mgmt_stat_t));
}

/*============================================================================
 * Initialize Global Statistics
 */
static void init_global_statistics( sdla_t *card )
{
    bzero(&card->statistics, sizeof(global_stats_t));
}

/*============================================================================
 * Initialize Receive and Transmit Buffers.
 */
static void init_ppp_tx_rx_buff( sdla_t *card )
{
    ppp508_buf_info_t* info = NULL;

    if (card->hw.type == SDLA_S514) 
    {
        info = (void*)((u_char*)card->hw.dpmbase + PPP514_BUF_OFFS);
        card->u.p.txbuf_base = (void*)((u_char*)card->hw.dpmbase + 
							info->txb_ptr);
        card->u.p.txbuf_last = (ppp_buf_ctl_t*)card->u.p.txbuf_base +
							(info->txb_num - 1);
        card->u.p.rxbuf_base = (void*)((u_char*)card->hw.dpmbase +
				                      info->rxb_ptr);
        card->u.p.rxbuf_last = (ppp_buf_ctl_t*)card->u.p.rxbuf_base +
		                        (info->rxb_num - 1);
    } 
    else 
    {
        info = (void*)((u_char*)card->hw.dpmbase + PPP508_BUF_OFFS);
        card->u.p.txbuf_base = (void*)((u_char*)card->hw.dpmbase +
				(info->txb_ptr - PPP508_MB_VECT));
        card->u.p.txbuf_last = (ppp_buf_ctl_t*)card->u.p.txbuf_base +
					(info->txb_num - 1);
        card->u.p.rxbuf_base = (void*)((u_char*)card->hw.dpmbase +
				(info->rxb_ptr - PPP508_MB_VECT));
        card->u.p.rxbuf_last = (ppp_buf_ctl_t*)card->u.p.rxbuf_base +
						(info->rxb_num - 1);
    }

    card->u.p.rx_base = info->rxb_base;
    card->u.p.rx_top  = info->rxb_end;
    card->u.p.txbuf = card->u.p.txbuf_base;
    card->rxmb = card->u.p.rxbuf_base;
}

/*=============================================================================
 * Read Connection Information (ie for Remote IP address assginment).
 * Called when ppp interface connected.
 */
static int read_info( sdla_t *card )
{
    struct ifnet *ifp = NULL;
    struct ifaddr *ifa = NULL; 
    struct sockaddr_in *saddr = NULL;
    ppp_private_area_t *ppp_priv_area = NULL;
    int err = 0;

    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;

    ifa = TAILQ_FIRST(&ifp->if_addrhead); 
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL)
    {
	/* Local IP address */
	saddr = (struct sockaddr_in*)ifa->ifa_addr;
	if (saddr)
	{
	    /* write new local IP address */
	    saddr->sin_addr.s_addr = ppp_priv_area->ip_local;
	}
	/* Remote IP address */
	saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
	if (saddr)
	{
	    /* write new remote IP address */
	    saddr->sin_addr.s_addr = ppp_priv_area->ip_local;
	}
	/* Add a new route instead od deleted one */
	err = rtrequest((int)RTM_ADD,			/* action      */
		  ifa->ifa_dstaddr,			/* destination */
		  ifa->ifa_addr,			/* gateway     */
		  ifa->ifa_netmask,			/* mask        */
		  RTF_HOST,				/* flag	       */ 	
	          (struct rtentry**)0);
    }

    if (!err)
    {
	log(LOG_INFO, "%s: Assigned Lcl. Addr: %s\n", 
			card->devname, in_ntoa(ppp_priv_area->ip_local));
	log(LOG_INFO, "%s: Assigned Rmt. Addr: %s\n", 
			card->devname, in_ntoa(ppp_priv_area->ip_local));
    } 
    else
    {
	log(LOG_ERR, "%s: Adding of route failed:\n", card->devname);
	log(LOG_ERR, "%s:	Local : %s\n", 
			card->devname,in_ntoa(ppp_priv_area->ip_local));
	log(LOG_ERR, "%s:	Remote: %s\n",
			card->devname,in_ntoa(ppp_priv_area->ip_remote));
	log(LOG_ERR, "%s:   An error occurred in IP assignment.\n", 
							card->devname);
    }
    return 0;
}


/*=============================================================================
 * Remove Dynamic Route.
 * Called when ppp interface disconnected.
 */
static int remove_route( sdla_t *card )
{
    /* 
     * ok, i was too lazy to allocate new sockaddr structures - so i'm 
     * going to temporary replace interface ip addresses, then call delete 
     * route and then restore them back.
     * Note that all this is done only if old and new addresses are different.
     */
    ppp_private_area_t *ppp_priv_area = NULL;
    struct ifnet *ifp = NULL;
    struct ifaddr	      *ifa = NULL;
    int err = 0;

    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;
    ifa = TAILQ_FIRST(&ifp->if_addrhead);

    /*  get the next one - its where all the stuff is */
    if ((ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL) 
    {
        /* if the entry is not found. we'll give it another try then 	  */
        if ((err = rtrequest((int)RTM_DELETE, 		/* action */
			 ifa->ifa_dstaddr,		/* destination */
			 ifa->ifa_addr,		/* gateway */
			 ifa->ifa_netmask,		/* mask */
			 RTF_HOST,			/* flag */
			 (struct rtentry **)0)) == 0)
        {
	    log(LOG_INFO, "%s: PPP deleting dynamic route successfuly!\n",
					card->devname); 
        }
	else
	    log(LOG_ERR, "%s: Deleting dynamic route failed %d!\n",
					card->devname, err);
    }
    return err;
}



/*
 * ============================================================================
 * Perform the Interrupt Test by running the READ_CODE_VERSION command MAX_INTR
 * _TEST_COUNTER times.
 */
static int ppp_intr_test( sdla_t *card )
{
    ppp_mbox_t *mb = card->mbox;
    ppp_flags_t *flags = card->flags;
    int intr_mode = 0, intr_permission = 0;
    int err,i;

    Intr_test_counter = 0;

    intr_permission = flags->imask;
    intr_mode	    = ppp_read_intr_mode(card);
 
    err = ppp_set_intr_mode(card, PPP_INTR_CMD);
    if (err == CMD_OK) 
    { 
	for (i = 0; i < MAX_INTR_TEST_COUNTER; i ++) 
	{	
	    /* Run command READ_CODE_VERSION */
	    bzero(&mb->cmd, sizeof(ppp_cmd_t));
	    mb->cmd.length  = 0;
	    mb->cmd.command = PPP_READ_CODE_VERSION;
	    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
	    if (err != CMD_OK) 
	        ppp_error(card, err, mb);
	}
    	err = ppp_set_intr_mode(card, intr_mode);
	flags->imask = intr_permission;
    }

    if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER))
    {
	log(LOG_ERR, "%s: Interrupt Test Failed, Counter: %d\n",
					card->devname, Intr_test_counter);
	log(LOG_ERR, "%s: Please choose another interrupt!!\n", 
					card->devname); 
	return -EIO;
    }
    log(LOG_INFO, "%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
    card->configured = 1;
 
    return 0;
}

/*==============================================================================
 * Determine what type of UDP call it is. DRVSTATS or PTPIPEAB ?
 */
static int udp_pkt_type( struct mbuf *m0, sdla_t *card )
{
    unsigned char *sendpacket;
    unsigned char buf2[5]; 
    /* FIXME: Use the structure */
    ppp_udp_pkt_t *ppp_udp_pkt = mtod(m0, ppp_udp_pkt_t *); 

    sendpacket = mtod(m0, char*);
    bcopy(&card->wandev.udp_port, &buf2, 2);

    if (ppp_udp_pkt->ip_pkt.ver_inet_hdr_length  == 0x45 &&     /* IP packet */ 
	sendpacket[9]  == 0x11 &&        /* UDP packet */
	sendpacket[22] == buf2[1] &&     /* UDP Port */
	sendpacket[23] == buf2[0] &&
	sendpacket[36] == 0x01 ) 
    {
	if (sendpacket[28] == 0x50 &&    /* PTPIPEAB: Signature */ 
	    sendpacket[29] == 0x54 &&      
	    sendpacket[30] == 0x50 &&    
	    sendpacket[31] == 0x49 &&   
	    sendpacket[32] == 0x50 &&  
	    sendpacket[33] == 0x45 && 
	    sendpacket[34] == 0x41 &&
	    sendpacket[35] == 0x42 )
	{ 
	    return UDP_PTPIPE_TYPE;
	} 
	else 
  	    if (sendpacket[28] == 0x44 &&  /* DRVSTATS: Signature */
	        sendpacket[29] == 0x52 &&      
      		sendpacket[30] == 0x56 &&      
      		sendpacket[31] == 0x53 &&    
      		sendpacket[32] == 0x54 &&   
      		sendpacket[33] == 0x41 && 
      		sendpacket[34] == 0x54 &&      
      		sendpacket[35] == 0x53 )
	    {
		return UDP_DRVSTATS_TYPE;
	    } 
	    else
		return UDP_INVALID_TYPE;
    } 
    else
	return UDP_INVALID_TYPE;

}

/*============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp, struct mbuf *m)
{
    uint32_t src_ip_addr = 0;
    uint32_t broadcast_ip_addr = 0;
    struct ifaddr *ifa = TAILQ_FIRST(&ifp->if_addrhead);
    struct sockaddr_in *saddr = NULL;

    /* read the IP source address from the outgoing packet */
    src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 12);

    /* read the IP broadcast address for the device */
    if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL) 
        broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);

    /* check if the IP Source Address is a Broadcast address */
    if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)) 
    {
        log(LOG_INFO, "%s: Broadcast Source Address silently discarded\n",
							card->devname);
        m_freem(m);
        ++ card->wandev.stats.tx_dropped;
        return 1;
     } 

     /* check if the IP Source Address is a Multicast address */
     if ((ntohl(src_ip_addr) >= 0xE0000001) && 
		 (ntohl(src_ip_addr) <= 0xFFFFFFFE))
     {
         log(LOG_INFO, "%s: Multicast Source Address silently discarded\n", 
					card->devname);
         m_freem(m);
	 card->wandev.stats.multicast++;
         card->wandev.stats.tx_dropped++;
         return 1;
     }

    return 0;
}

/****** PPP Firmware Interface Functions ************************************/

/*============================================================================
 * Read firmware code version.
 *	Put code version as ASCII string in str. 
 */
static int ppp_read_version(sdla_t *card, char *str)
{
    ppp_mbox_t *mb = card->mbox;
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    mb->cmd.command = PPP_READ_CODE_VERSION;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;

    if (err != CMD_OK)
	ppp_error(card, err, mb);
    else 
	if (str) 
	{
	    int len = mb->cmd.length;

	    bcopy(mb->data, str, len);
	    str[len] = '\0';
	}

    return err;
}

/*===========================================================================
 * Set Out-Bound Authentication.
*/
static int ppp_set_outbnd_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    ppp_mbox_t *mb = card->mbox;
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    bzero(&mb->data, (strlen(ppp_priv_area->userid) + 
				strlen(ppp_priv_area->passwd) + 2 ) );
    bcopy(ppp_priv_area->userid, mb->data, strlen(ppp_priv_area->userid));
    bcopy(ppp_priv_area->passwd, (mb->data + strlen(ppp_priv_area->userid) + 1),
					strlen(ppp_priv_area->passwd));	
    mb->cmd.length  = strlen(ppp_priv_area->userid) + 
				strlen(ppp_priv_area->passwd) + 2 ;
    mb->cmd.command = PPP_SET_OUTBOUND_AUTH;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
    if (err != CMD_OK)
	ppp_error(card, err, mb);

    return err;
}


/*===========================================================================
 * Set In-Bound Authentication.
*/
static int ppp_set_inbnd_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    ppp_mbox_t *mb = card->mbox;
    int err, i;
    char* user_tokens[32];
    char* pass_tokens[32];
    int userids, passwds;
    int add_ptr;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    bzero(&mb->data, 1008);
    bcopy(ppp_priv_area->sysname, mb->data,  
					strlen(ppp_priv_area->sysname));
	
    /* Parse the userid string and the password string and build a string
     * to copy it to the data area of the command structure.   The string
     * will look like "SYS_NAME<NULL>USER1<NULL>PASS1<NULL>USER2<NULL>PASS2
     * ....<NULL> " 
     */
    userids = tokenize( ppp_priv_area->userid, user_tokens);
    passwds = tokenize( ppp_priv_area->passwd, pass_tokens);

    if (userids != passwds)
    {
     log(LOG_ERR, "%s: Number of passwords does not equal the number of user ids\n",
							 card->devname);
	return 1;
    }

    add_ptr = strlen(ppp_priv_area->sysname) + 1;
    for (i=0; i<userids; i++)
    {
	bcopy(user_tokens[i], (mb->data + add_ptr),  
					strlen(user_tokens[i]));
	bcopy(pass_tokens[i], (mb->data + add_ptr + strlen(user_tokens[i]) + 1),
					strlen(pass_tokens[i]));
	add_ptr = add_ptr + strlen(user_tokens[i]) + 1 + 
					strlen(pass_tokens[i]) + 1;
    }

    mb->cmd.length  = add_ptr + 1;
    mb->cmd.command = PPP_SET_INBOUND_AUTH;

    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;

    if (err != CMD_OK)
	ppp_error(card, err, mb);

    return err;
}


static char *my_strtok(char *p, const char *tok)
{
    static char *phead = NULL, *tmp = NULL;
    char *tk = NULL;

    if (p)
    {
	if (phead)
	    free(phead, M_DEVBUF);
	phead = malloc(strlen(p)+1, M_DEVBUF, M_NOWAIT);
	if (phead == NULL)
	    return NULL;
	bcopy(p, phead, strlen(p));
	tmp = phead;
    }
    for(;; ++tmp)
    {
	tk = (char*)tok;
	for(;*tk != '\0'; ++tk)
	    if (*tmp == *tk)
		return tmp;
	if (!*tmp)
	    return NULL; 
    }
	
}

/*
 * ============================================================================
 * Tokenize string.
 *      Parse a string of the following syntax:
 *              <arg1>,<arg2>,...
 *      and fill array of tokens with pointers to string elements.
 *
 */
static int tokenize (char *str, char **tokens)
{
    char* eos = NULL;
    int cnt = 0;

    if (str == NULL || str[0] == '\0')
	return 0;
    tokens[cnt] = str;
    if (*tokens[cnt] == '/')
	tokens[cnt]++;
    while (tokens[cnt] && (cnt < 32 - 1))
    {
	eos = my_strchr(tokens[cnt], '/');
	tokens[++cnt] = eos;
	if (eos){
	    tokens[cnt] = eos + 1;
	    *eos = '\0';
	}
        /*tokens[cnt] = strstrip(tokens[cnt], " \t");*/
    }
    return cnt;
}

/*
 * ============================================================================
 *
 */
static char *my_strchr(const char *p, char ch)
{
    for(;p; ++p)
    {
	if (*p == ch)
	    return ((char*)p);
	if (!(*p))
	    return NULL;
    }
    return NULL;
}

/*============================================================================
 * Strip leading and trailing spaces off the string str.
 */
static char* strstrip (char *str, char* s)
{
    char *eos = str + strlen(str);          /* -> end of string */

    while (*str && my_strchr(s, *str))
        ++str;                           /* strip leading spaces */
    
    while ((eos > str) && my_strchr(s, *(eos - 1)))
        --eos;                           /* strip trailing spaces */
     
    *eos = '\0';
    return str;
}

/*============================================================================
 * Configure PPP firmware.
 */
static int ppp_configure(sdla_t *card, void *data)
{
    ppp_mbox_t *mb = card->mbox;
    int data_len = sizeof(ppp508_conf_t); 
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    bcopy(data, mb->data, data_len);
    mb->cmd.length  = data_len;
    mb->cmd.command = PPP_SET_CONFIG;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;

    if (err != CMD_OK) 
	ppp_error(card, err, mb);
	
    return err;
}

/*============================================================================
 * Set interrupt mode.
 */
static int ppp_set_intr_mode(sdla_t *card, unsigned char mode)
{
    ppp_mbox_t *mb = card->mbox;
    ppp_intr_info_t *ppp_intr_data = (ppp_intr_info_t *) &mb->data[0];
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    ppp_intr_data->i_enable = mode;

    ppp_intr_data->irq = card->hw.irq;
    mb->cmd.length = 2;

    /* If timer has been enabled, set the timer delay to 1sec */
    if (mode & 0x80)
    {
   	ppp_intr_data->timer_len = 5;/* 100; 250; */
        mb->cmd.length = 4;
    }

    mb->cmd.command = PPP_SET_INTR_FLAGS;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;

    if (err != CMD_OK) 
	ppp_error(card, err, mb);
 
    return err;
}

/*
 * ===========================================================================
 * Read interrupt mode
 *
 */
static int ppp_read_intr_mode(sdla_t* card)
{
    ppp_mbox_t *mb = card->mbox;
    int err = 0;
 
    mb->cmd.command = PPP_SET_INTR_FLAGS;
    mb->cmd.length  = 0x0;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
    if (err != CMD_OK) 
	ppp_error(card, err, mb);
    else
    {
        ppp_intr_info_t *ppp_intr_data = (ppp_intr_info_t *) &mb->data[0];
 	return ppp_intr_data->i_enable;	
    }
    return err;
}
 
/*============================================================================
 * Enable communications.
 */
static int ppp_comm_enable(sdla_t *card)
{
    ppp_mbox_t *mb = card->mbox;
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    mb->cmd.command = PPP_COMM_ENABLE;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
	
    if (err != CMD_OK)
    {
	if (err == PPPRES_COMMS_ENABLED)
	    dlog((LOG_DEBUG, "%s: Communication already enabled\n", card->devname));
	else
	    ppp_error(card, err, mb);
    }
    else
	dlog((LOG_DEBUG, "%s: Communication enabled.\n", card->devname));
	
    return err;
}

/*============================================================================
 * Disable communications.
 */
static int ppp_comm_disable(sdla_t *card)
{
    ppp_mbox_t *mb = card->mbox;
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    mb->cmd.command = PPP_COMM_DISABLE;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
    if (err != CMD_OK) 
	ppp_error(card, err, mb);
    else 
	dlog((LOG_DEBUG, "%s: Communication disabled.\n", card->devname));
    return err;
}

/*============================================================================
 * Get communications error statistics.
 */
static int ppp_get_err_stats(sdla_t *card)
{
    ppp_mbox_t *mb = card->mbox;
    int err;

    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    mb->cmd.command = PPP_READ_ERROR_STATS;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;
	
    if (err == CMD_OK) 
    {
	ppp_err_stats_t* stats = (void*)mb->data;
	card->wandev.stats.rx_over_errors    = stats->rx_overrun;
	card->wandev.stats.rx_crc_errors     = stats->rx_bad_crc;
	card->wandev.stats.rx_missed_errors  = stats->rx_abort;
	card->wandev.stats.rx_length_errors  = stats->rx_lost;
	card->wandev.stats.tx_aborted_errors = stats->tx_abort;
    } 
    else 
	ppp_error(card, err, mb);
	
    return err;
}

static int read_connection_info (sdla_t *card)
{
    ppp_mbox_t *mb = card->mbox;
    struct ifnet *ifp = NULL;
    ppp_private_area_t *ppp_priv_area = NULL;
    ppp508_connect_info_t *ppp508_connect_info = NULL;
    int err;

    ifp = card->wandev.if_head->ifp;
    ppp_priv_area = ifp->if_softc;
    bzero(&mb->cmd, sizeof(ppp_cmd_t));
    mb->cmd.length  = 0;
    mb->cmd.command = PPP_GET_CONNECTION_INFO;
    err = sdla_exec(mb) ? mb->cmd.result : CMD_TIMEOUT;

    if (err != CMD_OK)
	ppp_error(card, err, mb);
    else 
    {
	ppp508_connect_info = (ppp508_connect_info_t *)mb->data;
	ppp_priv_area->ip_remote = ppp508_connect_info->ip_remote;
	ppp_priv_area->ip_local = ppp508_connect_info->ip_local;
    }
    return err;
}

/*============================================================================
 * Send packet.
 *	Return:	0 - o.k.
 *		1 - no transmit buffers available
 */
static int ppp_send (sdla_t *card, void *data, unsigned len, unsigned proto)
{
    ppp_buf_ctl_t *txbuf = card->u.p.txbuf;

    if (txbuf->flag)
        return 1;
	
    sdla_poke(&card->hw, txbuf->buf.ptr, data, len);

    txbuf->length = len;		/* frame length */
	
    if (proto == htons(ETHERTYPE_IPX))
	txbuf->proto = 0x01;	/* protocol ID */
    else
	txbuf->proto = 0x00;	/* protocol ID */

    txbuf->flag = 1;		/* start transmission */

    /* Update transmit buffer control fields */
    card->u.p.txbuf = ++txbuf;

    if ((void*)txbuf > card->u.p.txbuf_last)
	card->u.p.txbuf = card->u.p.txbuf_base;

    return 0;
}

/****** Firmware Error Handler **********************************************/

/*============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int ppp_error(sdla_t *card, int err, ppp_mbox_t *mb)
{
    unsigned cmd = mb->cmd.command;

    switch (err) 
    {
    case CMD_TIMEOUT:
	log(LOG_ERR, "%s: command 0x%02x timed out!\n", card->devname, cmd);
	break;

    default:
	log(LOG_ERR, "%s: command 0x%02x returned 0x%02x!\n",
						   card->devname, cmd, err);
    }
    return 0;
}


/****** End *****************************************************************/
