/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdlamain.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
* sdlamain.c	WANPIPE(tm) Multiprotocol WAN Link Driver.  Main module.
*
* Author:	Alex Feldman    <al.feldman@sangoma.com> 
*
* ===========================================================================
* Nov 20, 1999	Alex Feldman 	Updated to FreeBSD 3.2
* Jan 02, 1997	Gene Kozin	Initial version.
*****************************************************************************/


#include "bpfilter.h"
#include "sdlamain.h"

#include <stddef.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/interrupt.h>

#include <net/sdlapci.h>
#include <net/sdla_dev.h> 	
#include <net/wanpipe.h>	/* WANPIPE common user API definitions */

#include <machine/types.h>
#include </usr/src/sys/i386/isa/isa_device.h>

/****** Defines & Macros ****************************************************/
/* For debugging convenience */
#ifdef WANDEBUG
#define dlog(x)	log x
#else
#define dlog(x)
#endif /* WANDEBUG */

#define	DRV_VERSION	1		/* version number */
#define	DRV_RELEASE	1		/* release (minor version) number */

#define	MAX_CARDS	8		/* max number of adapters */
#ifndef	CONFIG_WANPIPE_CARDS		/* configurable option */
 #define CONFIG_WANPIPE_CARDS 1
#endif
 
#define	CMD_OK		0		/* normal firmware return code */
#define	CMD_TIMEOUT	0xFF		/* firmware command timed out */
#define	MAX_CMD_RETRY	10		/* max number of firmware retries */

#define SYSTEM_TICK ticks
/****** Function Prototypes *************************************************/

/* Loadable kernel module function interface  */
static int wanpipe_load __P((module_t, int, void*));

/* WAN link driver entry points */
static int setup    (wan_device_t* wandev, wandev_conf_t* conf);
static int shutdown (wan_device_t* wandev);
static int ioctl    (wan_device_t* wandev, u_long cmd, caddr_t arg);

/* IOCTL handlers */
static int ioctl_dump	(sdla_t* card, sdla_dump_t* u_dump);
static int ioctl_exec	(sdla_t* card, sdla_exec_t* u_exec);

/* ISA/PCI interrupt handler */
static void 	     sdla_isr(void *arg);
static int 	     set_intrhand(sdla_t *card);
static int 	     restore_intrhand(sdla_t *card);

/* Miscellaneous functions */
static void release_hw(sdla_t *card);

/* SDLA ISA/PCI varibles */
extern int number_cards;		/* total number of SDLA cards */
extern sdladev_t sdladev[];		/* SDLA info structure 	      */
extern unsigned int number_S514_cards;  /* total number of S514 cards */

/* SDLA ISA/PCI device functions */
static sdladev_t *find_adapter(unsigned, unsigned, int, char*);
static int update_isa_resource(wan_device_t *, wandev_conf_t *);
static int update_pci_resource(sdla_t *card, wandev_conf_t *conf);

/****** Global Data **********************************************************
 * Note: All data must be explicitly initialized!!!
 */
/* private data */
static char modname[]	= "wanpipe";
static char drvname[]	= "wanpipe";

static char fullname[]	= "WANPIPE(tm) Multiprotocol Driver";
static char copyright[]	= "(c) 1995-1999 Sangoma Technologies Inc.";

static int ncards = CONFIG_WANPIPE_ROUTERS;
static int active = 0;			/* number of active cards */
static sdla_t* card_array = NULL;	/* adapter data space */

/******* Kernel Loadable Module Entry Points ********************************/
/* 
 *============================================================================
 * Module entry point.
 *
 * Return:	0	Ok
 *	  	  < 0	error.
 * Context:	process
 */
int wanpipe_load (module_t mod, int cmd, void *arg)
{
    int cnt = 0, err = 0, i = 0;

    switch(cmd)
    {
    case MOD_LOAD:
        /*
         * 'insert' entry point.
         *  - print announcement
         *  - allocate adapter data space
         *  - initialize static data
         *  - calibrate SDLA shared memory access delay.
         */
        log(LOG_INFO, "%s v%u.%u %s\n", 
			fullname, DRV_VERSION, DRV_RELEASE, copyright); 
        /* Verify number of cards and allocate adapter data space */
	/*
	 *ncards = min(ncards, MAX_CARDS);
	 *ncards = max(ncards, 1);
	 */
	card_array = malloc(sizeof(sdla_t)*ncards, M_DEVBUF, M_NOWAIT);
	if (card_array == NULL)
	    return -ENOMEM;
	bzero(card_array, sizeof(sdla_t)*ncards); 
        /* Register adapters with WAN router */
        for (cnt = 0; cnt < ncards; ++ cnt) 
        {
            sdla_t* card  	   = &card_array[cnt];
            wan_device_t* wandev   = &card->wandev;

            sprintf(card->devname, "%s%d", drvname, cnt+1);
	    card->unit 	  	   = cnt;
	    card->next 	  	   = NULL;
            wandev->magic          = ROUTER_MAGIC;
            wandev->name           = card->devname;
            wandev->private        = card;
            wandev->enable_tx_int  = 0;
            wandev->setup          = &setup;
            wandev->shutdown       = &shutdown;
            wandev->ioctl          = &ioctl;
            err = register_wan_device(wandev);
            if (err)
            {
                log(LOG_ERR, "%s: %s registration failed with error %d!\n",
						 drvname, card->devname, err);
                break;
            }
        }
        if (cnt)
            ncards = cnt;	/* adjust actual number of cards */
        else
	{
	    free(card_array, M_DEVBUF);
	    log(LOG_ERR, "*** Init Module: No cards registered.\n"); 
	    err = -ENODEV;
	}
        break;
    case MOD_UNLOAD:
	/*
	 * 'remove' entry point.
	 *   - release all remaining system resources
	*/
	for(i = 0; i < ncards; i++)
	{
	   sdla_t *card = &card_array[i];
	   unregister_wan_device(card->devname);
	}
	free(card_array, M_DEVBUF);
	break;
    default :
	err = EINVAL; /* unknown command */
	break;
    }
    return err;
}
/* Now declare the module to the system */
static moduledata_t wanpipe_mod = {
	"wanpipe",
	wanpipe_load,
	0
};
DECLARE_MODULE(wanpipe, wanpipe_mod, SI_SUB_KLD, SI_ORDER_MIDDLE);

/******* WAN Device Driver Entry Points *************************************/

/* 
 * ============================================================================
 * Setup/configure WAN link driver.
 * o check adapter state
 * o make sure firmware is present in configuration
 * o make sure I/O port and IRQ are specified
 * o make sure I/O region is available
 * o allocate interrupt vector
 * o setup SDLA hardware
 * o call appropriate routine to perform protocol-specific initialization
 * o mark I/O region as used
 * o if this is the first active card, then schedule background task
 *
 * This function is called when router handles ROUTER_SETUP IOCTL. The
 * configuration structure is in kernel memory (including extended data, if
 * any).
 */
 
static int setup (wan_device_t* wandev, wandev_conf_t* conf)
{
    sdla_t* card = NULL;
    int err = 0;
    int irq = 0;
    int i;
    int s = splimp();

    dlog((LOG_DEBUG, "%s: Starting device setup...\n", wandev->name));	
    /* Sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL) || (conf == NULL))
    {
	splx(s);
        return -EFAULT;
    }
	
    card = wandev->private;
    if (wandev->state != WAN_UNCONFIGURED)
    {
	log(LOG_ERR, "%s: Device already configured!\n", wandev->name);
	splx(s);
	return -EBUSY;		/* already configured */
    }

    log(LOG_INFO, "\nProcessing WAN device %s...\n", wandev->name);

    if (!conf->data_size || (conf->data == NULL)) 
    {
        log(LOG_ERR, "%s: firmware not found in configuration data!\n", 
						wandev->name);
	splx(s);
	return -EINVAL;
    }

    /* only check I/O port and IRQ if not an S514 adapter */
    if (!conf->S514_CPU_no[0]) 
    {
        if (conf->ioport <= 0) 
        {
            log(LOG_ERR, "%s: can't configure without IO port address!\n",
							wandev->name);
	    splx(s);
            return -EINVAL;
        }
	
        if (conf->irq <= 0)
	{ 
	    log(LOG_ERR, "%s: can't configure without IRQ!\n", wandev->name);
	    splx(s);
            return -EINVAL;
        }

	/* Try to find corresponding isa_device structure and read 
	 * resource value to wandev structure. If ISA device is not 
	 * found exit!! (device is not found by kernel
	 */
	if (update_isa_resource(wandev, conf))
	{
	    log(LOG_ERR, "%s: Driver failed to found current ISA device!\n",
						wandev->name);
	    log(LOG_ERR, "Check your kernel configuration or %s.conf has the\n",
						wandev->name);
	    log(LOG_ERR, "same resources as your kernel configuration!\n");
	    splx(s);
	    return -EINVAL;
	}
	/* 
	 * Check for already loaded card with the same IO port and IRQ.
   	 * If found, copy its hardware configuration and use its resources
	 * (i.e. piggybacking)
	 */
	if (!card->configured)
 	{
	    for(i = 0; i < ncards; i++)
	    {
	        sdla_t *nxt_card = &card_array[i];
		if (nxt_card == card || nxt_card->next != NULL)
		    continue;

	        if (nxt_card->hw.type == SDLA_S508 &&
		    nxt_card->hw.port == conf->ioport && 
		    conf->config_id == WANCONFIG_CHDLC && 
		    nxt_card->wandev.config_id == WANCONFIG_CHDLC)
	        {
		    log(LOG_INFO, "%s: This device has the same resources like %s\n",
					wandev->name, nxt_card->devname);
		    irq = nxt_card->hw.irq;
		    bcopy(&nxt_card->hw, &card->hw, sizeof(sdlahw_t));
		    nxt_card->next = card;
		    card->next = nxt_card;
		    card->wandev.piggyback = WANOPT_YES;
	        }
	    }
	    /* Make sure I/O port region is available */
	    /* (In FreeBSD is done before kernel recompilation) */
	}
    }
    /* 
     * For a S514 adapter, check for a possible configuration error in that
     * we are loading an adapter in the same slot as a previously loaded S514
     * card.
     */ 
    else 
    {
	/* Find corresponding PCI device structure. If PCI device not exists
	 * exit!!!.
	 */ 
	if (!find_adapter(SDLA_S514, conf->PCI_slot_no, 
			  conf->S514_CPU_no[0], card->devname))
	{
	    log(LOG_ERR, "%s: Driver failed to found current PCI device!\n",
						wandev->name);
	    log(LOG_ERR, "Check your %s.conf file!\n", wandev->name);
	    splx(s);
	    return -EINVAL;
	}
	if (!card->configured)
	{
	    for (i = 0; i < ncards; i ++) 
	    {
	        sdla_t* nxt_card = &card_array[i];
                if (nxt_card == card || nxt_card->next != NULL) 
		    continue;

                if ((nxt_card->hw.type == SDLA_S514) &&
                    (nxt_card->hw.S514_slot_no == conf->PCI_slot_no) &&
                    (nxt_card->hw.S514_cpu_no[0]  == conf->S514_CPU_no[0]) &&
		    (conf->config_id == WANCONFIG_CHDLC) &&
		    (nxt_card->wandev.config_id == WANCONFIG_CHDLC)) 
                {
		    irq = nxt_card->hw.irq;
		    bcopy(&nxt_card->hw, &card->hw, sizeof(sdlahw_t));
		    nxt_card->next = card;
		    card->next = nxt_card;
		    card->wandev.piggyback = WANOPT_YES;
                }
	    }
	}
    }
    /*
     * If the current card has already configured or its a piggybacking card, 
     * do not try allocate resources...
     */ 
    if (!card->wandev.piggyback && !card->configured)
    {
	log(LOG_INFO, "%s: Starting configure hardware...\n", wandev->name);
	/* Configure hardware, load firmware, etc. */
	bzero((void*)&card->hw, sizeof(sdlahw_t));

	/* for an S514 adapter, pass the CPU number and the slot number read 
	 * from 'router.conf' to the 'sdla_setup()' function via the 'port'
	 * parameter.
         */
	if (conf->S514_CPU_no[0])
	{
	    /* Find corresponding PCI device structure and read bus address and
	     * function number.
	     */ 
	    update_pci_resource(card, conf);
	    card->hw.S514_cpu_no[0] = conf->S514_CPU_no[0];
	    card->hw.S514_slot_no = conf->PCI_slot_no;
	    log(LOG_INFO, "%s: Setting CPU to %c and SLOT to %d\n", 
		wandev->name, card->hw.S514_cpu_no[0], card->hw.S514_slot_no);
	}
	else
	{
            card->hw.port = conf->ioport;
	    card->hw.irq = conf->irq;
	}
	/* wandev->maddr already kernel virtual address */
	if (wandev->maddr)
	    card->hw.dpmbase = wandev->maddr;
	else
	    card->hw.dpmbase = NULL;

	card->hw.dpmsize = SDLA_WINDOWSIZE;
	/* set the adapter type if using an S514 adapter */
	card->hw.type = (conf->S514_CPU_no[0]) ? SDLA_S514 : conf->hw_opt[0]; 
	card->hw.pclk = conf->hw_opt[1];
	log(LOG_INFO, "%s: Starting hardware setup...\n", card->devname);
	err = sdla_setup(&card->hw, conf->data, conf->data_size);
	if (err)
	{
	    splx(s);
	    return err;
	}
        log(LOG_INFO, "%s: Set interrupt handler ...\n", card->devname);
 	err = set_intrhand(card);	
	if (err)
	{
	    splx(s);
	    return err;
	}
    }
    else
   	log(LOG_INFO, "%s: Card configured %d or Piggybacking %d!\n",
		wandev->name, card->configured, card->wandev.piggyback);

    if (!card->configured)
    {
#ifdef SMP
        /* Initialize the Spin Lock */
        log(LOG_INFO, "%s: Initializing SMP\n", wandev->name);
#else
        log(LOG_INFO, "%s: NO SMP detected\n", wandev->name);
#endif
 
        /* Intialize WAN device data space */
        wandev->dma       = 0;
        if (card->hw.type != SDLA_S514)
            wandev->ioport = card->hw.port; /* important for secondary port */
        else
        {
	    wandev->S514_cpu_no[0] = card->hw.S514_cpu_no[0];
	    wandev->S514_slot_no = card->hw.S514_slot_no;
        }
	wandev->msize     = card->hw.dpmsize;
        wandev->hw_opt[0] = card->hw.type;
        wandev->hw_opt[1] = card->hw.pclk;
        wandev->hw_opt[2] = card->hw.memory;
        wandev->hw_opt[3] = card->hw.fwid;
    }

    /* Initialize counters for each device.
     * Used for counting number if times new interface and delete interface
     * called.
     */
    wandev->new_if_cnt = 0;
    wandev->del_if_cnt = 0;
    wandev->config_id = conf->config_id; 

    /* Protocol-specific initialization */
    switch (card->hw.fwid) 
    {
#ifdef	CONFIG_WANPIPE_X25
    case SFID_X25_502:
    case SFID_X25_508:
	err = wpx_init(card, conf);
	break;
#endif

#ifdef	CONFIG_WANPIPE_FR
    case SFID_FR502:
    case SFID_FR508:
	/* 
	 * Return priority to previous value, because in this function I run
	 * interrupt test.
	 */
	splx(s);
	err = wpf_init(card, conf);
	s = splimp();
	break;
#endif

#ifdef	CONFIG_WANPIPE_PPP
    case SFID_PPP502:
    case SFID_PPP508:
	err = wpp_init(card, conf);
	break;
#endif

#ifdef CONFIG_WANPIPE_CHDLC
    case SFID_CHDLC508:
    case SFID_CHDLC514:
	if (conf->ft1)
	{
	    log(LOG_INFO, "%s: Starting FT1 Configurator\n", card->devname);
	    err = wpft1_init(card, conf);
	}
	else
	    err = wpc_init(card, conf);
	break;
#endif

#ifdef CONFIG_WANPIPE_BSTRM
    case SFID_BSC502:
	err = bsc_init(card, conf);
	break;
#endif

#ifdef CONFIG_WANPIPE_HDLC
    case SFID_HDLC508:
	err = hdlc_init(card, conf);
	break;
#endif

    default:
        log(LOG_ERR, "%s: this firmware is not supported!\n", wandev->name);
	err = -EINVAL;
    }

    if (err) 
    {
	release_hw(card);
	splx(s);
	return err;
    }

    active++;  /* increment number of active cards */

    dlog((LOG_DEBUG, "%s: Device setup is finished\n", wandev->name));
    splx(s);
    return 0;
}

/*
 ******************************************************************************
 * update_isa_resource() - Update ISA resources. 
 * o Try to find Sangoma ISA device
 * o Read I/O port, irq, memory addr to private structure.
 */
static int 
update_isa_resource(wan_device_t *wandev, wandev_conf_t *conf)
{
   sdladev_t *adapter = 
	find_adapter(SDLA_S508, conf->ioport, conf->irq, wandev->name);

    if (adapter == NULL)
	return EINVAL;
    dlog((LOG_DEBUG, "%s: ISA device structure is found\n", wandev->name));
    wandev->ioport = adapter->u.isa.ioport;
    wandev->irq    = adapter->u.isa.irq;
    wandev->maddr  = adapter->u.isa.maddr; /* already virtual address */
    return 0;
}

/*
 *****************************************************************************
 * set_intrhand() - Change interrupt handler for ISA devices. 
 * o For S508, use local structure for store pointer to real interrupt handler.
 * o For S514, register new interrupt handler for each CPU. 
 */
static int set_intrhand(sdla_t *card)
{
    sdladev_t *adapter = NULL; 

    switch(card->hw.type)
    {
    case SDLA_S508:
    	adapter = find_adapter(SDLA_S508, card->wandev.ioport, 
					card->wandev.irq, card->devname);
	break;
    case SDLA_S514:
    	adapter = 
	   find_adapter(SDLA_S514, card->hw.S514_slot_no,  
			card->hw.S514_cpu_no[0], card->devname);
	break;
    }
    if (adapter == NULL)
	return -EINVAL;
    adapter->isr  = sdla_isr;
    adapter->arg = card;
    return 0;
}

/*
 *****************************************************************************
 * restore_intrhand() - Restore interrupt handler for ISA/PCI devices.
 * o For S508, set to NULL pointer to interrupt handler in local structure.
 * o For S514, call pci_unmap_int().
 */
static int restore_intrhand(sdla_t *card)
{
    sdladev_t *adapter = NULL;

    switch(card->hw.type)
    {
    case SDLA_S508:
    	adapter = find_adapter(SDLA_S508, card->wandev.ioport, 
					card->wandev.irq, card->devname);
	break;
    case SDLA_S514:
	adapter = 
	   find_adapter(SDLA_S514, card->wandev.S514_slot_no,  
			card->wandev.S514_cpu_no[0], card->devname);
	if (adapter)
	    adapter->u.pci.cpu_no = ' ';
	break;
    }
    if (adapter == NULL)
	return -EINVAL;
    /* Restore local vector */
    adapter->isr  = NULL;
    adapter->arg  = NULL;
    return 0;
}

/*
 ***************************************************************************
 * update_pci_resource() - Update PCI resources.
 * o Find S514 in driver local structure.
 * o Store bus number, function number, pointer to configuration structure in
 *   hardware structure.  
 */
static int 
update_pci_resource(sdla_t *card, wandev_conf_t *conf)
{
    sdladev_t *adapter = NULL;

    adapter = 
	find_adapter(SDLA_S514, conf->PCI_slot_no, 
		     conf->S514_CPU_no[0], card->devname);
    if (adapter == NULL)
	return EINVAL;
    if (!adapter->u.pci.cfg)
	return EINVAL;
    
    card->hw.pci_cfg = adapter->u.pci.cfg;
    return 0;
}

/*
 ***************************************************************************
 * find_adapter() - Find S514/S508 in driver local structure. 
 * o Return error in nxt cases:
 *   o number of S514 cards is 0.
 *   o number of S514 cards is bigger than 1, but slot no is 0.
 *   o specific card not found.
 * o otherwise, return index to specific S514 card in driver local structure.
 */
static sdladev_t *
find_adapter(unsigned type, unsigned arg1, int arg2, char *devname)
{
    int i; 

    if (type == SDLA_S508)
    	dlog((LOG_DEBUG, "%s: Searching for S508 Adapter (ioport=0x%x, irq=%d)\n",
					devname, arg1, arg2)); 
    else
	dlog((LOG_DEBUG, "%s: Searching for S514 Adapter (slot_no=%d, cpu=%c)\n", 
					devname, arg1, arg2)); 
    for(i = 0; i < number_cards; i++)
    {
	if (sdladev[i].type == type)
	{
	    switch(type)
	    {
	    case SDLA_S508:
	        {
		    int ioport = arg1, irq = arg2;
		
		    dlog((LOG_DEBUG, "S508 Adapter #%d (ioport=0x%x, irq=%d)\n",
		             i, sdladev[i].u.isa.ioport, sdladev[i].u.isa.irq));
		    if ((sdladev[i].u.isa.irq == irq) && 
		        (sdladev[i].u.isa.ioport == ioport))
		        return &(sdladev[i]);
	        } 
		break;
	    
	    case SDLA_S514:
		{
		    int slot_no = arg1;
		    int  cpu_no = arg2;
		    if (number_S514_cards == 0)
    		    {
		        log(LOG_ERR, "%s: no S514 adapters found\n", devname);
		        return NULL; /* no PCI board was found */
    		    }
    		    if (number_S514_cards > 1 && !slot_no)
    		    {
		        log(LOG_ERR, "%s: More than one(%d) S514 adapter found\n", 
						devname, number_S514_cards);
		        log(LOG_ERR, "Define a PCI slot number for this adapter\n");
		        return NULL; 
    		    }
		    dlog((LOG_DEBUG, "S514 Adapter #%d (slot_no=%d, cpu=%c)\n",
		        	i, sdladev[i].u.pci.cfg->slot, sdladev[i].u.pci.cpu_no));
		    if ((sdladev[i].u.pci.cfg && 
		         sdladev[i].u.pci.cfg->slot == slot_no) ||
		        (!slot_no && number_S514_cards == 1))
		    {
			if (sdladev[i].u.pci.cpu_no == ' ')
			{
			    sdladev[i].u.pci.cpu_no = cpu_no;
	    	       	    return &(sdladev[i]);
			}
			else
			    if (sdladev[i].u.pci.cpu_no == cpu_no)
	    	        	return &(sdladev[i]);
		    }
    	        }
		break;
	    }
	}
    }
    return NULL;
}

/* 
 * ============================================================================
 * Shut down WAN link driver. 
 * o shut down adapter hardware
 * o release system resources.
 *
 * This function is called by the router when device is being unregistered or
 * when it handles ROUTER_DOWN IOCTL.
 */
static int shutdown (wan_device_t* wandev)
{
    sdla_t* card = NULL;
    int s = 0;

    dlog((LOG_DEBUG, "%s: Shutdown device(sdlamain.c)...\n", wandev->name));
    /* sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL))
	return -EFAULT;

    if (wandev->state == WAN_UNCONFIGURED)
	return 0;

    /* If we are in a critical section we lose */
    s = splimp();
    card = wandev->private;
    wandev->state = WAN_UNCONFIGURED;

    /* Release resources */
    release_hw(card);

    active--;		/* decrement number of active cards */

    if (!card->configured)
        bzero((void*)&card->hw, sizeof(sdlahw_t));
    splx(s);
    dlog((LOG_DEBUG, "%s: Shutdown finished\n", wandev->name));
    return 0;
}

/*
 * ============================================================================
 * release_hw() -
 *
 */
static void release_hw(sdla_t *card)
{
    sdla_t *nxt_card = NULL;

    dlog((LOG_DEBUG, "%s: Release hw..\n", card->devname));
    /* Check if next device exists */
    if (card->next)
    {
	nxt_card = card->next;
	/* If next device is down release resources...*/
	if (nxt_card != WAN_UNCONFIGURED)
	{
	    log(LOG_INFO, "%s: Device still running!\n", card->devname);
	    if (!card->wandev.piggyback)
    	    {  /* Change interrupt vector argument to piggybacking card */
		set_intrhand(nxt_card);
		nxt_card->wandev.piggyback = WANOPT_NO;
	    }
	    else
		card->wandev.piggyback = WANOPT_NO;
	    /* `Disconnect' two cards: primary & seconadary */
	    card->next = NULL;
	    nxt_card->next = NULL;
	    card->configured = WANOPT_NO;
	}
	else
	    log(LOG_INFO, "Uops... never can be heppened, but...\n");
    }
    else
    {
	log(LOG_INFO, "%s: Master device shutting down!\n", card->devname);
        /* Free irq !!! */
        restore_intrhand(card);
	sdla_down(&card->hw);
	card->configured = 0;
    }
}

/*
 * ============================================================================
 * Driver I/O control. 
 * o verify arguments
 * o perform requested action
 *
 * This function is called when router handles one of the reserved user
 * IOCTLs.  Note that 'arg' stil points to user address space.
 */
static int ioctl (wan_device_t* wandev, u_long cmd, caddr_t arg)
{
    sdla_t* card = NULL;
    int err;
    int s = 0;

    /* sanity checks */
    if ((wandev == NULL) || (wandev->private == NULL))
	return -EFAULT;
    if (wandev->state == WAN_UNCONFIGURED)
	return -ENODEV;

    card = wandev->private;
    s = splimp();

    switch (cmd) 
    {
    case WANPIPE_DUMP:
	err = ioctl_dump(wandev->private, (void*)arg);
	break;

    case WANPIPE_EXEC:
	err = ioctl_exec(wandev->private, (void*)arg);
	break;

    default:
	err = -EINVAL;
    }

    splx(s);
    return err;
}

/****** Driver IOCTL Handlers ***********************************************/

/* 
 * ============================================================================
 * Dump adapter memory to user buffer.
 * o verify request structure
 * o copy request structure to kernel data space
 * o verify length/offset
 * o verify user buffer
 * o copy adapter memory image to user buffer
 *
 * Note: when dumping memory, this routine switches curent dual-port memory
 *	 vector, so care must be taken to avoid racing conditions.
 */
static int ioctl_dump (sdla_t* card, sdla_dump_t* u_dump)
{
    sdla_dump_t dump;
    unsigned winsize;
    unsigned long oldvec;	/* DPM window vector */
    int err = 0;
    int s = 0;

    if (u_dump == NULL)
	return -EFAULT;
    /* Copy from user space to kernel */
    if(copyin((void*)u_dump, (void*)&dump, sizeof(sdla_dump_t)))
	return -EFAULT;
		
    if ((dump.magic != WANPIPE_MAGIC) || 
	(dump.offset + dump.length > card->hw.memory))
	return -EINVAL;
	
    if (dump.ptr == NULL) 
	return -EFAULT;

    winsize = card->hw.dpmsize;
    s = splimp();	/* critical section */

    if (card->hw.type != SDLA_S514) 
    {
	oldvec = card->hw.vector;
        while (dump.length) 
	{
	    /* current offset */
            unsigned pos = dump.offset % winsize;
	    /* current vector */
            unsigned long vec = dump.offset - pos;
            unsigned len = (dump.length > (winsize - pos)) ?
                     	(winsize - pos) : dump.length;
	    /* relocate window */
            if (sdla_mapmem(&card->hw, vec) != 0) 
	    {
		err = -EIO;
                break;
            }
            /* FIXME::: COPY TO KERNEL BUFFER FIRST ?? */
            if (copyout((void*)((char*)card->hw.dpmbase + pos), 
						(void*)dump.ptr, len))
	    {
		splx(s);
		return -EFAULT; 
	    } 
            dump.length     -= len;
            dump.offset     += len;
            (char*)dump.ptr += len;
        }
	sdla_mapmem(&card->hw, oldvec);/* restore DPM window position */
    }
    else 
    {
	/* FIXME::: COPY TO KERNEL BUFFER FIRST ?? */
	if (copyout((void*)((char*)card->hw.dpmbase + dump.offset), 
				(void*)dump.ptr, dump.length))
	    err = -EFAULT; 
    }

    splx(s);		/* >>> critical section end <<< */
    return err;
}

/* 
 * ============================================================================
 * Execute adapter firmware command.
 * o verify request structure
 * o copy request structure to kernel data space
 * o call protocol-specific 'exec' function
 */
static int ioctl_exec (sdla_t* card, sdla_exec_t* u_exec)
{
    sdla_exec_t exec;

    if (card->exec == NULL)
	return -ENODEV;
	
    if (u_exec == NULL) 
	return -EFAULT;
    if (copyin((void*)u_exec, (void*)&exec, sizeof(sdla_exec_t)))
	return -EFAULT;

    if ((exec.magic != WANPIPE_MAGIC) || (exec.cmd == NULL))
	return -EINVAL;
    return card->exec(card, exec.cmd, exec.data);
}

/******* Miscellaneous ******************************************************/

/*
 * ============================================================================
 * SDLA Interrupt Service Routine (S514).
 *  o acknoledge SDLA hardware interrupt.
 *  o call protocol-specific interrupt service routine, if any.
 *  o Return 1, for internal error; otherwise - 0;
 */
static void sdla_isr(void *arg)
{
    sdla_t *card = (sdla_t*)arg;

    if (card == NULL)
    {
	log(LOG_EMERG, "%s?: Internal Error (line=%d)\n", modname, __LINE__); 
	return; 
    }

    if (card->hw.type == SDLA_S514)
    { /* handler interrupt on S514 */
       	uint32_t int_status;
    	unsigned char CPU_no = card->hw.S514_cpu_no[0];
    	unsigned char card_found_for_IRQ;
    	uint8_t  IRQ_count = 0;
      	for(;;)
    	{
            read_S514_int_stat(&card->hw, &int_status);
            /* check if the interrupt is for this device */
            if (!((unsigned char)int_status & (IRQ_CPU_A | IRQ_CPU_B)))
	    	return;
            /* If the IRQ is both CPUs on the same addapter, then alter
             * the interrupt status so as handle one CPU at a time.
             */
            if (((unsigned char)int_status &(IRQ_CPU_A | IRQ_CPU_B)) ==
					(IRQ_CPU_A | IRQ_CPU_B))
		int_status &= (CPU_no == S514_CPU_A) ? ~IRQ_CPU_B : ~IRQ_CPU_A;
            card_found_for_IRQ = 0;
            /* Check to see that the CPU number for this device corresponds to
             * the interrupt status read.
             */
            switch(CPU_no)
            {
            case S514_CPU_A:
    	    	if ((unsigned char)int_status & IRQ_CPU_A)
	            card_found_for_IRQ = 1;
    	    	break;
            case S514_CPU_B:
    	    	if ((unsigned char)int_status & IRQ_CPU_B)
	            card_found_for_IRQ = 1;
    	        break;
            }
            /* exit if the interrupt is for another CPU on the same IRQ */
            if (!card_found_for_IRQ)
	        return;
            if (card->wandev.state == WAN_UNCONFIGURED && !card->configured)
            {
	        log(LOG_INFO, "%s: IRQ for unconfigured adapter.\n", card->devname);
	        S514_intack(&card->hw, int_status);
	        return;
            }
	    if (card->in_isr)
	    {
	        log(LOG_INFO, "%s: interrupt re-entrancy on IRQ %d\n",
					card->devname, card->wandev.irq);
	        S514_intack(&card->hw, int_status);
	        return;
	    }
    	    S514_intack(&card->hw, int_status);

    	    if (card->isr)
	        card->isr(card);
            /* Handle a maximum of two interrupts (on for each CPU on the 
     	     * adapter) before returning.
             */
    	    if ((++IRQ_count) == 2)
	        return;
        }
    }
    else
    {
        if (card->wandev.state == WAN_UNCONFIGURED)
            return;
        if (card->in_isr)
        {
            log(LOG_INFO, "%s: interrupt re-entrancy on IRQ %d!\n",
				card->devname, card->wandev.irq);
            return;
        }
        sdla_intack(&card->hw);
        if (card->isr)
            card->isr(card);
    }
    return;
}

/* 
 * ============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being open.  
 */
void wanpipe_open (sdla_t* card)
{
	++card->open_cnt;
}

/* 
 * ============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being closed.  
 */
void wanpipe_close (sdla_t* card)
{
	--card->open_cnt;
}

/* 
 * ============================================================================
 * Set WAN device state.
 */
void wanpipe_set_state (sdla_t* card, int state)
{
    unsigned long flags;

    flags = read_eflags();
    disable_intr();
    if (card->wandev.state != state) 
    {
	switch (state) 
	{
	case WAN_CONNECTED:
	    log(LOG_INFO, "%s: link connected!\n", card->devname);
	    break;

	case WAN_CONNECTING:
	    log(LOG_INFO, "%s: link connecting...\n", card->devname);
	    break;

        case WAN_DISCONNECTED:
	    log(LOG_INFO, "%s: link disconnected!\n", card->devname);
	    break;
	}
	card->wandev.state = state;
    }
    card->state_tick = SYSTEM_TICK;
    write_eflags(flags);
}

/*
 ***************************************************************************
 * in_ntoa() - Convert IP address to string  
 */
char *in_ntoa(uint32_t in)
{
    static char buf[18];
    char *p = NULL;

    p = (char*)&in;
    sprintf(buf, "%d.%d.%d.%d",
         (p[0] & 0xFF), (p[1] & 0xFF), (p[2] & 0xFF), (p[3] & 0xFF));

    return buf;
}

/****** End *****************************************************************/
