/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wanmain.c,v 1.1.1.1 2000/01/17 22:01:24 alexf Exp $
 */

/*****************************************************************************
 * wanmain.c	WAN Multiprotocol Router Module. Main code.
 *
 *		This module is completely hardware-independent and provides
 *		the following common services for the WAN Link Drivers:
 *		 o WAN device managenment (registering, unregistering)
 *		 o Network interface management
 *		 o Physical connection management (dial-up, incomming calls)
 *		 o Logical connection management (switched virtual circuits)
 *
 * Author:	Alex Feldman	<al.feldmang@sangoma.com>
 *
 * ============================================================================
 * Dec 12, 1999	Alex Feldman	Updated for FreeBSD 3.2 
 * Dec 27, 1996	Gene Kozin	Initial version (based on Sangoma's WANPIPE)
 ******************************************************************************
 */

#include "bpfilter.h" 

#include <stddef.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <sys/conf.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/module.h>

#include <net/if.h>
#include <net/route.h>
#include <net/wanpipe.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <machine/cpufunc.h>

#include "wanr_cdev.h"

/************************* Defines and Macros ****************************/ 

/* For debugging convenience */
#ifdef WANDEBUG
#define dlog(x) log x
#else
#define dlog(x)
#endif /* WANDEBUG */ 

#ifndef	min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef	max
#define max(a,b) (((a)>(b))?(a):(b))
#endif

/*
 * 	Function Prototypes 
 */

static int wanrouter_load __P((module_t, int, void*));

/* 
 *	WAN device IOCTL handlers 
 */

static int device_setup(wan_device_t *wandev, wandev_conf_t *u_conf);
static int device_stat(wan_device_t *wandev, wandev_stat_t *u_stat);
static int device_shutdown(wan_device_t *wandev);
static int device_new_if(wan_device_t *wandev, wanif_conf_t *u_conf);
static int device_del_if(wan_device_t *wandev, char *u_name);
static void wanpipe_attach(struct ifnet *ifp);
static void wanpipe_detach(struct ifnet *ifp);
 
/* 
 *	Miscellaneous 
 */

static wan_device_t *find_device (char *name); 
static int delete_interface (wan_device_t *wandev, char *name, int force);

/*
 *	Global Data
 */
static char fullname[]		= "WAN Router";
static char copyright[]		= "(c) 1995-1999 Sangoma Technologies Inc.";
static char modname[]		= ROUTER_NAME;	/* short module name */
wan_device_t* router_devlist 	= NULL;	/* list of registered devices */
static int devcnt 		= 0;

/********************* Kernel Loadable Module Entry Points *******************/
/* 
 *=============================================================================
 * 	Module entry point.
 *
 * 	Return:	0	Ok
 *		< 0	error.
 * 	Context:	process
 */
int wanrouter_load (module_t mod, int cmd, void *arg)
{
    int err = 0;

    switch(cmd)
    {
    case MOD_LOAD:
        /* 'insert' entry point.
         * 	o print announcement
         * 	o initialize static data
         */
        log(LOG_INFO, "%s v%u.%u %s\n", 
		fullname, ROUTER_VERSION, ROUTER_RELEASE, copyright);
        break;
    case MOD_UNLOAD:
        /*
         * 'remove' entry point.
         */
        break;
    }
    return err;
}
CDEV_MODULE(wanrouter, CDEV_MAJOR, wanrouter_devsw, wanrouter_load, NULL);

/*
 * 	Kernel APIs
 */

/* 
 * ============================================================================
 * Register WAN device.
 * o verify device credentials
 * o create an entry for the device in the /proc/net/router directory(LINUX)
 * o initialize internally maintained fields of the wan_device structure
 * o link device data space to a singly-linked list
 * o if it's the first device, then start kernel 'thread'
 * o increment module use count
 *
 * Return:
 * 0	Ok
 * < 0	error.
 *
 * Context:	process
 */
int register_wan_device(wan_device_t *wandev)
{
    int namelen;

    if ((wandev == NULL) || (wandev->magic != ROUTER_MAGIC) ||
        (wandev->name == NULL))
	return -EINVAL;
 
    namelen = strlen(wandev->name);
    if (!namelen || (namelen > WAN_DRVNAME_SZ))
	return -EINVAL;
	
    if (find_device(wandev->name) != NULL)
	return -EEXIST;

    dlog((LOG_DEBUG, "%s: registering WAN device %s\n", modname, wandev->name));

    /*
     *	Initialize fields of the wan_device structure maintained by the
     *	router and update local data.
     */
 
    wandev->ndev    = 0;
    wandev->if_head = NULL;
    wandev->next    = router_devlist;
    router_devlist  = wandev;
    ++devcnt;
    return 0;
}

/* 
 * ============================================================================
 *	Unregister WAN device.
 *	o shut down device
 *	o unlink device data space from the linked list
 *	o delete device entry in the /proc/net/router directory
 *	o decrement module use count
 *
 *	Return:		0	Ok
 *			<0	error.
 *	Context:	process
 */

int unregister_wan_device(char *name)
{
    wan_device_t *wandev = NULL, *prev = NULL;
    char if_name[IFNAMSIZ+1];

    if (name == NULL)
	return -EINVAL;

    for (wandev = router_devlist, prev = NULL;
	wandev && strcmp(wandev->name, name);
	prev = wandev, wandev = wandev->next);
    if (wandev == NULL)
	return -ENODEV;

    dlog((LOG_DEBUG, "%s: unregistering WAN device %s\n", modname, wandev->name));

    if (wandev->state != WAN_UNCONFIGURED) 
    {
	struct ifnet *ifp = NULL;
        while(wandev->if_head)
	{
	    bzero(if_name, IFNAMSIZ+1);
	    if (wandev->if_head && wandev->if_head->ifp)
	    {
	        ifp = wandev->if_head->ifp;   
		sprintf(if_name, "%s%d", ifp->if_name, ifp->if_unit);
		delete_interface(wandev, if_name, 1);
            }
	    else
            { /* Vary bad case */
		struct ifnet_list *tmp = wandev->if_head;
		wandev->if_head = wandev->if_head->pnext;
		if (tmp)
		    free(tmp, M_DEVBUF);
            }
	}
        if (wandev->shutdown)	
	    wandev->shutdown(wandev);
    }
    if (prev)
        prev->next = wandev->next;
    else
        router_devlist = wandev->next;
    --devcnt;
    return 0;
}

int wanrouter_open(dev_t dev, int flag, int otyp, struct proc *procp)
{
   return (0);
}
int wanrouter_close(dev_t dev, int flag, int otyp, struct proc *procp)
{   
   return (0);
}
/*
 * ============================================================================
 * 	WAN device IOCTL.
 *	o find WAN device associated with this node
 *	o execute requested action or pass command to the device driver
 */
int wanrouter_ioctl( dev_t dev, u_long cmd, caddr_t arg, int mode, struct proc *procp)
{
    int err = 0;
    wan_device_t *wandev = NULL;
    wan_conf_t *cfg = NULL;

    if (((cmd >> 8) & ROUTER_IOCTL) != ROUTER_IOCTL)
	return -EINVAL;
    if (arg == NULL)
	return -EINVAL;
    /* The arg is points already to kernel space */
    cfg = (wan_conf_t*)arg;

    wandev = find_device(cfg->devname);
    if (wandev == NULL || wandev->magic != ROUTER_MAGIC)
        return -EINVAL;
		
    switch (cmd) 
    {
    case ROUTER_SETUP:
	dlog((LOG_DEBUG, "%s: ROUTER_SETUP \n", cfg->devname));
	err = device_setup(wandev, cfg->arg);
	break;

    case ROUTER_DOWN:
	dlog((LOG_DEBUG, "%s: ROUTER_DOWN\n", cfg->devname));
	err = device_shutdown(wandev);
	break;

    case ROUTER_STAT:
	dlog((LOG_DEBUG, "%s: ROUTER_STAT\n", cfg->devname)); 
	err = device_stat(wandev, cfg->arg);
	break;

    case ROUTER_IFNEW:
	dlog((LOG_DEBUG, "%s: ROUTER_IFNEW\n", cfg->devname));
	err = device_new_if(wandev, cfg->arg);
	break;

    case ROUTER_IFDEL:
	dlog((LOG_DEBUG, "%s: ROUTER_IFDEL\n", cfg->devname));
	err = device_del_if(wandev, cfg->devname);
	break;

    case ROUTER_IFSTAT:
        break;

    default:
	dlog((LOG_DEBUG, "%s: default\n", cfg->devname));
	if (wandev->ioctl)
	    err = wandev->ioctl(wandev, cmd, cfg->arg);
	else err = -EINVAL;
    }
    return err;
}

/*
 *	WAN Driver IOCTL Handlers
 */

/* 
 * ============================================================================
 *	Setup WAN link device.
 *	o verify user address space
 *	o allocate kernel memory and copy configuration data to kernel space
 *	o if configuration data includes extension, copy it to kernel space too
 *	o call driver's setup() entry point
 */

static int device_setup (wan_device_t *wandev, wandev_conf_t *u_conf)
{
    void *data = NULL;
    wandev_conf_t *conf = NULL;
    int err = 0;

    dlog((LOG_DEBUG, "%s: Device Setup...\n", wandev->name));
    if (wandev->setup == NULL)	/* Nothing to do ? */
	return 0;

    conf = malloc(sizeof(wandev_conf_t), M_DEVBUF, M_NOWAIT);
    if (conf == NULL)
    {
	log(LOG_ERR, "%s: Cannot allocate memory for configuration structure\n",
				wandev->name);
	return -ENOMEM;
    }

    if (copyin((void *)u_conf, (void*)conf, sizeof(wandev_conf_t)))
    {
	free(conf, M_DEVBUF);
    	return -EFAULT;
    }
	
    if (conf->magic != ROUTER_MAGIC) 
    {
        log(LOG_ERR, "%s: Wrong configuration structure\n", wandev->name);	
        free(conf, M_DEVBUF);
        return -EINVAL; 
    }

    if (conf->data_size && conf->data) 
    {
	if (conf->data_size > 128000 || conf->data_size < 0) 
	{
	    log(LOG_ERR, "%s: Large configuration structure\n", wandev->name);
            free(conf, M_DEVBUF);
	    return -EINVAL;;
	}
        data = malloc(conf->data_size, M_DEVBUF, M_NOWAIT);
        if (data) 
	{
	    if (!copyin((void*)conf->data, data, conf->data_size))
	    {
		conf->data = data;
                err = wandev->setup(wandev, conf);
	    }
	    else
		err = -EFAULT;
        }
        else
        {
            log(LOG_ERR, "%s: Cannot allocate data buffer\n", wandev->name);
	    err = -ENOMEM;
        }
        if (data)
            free(data, M_DEVBUF);
    }

    if (conf)
    	free(conf, M_DEVBUF);
    dlog((LOG_DEBUG, "%s: Device setup finished (%d)\n", wandev->name, err));
    return err;
}

/* 
 * ============================================================================
 *	Shutdown WAN device.
 *	o delete all not opened logical channels for this device
 *	o call driver's shutdown() entry point
 */
 
static int device_shutdown (wan_device_t *wandev)
{
    struct ifnet_list *ifnode = NULL, *prev = NULL;
    struct ifnet *ifp = NULL;
    char if_name[IFNAMSIZ+1];

    dlog((LOG_DEBUG, "%s: device_shutdown called\n", wandev->name));	
    if (wandev->state == WAN_UNCONFIGURED)
	return 0;

    /* Delete all interfaces */
    ifnode = wandev->if_head;
    prev = NULL;
    while(ifnode)
    {
	bzero(if_name, IFNAMSIZ+1);
        ifp = ifnode->ifp;
        if (ifp)
	{
	    sprintf(if_name, "%s%d", ifp->if_name, ifp->if_unit);
            if (delete_interface(wandev, if_name, 0))
		prev = ifnode;
	}
	else
	{ /* Bad case or error */
	    struct ifnet_list *tmp = ifnode;
	    if (prev)
	        prev->pnext = ifnode->pnext;
    	    else	
	        wandev->if_head = ifnode->pnext;
	    if (tmp)
	    	free(tmp, M_DEVBUF);
	}
	ifnode = (prev) ? prev->pnext : wandev->if_head;
    }
    if (wandev->ndev)
	return -EBUSY;	/* there are opened interfaces  */
		
    if (wandev->shutdown)
	return wandev->shutdown(wandev);
    dlog((LOG_DEBUG, "%s: device_shutdown end\n", wandev->name));	
    return 0;
}

/* 
 * ============================================================================
 *	Get WAN device status & statistics.
 */

static int device_stat (wan_device_t *wandev, wandev_stat_t *u_stat)
{
    wandev_stat_t stat;

    bzero((void*)&stat, sizeof(stat));

    /* Ask device driver to update device statistics */
    if ((wandev->state != WAN_UNCONFIGURED) && wandev->update)
	wandev->update(wandev);

    /* Fill out structure */
    stat.ndev  = wandev->ndev;
    stat.state = wandev->state;

    if (copyout((void*)&stat, (void*)u_stat, sizeof(stat)))
	return -EFAULT;

    return 0;
}

/* 
 * ============================================================================
 *	Create new WAN interface.
 *	o verify user address space
 *	o copy configuration data to kernel address space
 *	o allocate network interface data space
 *	o call driver's new_if() entry point
 *	o make sure there is no interface name conflict
 *	o register network interface
 */

static int device_new_if (wan_device_t *wandev, wanif_conf_t *u_conf)
{
    wanif_conf_t conf;
    struct ifnet *ifp = NULL;
    char ifname[IFNAMSIZ+1];
    int interface_exist = 0;
    int err;
    int s = splimp();

    log(LOG_INFO, "%s: Creating new WAN interface %s...\n", 
						wandev->name, u_conf->name);
    if ((wandev->state == WAN_UNCONFIGURED) || (wandev->new_if == NULL))
    {
	splx(s);
	return -ENODEV;
    }

    if (copyin((void*)u_conf, (void*)&conf, sizeof(wanif_conf_t)))
    {
	splx(s);
	return -EFAULT;
    }
		
    if (conf.magic != ROUTER_MAGIC)
    {
	splx(s);
	return -EINVAL;
    }

    /* The network interface contanins of the interface in two forms:
     * character string plus a unit number for the driver and binary system
     * wide index number. The user define the network interface name, he 
     * responsible that these names must be different. The unit number will
     * always 0 (I cannot do nothing with this.
     */
    if (conf.config_id == WANCONFIG_FR)
    {
	int dlci, ind;
	char c;

	sscanf(conf.addr, "%d", &dlci);
	ind = strlen(conf.name);
        while((ind != 0) && (c = conf.name[ind-1]))
	{
	    if ((c < '0') || (c > '9'))
		break;
	    conf.name[ind-1] = '\0';
	    ind--;    
	}
	conf.name[ind] = '\0';
        sprintf(ifname, "%s%d", conf.name, dlci);
    }
    else 
        sprintf(ifname, "%s%d", conf.name, 0);
    ifp = ifunit(ifname);
    if (ifp)
        interface_exist = 1;    
    else
    {
        ifp = malloc(sizeof(struct ifnet), M_DEVBUF, M_NOWAIT);
        if (ifp == NULL)
        {
   	    log(LOG_ERR, "Cannot allocate memory for network interface!\n");
	    splx(s);
	    return -ENOMEM;
        }
	bzero(ifp, sizeof(struct ifnet));
        interface_exist = 0;
    }
 
    err = wandev->new_if(wandev, ifp, &conf);
    if (!err) 
    {
	/* Register network interface. This will invoke init()
	 * function supplied by the driver.  If device registered
	 * successfully, add it to the interface list.
	 */
	if (ifp->if_name == NULL)
	    err = -EINVAL;
	else 
	{
	    struct ifnet_list *if_node = NULL;
            if_node = malloc(sizeof(struct ifnet_list), M_DEVBUF, M_NOWAIT);
	    if (if_node == NULL)
	    {
   	        log(LOG_ERR, "Cannot allocate memory for network interface list!\n");
	        splx(s);
	        return -ENOMEM;
	    }
            if_node->ifp = ifp;	
	    if_node->pnext = wandev->if_head;
	    wandev->if_head = if_node;
	    log(LOG_INFO, "Attaching network interface %s%d\n", 
					       ifp->if_name, ifp->if_unit);
            if (!interface_exist)
		wanpipe_attach(ifp);
	    ++wandev->ndev;
	    splx(s);
	    return 0;	/* done !!! */
	}
	if (wandev->del_if)
	    wandev->del_if(wandev, ifp);
    }
    free(ifp, M_DEVBUF);
    log(LOG_INFO, "%s: !!! Failed to dprintfte new interface (%d)!!!\n",
							wandev->name, err);
    splx(s);
    return err;
}


/* 
 * ============================================================================
 *	Delete WAN logical channel.
 *	 o verify user address space
 *	 o copy configuration data to kernel address space
 */

static int device_del_if (wan_device_t *wandev, char *u_name)
{
    char name[WAN_IFNAME_SZ + 1];

    if (wandev->state == WAN_UNCONFIGURED)
	return -ENODEV;
	
    bzero((void*)name, sizeof(name));

    if (copyin((void*)u_name, (void*)name, WAN_IFNAME_SZ))
	return -EFAULT;

    return delete_interface(wandev, name, 0);
}


/*
 *	Miscellaneous Functions
 */

/* 
 * ===========================================================================
 *	Find WAN device by name.
 *	Return pointer to the WAN device data space or NULL if device not found.
 */

static wan_device_t *find_device(char *name)
{
    wan_device_t *wandev = NULL;

    for (wandev = router_devlist;wandev && strcmp(wandev->name, name);
	wandev = wandev->next);
    return wandev;
}

/* 
 * ============================================================================
 *	Delete WAN logical channel identified by its name.
 *	o find logical channel by its name
 *	o call driver's del_if() entry point
 *	o unregister network interface
 *	o unlink channel data space from linked list of channels
 *	o release channel data space
 *
 *	Return:	0		success
 *		-ENODEV		channel not found.
 *		-EBUSY		interface is open
 *
 *	Note: If (force != 0), then device will be destroyed even if interface
 *	associated with it is open. It's caller's responsibility to make
 *	sure that opened interfaces are not removed!
 */

static int delete_interface (wan_device_t *wandev, char *name, int force)
{
    struct ifnet *ifp = NULL;
    struct ifnet_list *if_node = NULL,
		      *prev    = NULL;
    char if_name[IFNAMSIZ+1];
    int s = 0;

    log(LOG_INFO, "%s: Deleting interface %s...\n", wandev->name, name);
    /* Searching for all interfaces */
    for(if_node = wandev->if_head, prev = NULL; if_node && if_node->ifp; 
	prev = if_node, if_node = if_node->pnext)
    {
	sprintf(if_name, "%s%d", if_node->ifp->if_name, if_node->ifp->if_unit);
	if ((strcmp(name, if_name) == 0))
	    break;
    }
    if (if_node == NULL)
    {
	log(LOG_ERR, "%s: Network interface %s is not found.\n", wandev->name, name);
	return -ENODEV;
    } 
    ifp = if_node->ifp;
    if (ifp->if_flags & IFF_UP == IFF_UP) 
    {
	if (force) 
	    log(LOG_INFO, "%s: deleting opened interface %s!\n", modname, name);
	else
  	    return -EBUSY;	/* interface in use */
	ifp->if_flags &= ~IFF_UP;
    }

    if (wandev->del_if)
	wandev->del_if(wandev, ifp);
    if_node->ifp = NULL;
 
    s = splimp(); 			/***** critical section start *****/
    if (prev)
	prev->pnext = if_node->pnext;
    else
	wandev->if_head = if_node->pnext;
    --wandev->ndev;
    splx(s); 				/****** critical section end ******/

    wanpipe_detach(ifp);
   
    free(if_node, M_DEVBUF);
    log(LOG_INFO, "%s: Interface %s deleted.\n", wandev->name, name);

    return 0;
}

static void wanpipe_attach(struct ifnet *ifp)
{
    if_attach(ifp);
#if NBPFILTER > 0
    log(LOG_INFO, "%s%d: Attaching to BPF ...\n", ifp->if_name, ifp->if_unit);
    bpfattach(ifp, DLT_NULL, 4);
#endif
}

static void wanpipe_detach(struct ifnet *ifp)
{
    struct ifaddr *ifa = NULL, *dst_ifa = NULL;
    int s;

    s = splnet();
    /* Remove addresses from routing table */
    ifa = TAILQ_FIRST(&ifp->if_addrhead);
    for(; ifa; ifa = TAILQ_NEXT(ifa, ifa_link))
	rtinit(ifa, (int)RTM_DELETE, 0);
    if ((ifa = TAILQ_FIRST(&ifp->if_addrhead)) != NULL)
        if ((dst_ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL)
    	    TAILQ_REMOVE(&ifp->if_addrhead, dst_ifa, ifa_link);
 
#ifdef IF_DETACH
    TAILQ_REMOVE(&ifnet, ifp, if_link);
    free(ifp, M_DEVBUF);
#endif
    splx(s);
}

/************************ End of Wanmain.c ******************************/
